package dtl_algo;

import java.io.File;

import java.io.IOException;

import java.io.PrintWriter;
import java.math.BigInteger;
import java.util.*;
import mapping.*;
import node.DomainNodeRooted;
import node.GeneNodeRooted;
import node.Node;
import node.SpeciesNodeRooted;

import tree.*;
import utility.FileUtility;;


public class DTLReconciliation {
	SpeciesTreeRooted speciesTree; // species tree instance
	
	ArrayList<GeneTreeRooted>geneTrees = new ArrayList<GeneTreeRooted>();
	ArrayList<DomainTreeRooted>domainTrees = new ArrayList<DomainTreeRooted>();
	public static int DOMAIN_DUPLICATION = 2; // cost of domain duplication
	public static int DOMAIN_LOSS = 1; // cost of domain loss
	
	public static int INTRA_SPECIES_DOMAIN_TRANSFER_SAME =3; // cost of intra species transfer within same gene family
	public static int INTRA_SPECIES_DOMAIN_TRANSFER_DIFFERENT = 4; // cost of intra species transfer between different gene family
	public static int INTER_SPECIES_DOMAIN_TRANSFER_SAME =4; // cost of inter species transfer within same gene family
	public static int INTER_SPECIES_DOMAIN_TRANSFER_DIFFERENT =5; // cost of inter species transfer between different gene family
	
	public static int GENE_TRANSFER = 3; // cost of gene transfer event
	public static int GENE_DUPLICATION = 2; // cost of gene duplication
	public static int GENE_LOSS = 1; // cost of gene loss
	int reconciliationCost; // final cost of reconciliation
	HashMap<Node, ArrayList<RootedBinaryTreeMapping>> optimalMappings = new HashMap<>(); // list of all optimal mappings
	DomainMapping[][] allPossibleDomainMappings;
	GeneMapping[][] allPossibleGeneMappings;
	boolean inputError = false;
	int numberOfTransfer = 0; //number of transfer events
	int numberOfDuplication = 0; // number of duplication events
	public boolean errorFlag = false;

	private int geneSpeciesReconciliationCost;
	// setters and getters of all species tree and gene tree
	public SpeciesTreeRooted getSpeciesTree() {
		return speciesTree;
	}

	public void setSpeciesTree(SpeciesTreeRooted speciesTree) {
		this.speciesTree = speciesTree;
	}
	public DomainTreeRooted getSingleDomainTree() {
		return domainTrees.get(0);
	}
	

	public ArrayList<GeneTreeRooted> getGeneTrees() {
		return geneTrees;
	}

	public void setGeneTrees(ArrayList<GeneTreeRooted> geneTrees) {
		this.geneTrees = geneTrees;
	}

	public boolean isInputError() {
		return inputError;
	}

	public void setInputError(boolean inputError) {
		this.inputError = inputError;
	}

	
	public static void main(String[] args) throws IOException {

		String speciesTreeFileName = "";
		String domainTreeFileName = "";
		String geneTreeFolderName = "";
		String outputFileName = "";
		String heuristic = "improvedApprox";
		for (int i = 0; i < args.length; i++) {
			if (args[i].equals("-s")) {
				i++;
				speciesTreeFileName = args[i];
			} else if (args[i].equals("-d")) {
				i++;
				domainTreeFileName = args[i];
			}else if (args[i].equals("-g")) {
				i++;
				geneTreeFolderName = args[i];
			}
			else if (args[i].equals("-o")) {
				i++;
				outputFileName = args[i];
			}
			
			else if (args[i].equals("-DL")) {
				i++;
				DOMAIN_LOSS = Integer.parseInt(args[i]);
			} else if (args[i].equals("-DD")) {
				i++;
				DOMAIN_DUPLICATION = Integer.parseInt(args[i]);
			} else if (args[i].equals("-DTWW")) {
				i++;
				INTRA_SPECIES_DOMAIN_TRANSFER_SAME = Integer.parseInt(args[i]);
			} 
			else if (args[i].equals("-DTAW")) {
				i++;
				INTRA_SPECIES_DOMAIN_TRANSFER_DIFFERENT = Integer.parseInt(args[i]);
			}
			else if (args[i].equals("-DTWA")) {
				i++;
				INTER_SPECIES_DOMAIN_TRANSFER_SAME = Integer.parseInt(args[i]);
			}
			else if (args[i].equals("-DTAA")) {
				i++;
				INTER_SPECIES_DOMAIN_TRANSFER_DIFFERENT = Integer.parseInt(args[i]);
			}
			else if(args[i].equals("-GT")) {
				i++;
				GENE_TRANSFER = Integer.parseInt(args[i]);
			}
			else if(args[i].equals("-GL")) {
				i++;
				GENE_LOSS = Integer.parseInt(args[i]);
			}
			else if(args[i].equals("-GD")) {
				i++;
				GENE_DUPLICATION = Integer.parseInt(args[i]);
			}
			else if (args[i].equals("-a")) {
				i++;
				heuristic = args[i];
			}
			else if (args[i].equals("-h")) {
				System.out.println("List of arguments: ");
				System.out.println("-d The input domain tree file.\r\n"
						+ "-g Path to the directory containing the gene trees.\r\n"
						+ "-s The input species tree file.\r\n"
						+ "-o Output file name. By default the output file will be the input domain tree file name plus "
						+ "\".output\".\r\n"
						+"-a Algorithm selected for performing Gen-DGS reconciliation: 'simpleApprox' for selecting 'SimpleApprox' algorithm "
						+ "and \n 'improvedApprox' for selecting 'ImprovedAprrox' algorithm. Default is 'ImprovedApprox' algorithm\n"
						+ "-DD Domain duplication cost. Default value 2.\r\n"
						+ "-DL Domain loss cost. Default value 1.\r\n"
						+ "-DTWW Domain transfer cost when the donor and recipient are in the same gene family and within the same species. Default value 3.\r\n"
					+ "-DTAW Domain transfer cost when the donor and recipient are in different gene family but within the same species. Default value 4.\r\n"
					+ "-DTWA Domain transfer cost when the donor and recipient are in the same gene family but in different species. Default value 4.\r\n"
					+ "-DTAA Domain transfer cost when the donor and recipient are in different gene family and in different species. Default value 5.\r\n"
						+ "-GD Gene duplication cost. Default value 2.\r\n"
						+ "-GL Gene loss cost. Default value 1.");
			}
		}
		if(outputFileName.isEmpty()) {
			outputFileName = domainTreeFileName+".output";
		}
		PrintWriter writer = new PrintWriter(new File(outputFileName));
		DTLReconciliation reconciliation = new DTLReconciliation();
		FileUtility.processInput(speciesTreeFileName,geneTreeFolderName, domainTreeFileName, reconciliation);
		if (reconciliation.isInputError())
			return;
		reconciliation.performPreprocessings();
		if(heuristic.equals("simpleApprox"))
			reconciliation.runApproximationAlgorithm(writer);
		else if(heuristic.equals("improvedApprox")) {
			reconciliation.runExtendedDGTL(writer);
		}
	
		writer.close();
	
	
	}




	private void runApproximationAlgorithm(PrintWriter writer) {
    	// create a mapping from each gene node to every species node
    	initializeAllGeneMappings();
    	//reconcile every gene tree with species tree
    	for(GeneTreeRooted geneTree: geneTrees) {
    		reconcileGeneTree(geneTree);
    		 sampleGreedyOptionSolutionOfGeneTree(geneTree);
   	}
    	
    	geneSpeciesReconciliationCost = reconciliationCost;
    	// create a mapping from each domain node to every gene node
		initializeAllDomainMappings();
		for(DomainTreeRooted domainTree: domainTrees) {
			reconcileDomainTreeApproximately(domainTree);
			sampleReconciliationOfDomainToGeneApprox(domainTree);
		//	System.out.println(reconciliationCost);
		}
		reviseOutputsOfApprox();
		printOutputOfDGSReconciliation(writer);
	}

	
	
	
	private void runExtendedDGTL(PrintWriter writer) {
    	// create a mapping from each gene node to every species node
    	initializeAllGeneMappings();
    	//reconcile every gene tree with species tree
    	for(GeneTreeRooted geneTree: geneTrees) {
    		reconcileGeneTree(geneTree);
    	
    		 sampleGreedyOptionSolutionOfGeneTree(geneTree);
    		sampleUniformRandomSolution(geneTree);
   // 		System.out.println(reconciliationCost);
    	}
    	geneSpeciesReconciliationCost = reconciliationCost;
    	
    	// create a mapping from each domain node to every gene node
		initializeAllDomainMappings();
		for(DomainTreeRooted domainTree: domainTrees) {
			reconcileDomainTree(domainTree);
			sampleGreedyOptionSolutionOfDomainTree(domainTree);
		//	System.out.println(reconciliationCost);
		}
		printOutputOfDGSReconciliation(writer);
	}
   

	private void reviseOutputsOfApprox() {

		ArrayList<Node> allPostOrderDomainNodes = getAllDomainNodesByPostOrder();
		
		for (Node node : allPostOrderDomainNodes) {
			DomainNodeRooted domainNode = (DomainNodeRooted) node;
		
			if (domainNode.getEvent().contains("Transfer")) {
			 Node recipient = domainNode.getMapping().getAdditiveTransferRecipient();
			 Node donor = domainNode.getMapping().getSecondNode();
		//	 int previousMappingOption = domainNode.getMapping().getO
			 int previousMappingCost = domainNode.getMapping().getCost();
			 int changeOfCost = 0;
	
			 if(checkInterSameDomainTransfer(donor, recipient)) {
				 changeOfCost = INTER_SPECIES_DOMAIN_TRANSFER_DIFFERENT - INTER_SPECIES_DOMAIN_TRANSFER_SAME;
			
				 domainNode.setEvent("Transfer Intragene Interspecies");
				
			 }
			 else if (checkIntraSameDomainTransfer(donor, recipient)) {
				 changeOfCost = INTER_SPECIES_DOMAIN_TRANSFER_DIFFERENT - INTRA_SPECIES_DOMAIN_TRANSFER_SAME;
					
					 domainNode.setEvent("Transfer Intragene Intraspecies");
			 }
			 else if (checkIntraDifferentDomainTransfer(donor, recipient)) {
				 changeOfCost = INTER_SPECIES_DOMAIN_TRANSFER_DIFFERENT - INTRA_SPECIES_DOMAIN_TRANSFER_DIFFERENT;
					
					 domainNode.setEvent("Transfer Intergene Intraspecies");
			 }
			 domainNode.getMapping().setCost(previousMappingCost - changeOfCost);
			 reconciliationCost-=changeOfCost;
			}
			

		}
	
		

	}
	private boolean checkIntraDifferentDomainTransfer(Node firstNode, Node secondNode) {
		if(!checkSameGeneTreeCompatibility(firstNode, secondNode) && checkIntraSpeciesCompatibility(firstNode, secondNode)) {
			return true;
		}
		else return false;
	}
	private boolean checkIntraSameDomainTransfer(Node firstNode, Node secondNode) {
		if(checkSameGeneTreeCompatibility(firstNode, secondNode) && checkIntraSpeciesCompatibility(firstNode, secondNode)) {
			return true;
		}
		else return false;
	}
	private boolean checkInterSameDomainTransfer(Node firstNode, Node secondNode) {
		if(checkSameGeneTreeCompatibility(firstNode, secondNode) && !checkIntraSpeciesCompatibility(firstNode, secondNode)) {
			return true;
		}
		else return false;
	}
	private void reconcileGeneTree(GeneTreeRooted geneTree) {
    	// get all gene nodes by post order
    	ArrayList<Node> allPostOrderInternalGeneNodes = getAllPostOrderInternalGeneNodes();
    	// get all species nodes by post order
    	ArrayList<Node>allPostOrderSpeciesNodes = speciesTree.getAllNodesByPostOrder();
    	// get leaf nodes of gene trees and species tree
    	ArrayList<Node> leafGeneNodes = getAllLeafGeneNodes();
    	ArrayList<Node>leafSpeciesNodes = speciesTree.getLeafSet();
    	
    	// get all internal nodes of gene tree by pre-order
		ArrayList<Node> internalPreOrderSpeciesNodes = speciesTree.getAllPreOrderInternalNodes();
		
		//initialize all the gene to species leaf mappings
		initializeGeneToSpeciesLeafMappings(leafGeneNodes, leafSpeciesNodes);
		// for each internal gene node
		for (Node geneNode : allPostOrderInternalGeneNodes) {

			// get left child g' and right child g" of gene node
			Node leftGeneChild = geneNode.getLeft();
			Node rightGeneChild = geneNode.getRight();

			// assume the mapping cost to be infinity first
			int minimumMappingCost = Integer.MAX_VALUE;

			// Initialize an optimal mapping for this gene node
			ArrayList<RootedBinaryTreeMapping> optimalMappingsForGeneNode = new ArrayList<>();

			// for each species node
			for (Node speciesNode : allPostOrderSpeciesNodes) {

				// let the tentative mapping be (g,s)
				GeneMapping tentativeMapping = getGeneMappingFromNodes(geneNode, speciesNode);

				// let the left child to gene node mapping be (g', s)
				GeneMapping leftToParentMapping = getGeneMappingFromNodes(leftGeneChild, speciesNode);

				// let the right child to gene node mapping be (g", s)
				GeneMapping rightToParentMapping = getGeneMappingFromNodes(rightGeneChild, speciesNode);

				// if s is a leaf node
				if (speciesNode.isLeaf()) {

					handleIntToLeafGeneMapping(geneNode, speciesNode, leftGeneChild, rightGeneChild, tentativeMapping,
							leftToParentMapping, rightToParentMapping);
					setOptionsForIntToLeafGeneMapping(geneNode, speciesNode, leftGeneChild, rightGeneChild,
							tentativeMapping);
				} else {
					Node leftSpeciesChild = speciesNode.getLeft();
					Node rightSpeciesChild = speciesNode.getRight();
					handleIntToIntGeneMapping(tentativeMapping, geneNode, speciesNode, leftToParentMapping,
							rightToParentMapping);
					setOptionsForIntToIntGeneMapping(geneNode, speciesNode, leftGeneChild, rightGeneChild,
							leftSpeciesChild, rightSpeciesChild, tentativeMapping);
				}

				tentativeMapping.sumUpNumberOfOptions();

				// if this mapping has smaller cost then let it be present optimal mapping
				if (tentativeMapping.getCost() < minimumMappingCost) {
					minimumMappingCost = tentativeMapping.getCost();
					optimalMappingsForGeneNode.clear();
					optimalMappingsForGeneNode.add(tentativeMapping);
				} else if (tentativeMapping.getCost() == minimumMappingCost) {
					optimalMappingsForGeneNode.add(tentativeMapping);
				}

			}
			processGeneToSpeciesOutCalculation(internalPreOrderSpeciesNodes, geneNode);

			// add the optimal mapping to optimal mapping list

			optimalMappings.put(geneNode, optimalMappingsForGeneNode);

		}
		// the final reconciliation cost should be the cost of mapping for root of gene
		// tree
		RootedBinaryTreeMapping rootMap = optimalMappings.get(geneTree.getRoot()).get(0);
		reconciliationCost += rootMap.getCost();
    }
	private void reconcileDomainTreeApproximately(DomainTreeRooted domainTree) {

    	

		// get all nodes of gene tree  by post order		
		ArrayList<Node> allPostOrderGeneNodes = getAllGeneNodesByPostOrder();
		

		// get leaf nodes of domain tree and species tree
		ArrayList<Node> leafDomainNodes = domainTree.getLeafSet();
		ArrayList<Node> leafGeneNodes = getAllLeafGeneNodes();

		// get all internal nodes of domain tree by post-order
		ArrayList<Node> internalDomainNodes = domainTree.getAllPostOrderInternalNodes();

		// get all internal nodes of gene tree by pre-order
		ArrayList<Node> internalPreOrderGeneNodes = getAllPreOrderInternalGeneNodes();

		

		initDToGLeafMappingsApproximately(leafDomainNodes, leafGeneNodes);

		// for each internal domain node
		for (Node domainNode : internalDomainNodes) {

			// get left child g' and right child g" of gene node
			Node leftDomainChild = domainNode.getLeft();
			Node rightDomainChild = domainNode.getRight();

			// assume the mapping cost to be infinity first
			int minimumMappingCost = Integer.MAX_VALUE;

			// Initialize an optimal mapping for this gene node
			ArrayList<RootedBinaryTreeMapping> optimalMappingsForGeneNode = new ArrayList<>();

			// for each gene node
			for (Node geneNode : allPostOrderGeneNodes) {

				// let the tentative mapping be (g,s)
				DomainMapping tentativeMapping = getDomainMappingFromNodes(domainNode, geneNode);

				// let the left child to gene node mapping be (g', s)
				DomainMapping leftToParentMapping = getDomainMappingFromNodes(leftDomainChild, geneNode);

				// let the right child to gene node mapping be (g", s)
				DomainMapping rightToParentMapping = getDomainMappingFromNodes(rightDomainChild, geneNode);

				// if s is a leaf node
				if (geneNode.isLeaf()) {

					handleIntToLeafDomainMappingApprox(domainNode, geneNode, leftDomainChild, rightDomainChild, tentativeMapping,
							leftToParentMapping, rightToParentMapping);
					setOptionsForIntToLeafDomainMappingApprox(domainNode, geneNode, leftDomainChild, rightDomainChild,
							tentativeMapping);
				} else {
					Node leftSpeciesChild = geneNode.getLeft();
					Node rightSpeciesChild = geneNode.getRight();
					handleIntToIntDomainMappingApprox(tentativeMapping, domainNode, geneNode, leftToParentMapping,
							rightToParentMapping);
					setOptionsForIntToIntDomainMappingApprox(domainNode, geneNode, leftDomainChild, rightDomainChild,
							leftSpeciesChild, rightSpeciesChild, tentativeMapping);
				}

			//	tentativeMapping.sumUpNumberOfOptions();

				// if this mapping has smaller cost then let it be present optimal mapping
				if (tentativeMapping.getCost() < minimumMappingCost) {
					minimumMappingCost = tentativeMapping.getCost();
					optimalMappingsForGeneNode.clear();
					optimalMappingsForGeneNode.add(tentativeMapping);
				} else if (tentativeMapping.getCost() == minimumMappingCost) {
					optimalMappingsForGeneNode.add(tentativeMapping);
				}

			}
			processDomainToGeneOutCalculationApprox(internalPreOrderGeneNodes, domainNode);

			// add the optimal mapping to optimal mapping list

			optimalMappings.put(domainNode, optimalMappingsForGeneNode);

		}
		// the final reconciliation cost should be the cost of mapping for root of gene
		// tree
		RootedBinaryTreeMapping rootMap = optimalMappings.get(domainTree.getRoot()).get(0);
		reconciliationCost += rootMap.getCost();
	
	}
	
	private void reconcileDomainTree(DomainTreeRooted domainTree) {
    	

    		// get all nodes of gene tree  by post order
    		
    		ArrayList<Node> allPostOrderGeneNodes = getAllGeneNodesByPostOrder();

    		

    		// get leaf nodes of domain tree and species tree
    		ArrayList<Node> leafDomainNodes = domainTree.getLeafSet();
    		ArrayList<Node> leafGeneNodes = getAllLeafGeneNodes();

    		// get all internal nodes of domain tree by post-order
    		ArrayList<Node> internalDomainNodes = domainTree.getAllPostOrderInternalNodes();

    		// get all internal nodes of gene tree by pre-order
    		ArrayList<Node> internalPreOrderGeneNodes = getAllPreOrderInternalGeneNodes();

    		

    		initializeDomainToGeneLeafMappings(leafDomainNodes, leafGeneNodes);

    		// for each internal domain node
    		for (Node domainNode : internalDomainNodes) {

    			// get left child g' and right child g" of gene node
    			Node leftDomainChild = domainNode.getLeft();
    			Node rightDomainChild = domainNode.getRight();

    			// assume the mapping cost to be infinity first
    			int minimumMappingCost = Integer.MAX_VALUE;

    			// Initialize an optimal mapping for this gene node
    			ArrayList<RootedBinaryTreeMapping> optimalMappingsForGeneNode = new ArrayList<>();

    			// for each gene node
    			for (Node geneNode : allPostOrderGeneNodes) {

    				// let the tentative mapping be (g,s)
    				DomainMapping tentativeMapping = getDomainMappingFromNodes(domainNode, geneNode);

    				// let the left child to gene node mapping be (g', s)
    				DomainMapping leftToParentMapping = getDomainMappingFromNodes(leftDomainChild, geneNode);

    				// let the right child to gene node mapping be (g", s)
    				DomainMapping rightToParentMapping = getDomainMappingFromNodes(rightDomainChild, geneNode);

    				// if s is a leaf node
    				if (geneNode.isLeaf()) {

    					handleIntToLeafDomainMapping(domainNode, geneNode, leftDomainChild, rightDomainChild, tentativeMapping,
    							leftToParentMapping, rightToParentMapping);
    					setOptionsForIntToLeafDomainMapping(domainNode, geneNode, leftDomainChild, rightDomainChild,
    							tentativeMapping);
    				} else {
    					Node leftGeneChild = geneNode.getLeft();
    					Node rightGeneChild = geneNode.getRight();
    					handleIntToIntDomainMapping(tentativeMapping, domainNode, geneNode, leftToParentMapping,
    							rightToParentMapping);
    					setOptionsForIntToIntDomainMapping(domainNode, geneNode, leftDomainChild, rightDomainChild,
    							leftGeneChild, rightGeneChild, tentativeMapping);
    				}

    			//	tentativeMapping.sumUpNumberOfOptions();

    				// if this mapping has smaller cost then let it be present optimal mapping
    				if (tentativeMapping.getCost() < minimumMappingCost) {
    					minimumMappingCost = tentativeMapping.getCost();
    					optimalMappingsForGeneNode.clear();
    					optimalMappingsForGeneNode.add(tentativeMapping);
    				} else if (tentativeMapping.getCost() == minimumMappingCost) {
    					optimalMappingsForGeneNode.add(tentativeMapping);
    				}

    			}
    			processDomainToGeneDGTLOutCalculation(internalPreOrderGeneNodes, domainNode);

    			// add the optimal mapping to optimal mapping list

    			optimalMappings.put(domainNode, optimalMappingsForGeneNode);

    		}
    		// the final reconciliation cost should be the cost of mapping for root of gene
    		// tree
    		RootedBinaryTreeMapping rootMap = optimalMappings.get(domainTree.getRoot()).get(0);
    		reconciliationCost += rootMap.getCost();
    	}
	
    private void setOptionsForIntToIntGeneMapping(Node geneNode, Node speciesNode, Node leftGeneChild,
			Node rightGeneChild, Node leftSpeciesChild, Node rightSpeciesChild,
			RootedBinaryTreeMapping tentativeMapping) {

		// let the left child to species node mapping be (g', s)
		RootedBinaryTreeMapping leftToParentMapping = getGeneMappingFromNodes(leftGeneChild, speciesNode);

		// let the right child to species node mapping be (g", s)
		RootedBinaryTreeMapping rightToParentMapping = getGeneMappingFromNodes(rightGeneChild, speciesNode);

		// let the left child of gene node to left child of species node mapping be (g',
		// s')
		RootedBinaryTreeMapping leftToLeftMapping = getGeneMappingFromNodes(leftGeneChild, leftSpeciesChild);

		// let the left child of gene node to right child of species node mapping be
		// (g', s")
		RootedBinaryTreeMapping leftToRightMapping = getGeneMappingFromNodes(leftGeneChild, rightSpeciesChild);

		// let the right child of gene node to left child of species node mapping be
		// (g", s')
		RootedBinaryTreeMapping rightToLeftMapping = getGeneMappingFromNodes(rightGeneChild, leftSpeciesChild);

		// let the right child of gene node to right child of species node mapping be
		// (g", s")
		RootedBinaryTreeMapping rightToRightMapping = getGeneMappingFromNodes(rightGeneChild, rightSpeciesChild);

		// speciationCost = min{in(g',s') + in (g",s"), in(g",s') + in (g', s")}
		int speciationCost = Math.min(leftToLeftMapping.getInCost() + rightToRightMapping.getInCost(),
				leftToRightMapping.getInCost() + rightToLeftMapping.getInCost());
		if (tentativeMapping.getCost() >= Integer.MAX_VALUE / 10)
			return;
		if (tentativeMapping.getCost() == speciationCost) {
			if (tentativeMapping.getCost() == leftToLeftMapping.getInCost() + rightToRightMapping.getInCost()) {
				tentativeMapping.getOptions().add(4);
				tentativeMapping.addWeightByInAndIn(leftToLeftMapping, rightToRightMapping, this);
			}
			if (tentativeMapping.getCost() == leftToRightMapping.getInCost() + rightToLeftMapping.getInCost()) {
				tentativeMapping.getOptions().add(5);
				tentativeMapping.addWeightByInAndIn(leftToRightMapping, rightToLeftMapping, this);
			}
		}

		int[] duplicationOptions = getDuplicationOptions(leftToParentMapping, rightToParentMapping, leftToLeftMapping,
				leftToRightMapping, rightToLeftMapping, rightToRightMapping);

		// Set duplicationCost = DUPLICATION + min(duplicationOptions)
		int duplicationCost = GENE_DUPLICATION + getMinimum(duplicationOptions);
		if (tentativeMapping.getCost() == duplicationCost) {
			if (tentativeMapping.getCost() - GENE_DUPLICATION == leftToParentMapping.getCost()
					+ rightToRightMapping.getInCost() + GENE_LOSS) {
				tentativeMapping.getOptions().add(6);
				tentativeMapping.addWeightByNormalAndIn(leftToParentMapping, rightToRightMapping, this);
			}
			if (tentativeMapping.getCost() - GENE_DUPLICATION == leftToParentMapping.getCost()
					+ rightToLeftMapping.getInCost() + GENE_LOSS) {
				tentativeMapping.getOptions().add(7);
				tentativeMapping.addWeightByNormalAndIn(leftToParentMapping, rightToLeftMapping, this);

			}
			if (tentativeMapping.getCost() - GENE_DUPLICATION == rightToParentMapping.getCost()
					+ leftToRightMapping.getInCost() + GENE_LOSS) {
				tentativeMapping.getOptions().add(8);
				tentativeMapping.addWeightByNormalAndIn(rightToParentMapping, leftToRightMapping, this);
			}
			if (tentativeMapping.getCost() - GENE_DUPLICATION == rightToParentMapping.getCost()
					+ leftToLeftMapping.getInCost() + GENE_LOSS) {
				tentativeMapping.getOptions().add(9);
				tentativeMapping.addWeightByNormalAndIn(rightToParentMapping, leftToLeftMapping, this);

			}
			if (tentativeMapping.getCost() - GENE_DUPLICATION == rightToParentMapping.getCost()
					+ leftToParentMapping.getCost()) {
				tentativeMapping.getOptions().add(10);
				tentativeMapping.addWeightByNormalAndNormal(rightToParentMapping, leftToParentMapping);

			}
			if (tentativeMapping.getCost() - GENE_DUPLICATION == leftToLeftMapping.getInCost()
					+ rightToRightMapping.getInCost() + 2 * GENE_LOSS) {
				tentativeMapping.getOptions().add(11);
				tentativeMapping.addWeightByInAndIn(leftToLeftMapping, rightToRightMapping, this);
			}
			if (tentativeMapping.getCost() - GENE_DUPLICATION == leftToRightMapping.getInCost()
					+ rightToLeftMapping.getInCost() + 2 * GENE_LOSS) {
				tentativeMapping.getOptions().add(12);
				tentativeMapping.addWeightByInAndIn(leftToRightMapping, rightToLeftMapping, this);

			}
			if (tentativeMapping.getCost() - GENE_DUPLICATION == leftToLeftMapping.getInCost()
					+ rightToLeftMapping.getInCost() + 2 * GENE_LOSS) {
				tentativeMapping.getOptions().add(13);
				tentativeMapping.addWeightByInAndIn(leftToLeftMapping, rightToLeftMapping, this);

			}
			if (tentativeMapping.getCost() - GENE_DUPLICATION == leftToRightMapping.getInCost()
					+ rightToRightMapping.getInCost() + 2 * GENE_LOSS) {
				tentativeMapping.getOptions().add(14);
				tentativeMapping.addWeightByInAndIn(leftToRightMapping, rightToRightMapping, this);

			}
		}
		if (!speciesNode.isRoot()) {

			// transferCost(g,s) = TRANSFER + min(in(g',s) + out(g",s), in(g",s) + out(g's))
			int transferCost = GENE_TRANSFER + Math.min(leftToParentMapping.getInCost() + rightToParentMapping.getOutCost(),
					leftToParentMapping.getOutCost() + rightToParentMapping.getInCost());
			if (tentativeMapping.getCost() == transferCost) {
				if (tentativeMapping.getCost() == GENE_TRANSFER + leftToParentMapping.getInCost()
						+ rightToParentMapping.getOutCost()) {
					tentativeMapping.getOptions().add(15);
					tentativeMapping.addWeightByInOut(leftToParentMapping, rightToParentMapping, this);
				}
				if (tentativeMapping.getCost() == GENE_TRANSFER + leftToParentMapping.getOutCost()
						+ rightToParentMapping.getInCost()) {
					tentativeMapping.getOptions().add(16);
					tentativeMapping.addWeightByInOut(rightToParentMapping, leftToParentMapping, this);
				}
			}

		}

	}
    private void processGeneToSpeciesOutCalculation(ArrayList<Node> internalPreOrderSpeciesNodes, Node geneNode) {

		// for each species node in pre order
		for (Node speciesNode : internalPreOrderSpeciesNodes) {

			// let s' = left(s) and s" = right(s)
			Node leftSpeciesChild = speciesNode.getLeft();
			Node rightSpeciesChild = speciesNode.getRight();

			// get the mappings (g, s), (g, s'), (g, s")
			RootedBinaryTreeMapping tentativeMapping = getGeneMappingFromNodes(geneNode, speciesNode);
			RootedBinaryTreeMapping parentToLeftMapping = getGeneMappingFromNodes(geneNode, leftSpeciesChild);
			RootedBinaryTreeMapping parentToRightMapping = getGeneMappingFromNodes(geneNode, rightSpeciesChild);

			// int prevParentToLeftOutCost = parentToLeftMapping.getOutCost();
			// out(g, s') = { out(g, s), inAlt(g, s")}
			parentToLeftMapping
					.setOutCost(Math.min(tentativeMapping.getOutCost(), parentToRightMapping.getInAltCost()));
			// if (parentToLeftMapping.getOutCost() < prevParentToLeftOutCost) {
			parentToLeftMapping.getOutSpeciesNodes().clear();
			// }

			// int prevParentToRightOutCost = parentToRightMapping.getOutCost();
			// out(g, s") = { out(g, s), inAlt(g, s')}
			parentToRightMapping
					.setOutCost(Math.min(tentativeMapping.getOutCost(), parentToLeftMapping.getInAltCost()));
			// if (parentToRightMapping.getOutCost() < prevParentToRightOutCost) {
			parentToRightMapping.getOutSpeciesNodes().clear();
			// }
			updateOutSpeciesNodeForMultipleOptima(tentativeMapping, parentToLeftMapping, parentToRightMapping);

		}

	}
	private RootedBinaryTreeMapping getUniformRandomInMappingFromPair(Node geneNode, Node speciesNode) {
		RootedBinaryTreeMapping mapping = getGeneMappingFromNodes(geneNode, speciesNode);

		return mapping.getInMappingUniformRandomly(this);

	}
	private RootedBinaryTreeMapping getUniformRandomSolution(
			ArrayList<RootedBinaryTreeMapping> optimalMappingsOfGeneNode) {

		RootedBinaryTreeMapping uniformMapping = null;

		int minHeight = Integer.MAX_VALUE;
		for (RootedBinaryTreeMapping mapping : optimalMappingsOfGeneNode) {
			int height = mapping.getSecondNode().getHeight();
			//System.out.println(mapping.getSpeciesNode() + " " + height);
			if (height < minHeight) {
				uniformMapping = mapping;
				minHeight = height;
			}
		}
		return uniformMapping;

	}
    private void sampleUniformRandomSolution(GeneTreeRooted geneTree) {

		ArrayList<Node> allPreOrderGeneNodes = geneTree.getAllNodesByPreOrder();
		Random random = new Random();
		for (Node geneNode : allPreOrderGeneNodes) {
			RootedBinaryTreeMapping optimalMapping = null;
			ArrayList<RootedBinaryTreeMapping> optimalMappingsOfGeneNode = optimalMappings.get(geneNode);
			GeneNodeRooted geneNodeRooted = (GeneNodeRooted) geneNode;
			if (geneNode.isRoot()) {

				optimalMapping = getUniformRandomSolution(optimalMappingsOfGeneNode);// getGreedyOptionSolution(optimalMappingsOfGeneNode);

			} else {
				GeneNodeRooted parentNode = (GeneNodeRooted) geneNodeRooted.getParent();

				RootedBinaryTreeMapping parentMapping = parentNode.getMapping();
				Node parentImage = parentMapping.getSecondNode();
				Node rightOfParentImage = parentMapping.getSecondNode().getRight();
				Node leftOfParentImage = parentMapping.getSecondNode().getLeft();

				int parentOption = parentNode.getOption();
				if (parentOption == 1) {
					optimalMapping = getGeneMappingFromNodes(geneNode, parentImage);
				} else if (parentOption == 2 || parentOption == 15) {
					if (parentNode.getLeft().equals(geneNodeRooted)) {
						optimalMapping = getUniformRandomInMappingFromPair(geneNode, parentImage);
					//	parentMapping.updateInTransferLossesWithBranch(optimalMapping.getSecondNode());
					} else {
						optimalMapping = getUniformOutMappingFromPair(geneNode, parentImage);
						parentNode.getMapping().setAdditiveTransferRecipient(optimalMapping.getSecondNode());
					}
				} else if (parentOption == 3 || parentOption == 16) {
					if (parentNode.getRight().equals(geneNodeRooted)) {
						optimalMapping = getUniformRandomInMappingFromPair(geneNode, parentImage);
				//		parentMapping.updateInTransferLossesWithBranch(optimalMapping.getSecondNode());

					} else {
						optimalMapping = getUniformOutMappingFromPair(geneNode, parentImage);
						parentNode.getMapping().setAdditiveTransferRecipient(optimalMapping.getSecondNode());

					}
				} else if (parentOption == 4) {
					if (parentNode.getLeft().equals(geneNodeRooted)) {
						optimalMapping = getUniformRandomInMappingFromPair(geneNode, leftOfParentImage);

					} else {

						optimalMapping = getUniformRandomInMappingFromPair(geneNode, rightOfParentImage);

					}
				
				} else if (parentOption == 5) {
					if (parentNode.getLeft().equals(geneNodeRooted)) {
						optimalMapping = getUniformRandomInMappingFromPair(geneNode, rightOfParentImage);
					} else {
						optimalMapping = getUniformRandomInMappingFromPair(geneNode, leftOfParentImage);
					}
				
				} else if (parentOption == 6) {
					if (parentNode.getLeft().equals(geneNodeRooted)) {
						optimalMapping = getGeneMappingFromNodes(geneNode, parentImage);
					} else {
						optimalMapping = getUniformRandomInMappingFromPair(geneNode, rightOfParentImage);// inSearchWithLossTracking(optimalMappingsOfGeneNode,
						// rightOfParentImage);
					}
				} else if (parentOption == 7) {
					if (parentNode.getLeft().equals(geneNodeRooted)) {
						optimalMapping = getGeneMappingFromNodes(geneNode, parentImage);
					} else {
						optimalMapping = getUniformRandomInMappingFromPair(geneNode, leftOfParentImage);
					}
				} else if (parentOption == 8) {
					if (parentNode.getRight().equals(geneNodeRooted)) {
						optimalMapping = getGeneMappingFromNodes(geneNode, parentImage);
				
					} else {
						optimalMapping = getUniformRandomInMappingFromPair(geneNode, rightOfParentImage);
				
					}
				} else if (parentOption == 9) {
					if (parentNode.getRight().equals(geneNodeRooted)) {
						optimalMapping = getGeneMappingFromNodes(geneNode, parentImage);
				
					} else {
						optimalMapping = getUniformRandomInMappingFromPair(geneNode, leftOfParentImage);
				
					}
				} else if (parentOption == 10) {
					optimalMapping = getGeneMappingFromNodes(geneNode, parentImage);
				} else if (parentOption == 11) {
					if (parentNode.getLeft().equals(geneNodeRooted)) {
						optimalMapping = getUniformRandomInMappingFromPair(geneNode, leftOfParentImage);
					} else {
						optimalMapping = getUniformRandomInMappingFromPair(geneNode, rightOfParentImage);
					}
			//		parentMapping.updateLossesWithBranch(optimalMapping.getSecondNode());

				} else if (parentOption == 12) {
					if (parentNode.getLeft().equals(geneNodeRooted)) {
						optimalMapping = getUniformRandomInMappingFromPair(geneNode, rightOfParentImage);
					
					} else {
						optimalMapping = getUniformRandomInMappingFromPair(geneNode, leftOfParentImage);
					
					}
				
				} else if (parentOption == 13) {
					optimalMapping = getUniformRandomInMappingFromPair(geneNode, leftOfParentImage);
				} else if (parentOption == 14) {
					optimalMapping = getUniformRandomInMappingFromPair(geneNode, rightOfParentImage);
				
				}
			}

			geneNodeRooted.setMappedSpeciesNode((SpeciesNodeRooted) optimalMapping.getSecondNode());
			geneNodeRooted
					.setOption(optimalMapping.getOptions().get(random.nextInt(optimalMapping.getOptions().size())));
			geneNodeRooted.setMapping(optimalMapping);
			geneNodeRooted.setEventByOption();

		}

	}
	 private void sampleGreedyOptionSolutionOfGeneTree(GeneTreeRooted geneTree) {
	    	

		ArrayList<Node> allPreOrderGeneNodes = geneTree.getAllNodesByPreOrder();

		for (Node geneNode : allPreOrderGeneNodes) {
			RootedBinaryTreeMapping optimalMapping = null;
			ArrayList<RootedBinaryTreeMapping> optimalMappingsOfGeneNode = optimalMappings.get(geneNode);
			GeneNodeRooted geneNodeRooted = (GeneNodeRooted) geneNode;
			if (geneNode.isRoot()) {

				optimalMapping = getRandomSolution(optimalMappingsOfGeneNode);//getGreedyOptionSolution(optimalMappingsOfGeneNode);

			} else {
				GeneNodeRooted parentNode = (GeneNodeRooted) geneNodeRooted.getParent();

				RootedBinaryTreeMapping parentMapping = parentNode.getMapping();
				Node parentImage = parentMapping.getSecondNode();
				Node rightOfParentImage = parentMapping.getSecondNode().getRight();
				Node leftOfParentImage = parentMapping.getSecondNode().getLeft();

				int parentOption = parentNode.getOption();
				if (parentOption == 1) {
					optimalMapping = getGeneMappingFromNodes(geneNode, parentImage);
				} else if (parentOption == 2 || parentOption == 15) {
					if (parentNode.getLeft().equals(geneNodeRooted)) {
						optimalMapping = getRandomInMappingFromPair(geneNode, parentImage);
						parentMapping.updateInTransferLossesWithBranch(optimalMapping.getSecondNode());
					} else {
						optimalMapping = getRandomOutMappingFromPair(geneNode, parentImage);
						parentNode.getMapping().setAdditiveTransferRecipient(optimalMapping.getSecondNode());
					}

				} else if (parentOption == 3 || parentOption == 16) {
					if (parentNode.getRight().equals(geneNodeRooted)) {
						optimalMapping = getRandomInMappingFromPair(geneNode, parentImage);
						parentMapping.updateInTransferLossesWithBranch(optimalMapping.getSecondNode());

					} else {
						optimalMapping = getRandomOutMappingFromPair(geneNode, parentImage);
						parentNode.getMapping().setAdditiveTransferRecipient(optimalMapping.getSecondNode());

					}

				} else if (parentOption == 4) {
					if (parentNode.getLeft().equals(geneNodeRooted)) {
						optimalMapping = getRandomInMappingFromPair(geneNode, leftOfParentImage);

					} else {

						optimalMapping = getRandomInMappingFromPair(geneNode, rightOfParentImage);

					}
				//	parentMapping.updateLossesWithBranch(optimalMapping.getSecondNode());

				} else if (parentOption == 5) {
					if (parentNode.getLeft().equals(geneNodeRooted)) {
						optimalMapping = getRandomInMappingFromPair(geneNode, rightOfParentImage);
					} else {
						optimalMapping = getRandomInMappingFromPair(geneNode, leftOfParentImage);
					}
				//	parentMapping.updateLossesWithBranch(optimalMapping.getSecondNode());

				} else if (parentOption == 6) {
					if (parentNode.getLeft().equals(geneNodeRooted)) {
						optimalMapping = getGeneMappingFromNodes(geneNode, parentImage);
						parentMapping.addLoss(parentImage.getLeft());
						parentMapping.addLostBranch(parentImage.getLeftBranch());
					} else {
						optimalMapping = getRandomInMappingFromPair(geneNode, rightOfParentImage);// inSearchWithLossTracking(optimalMappingsOfGeneNode,
						// rightOfParentImage);
						parentMapping.updateLossesWithBranch(optimalMapping.getSecondNode());
					}
				} else if (parentOption == 7) {
					if (parentNode.getLeft().equals(geneNodeRooted)) {
						optimalMapping = getGeneMappingFromNodes(geneNode, parentImage);
						parentMapping.addLoss(parentImage.getRight());
						parentMapping.addLostBranch(parentImage.getRightBranch());
					} else {
						optimalMapping = getRandomInMappingFromPair(geneNode, leftOfParentImage);
					//	parentMapping.updateLossesWithBranch(optimalMapping.getSecondNode());
					}
				} else if (parentOption == 8) {
					if (parentNode.getRight().equals(geneNodeRooted)) {
						optimalMapping = getGeneMappingFromNodes(geneNode, parentImage);
					//	parentMapping.addLoss(parentImage.getLeft());
					//	parentMapping.addLostBranch(parentImage.getLeftBranch());

					} else {
						optimalMapping = getRandomInMappingFromPair(geneNode, rightOfParentImage);
						parentMapping.updateLossesWithBranch(optimalMapping.getSecondNode());

					}
				} else if (parentOption == 9) {
					if (parentNode.getRight().equals(geneNodeRooted)) {
						optimalMapping = getGeneMappingFromNodes(geneNode, parentImage);
						parentMapping.addLoss(parentImage.getRight());
						parentMapping.addLostBranch(parentImage.getRightBranch());

					} else {
						optimalMapping = getRandomInMappingFromPair(geneNode, leftOfParentImage);
						parentMapping.updateLossesWithBranch(optimalMapping.getSecondNode());

					}
				} else if (parentOption == 10) {
					optimalMapping = getGeneMappingFromNodes(geneNode, parentImage);
				} else if (parentOption == 11) {
					if (parentNode.getLeft().equals(geneNodeRooted)) {
						optimalMapping = getRandomInMappingFromPair(geneNode, leftOfParentImage);
						parentMapping.addLoss(parentImage.getRight());
						parentMapping.addLostBranch(parentImage.getRightBranch());
					} else {
						optimalMapping = getRandomInMappingFromPair(geneNode, rightOfParentImage);
						parentMapping.addLoss(parentImage.getLeft());
						parentMapping.addLostBranch(parentImage.getLeftBranch());
					}
					parentMapping.updateLossesWithBranch(optimalMapping.getSecondNode());

				} else if (parentOption == 12) {
					if (parentNode.getLeft().equals(geneNodeRooted)) {
						optimalMapping = getRandomInMappingFromPair(geneNode, rightOfParentImage);
						parentMapping.addLoss(parentImage.getRight());
						parentMapping.addLostBranch(parentImage.getRightBranch());

					} else {
						optimalMapping = getRandomInMappingFromPair(geneNode, leftOfParentImage);
						parentMapping.addLoss(parentImage.getLeft());
						parentMapping.addLostBranch(parentImage.getLeftBranch());

					}
					parentMapping.updateLossesWithBranch(optimalMapping.getSecondNode());

				} else if (parentOption == 13) {
					optimalMapping = getRandomInMappingFromPair(geneNode, leftOfParentImage);
					parentMapping.addLoss(parentImage.getRight());
					parentMapping.addLostBranch(parentImage.getRightBranch());
					parentMapping.updateLossesWithBranch(optimalMapping.getSecondNode());
				} else if (parentOption == 14) {
					optimalMapping = getRandomInMappingFromPair(geneNode, rightOfParentImage);
					parentMapping.addLoss(parentImage.getLeft());
					parentMapping.addLostBranch(parentImage.getLeftBranch());
					parentMapping.updateLossesWithBranch(optimalMapping.getSecondNode());
				}
			}
			
			geneNodeRooted.setMappedSpeciesNode((SpeciesNodeRooted)optimalMapping.getSecondNode());
			setGeneOptionRandomly(geneNodeRooted, optimalMapping);
			geneNodeRooted.setMapping(optimalMapping);
			geneNodeRooted.setEventByOption();

		}

	}
	

	private void processDomainToGeneOutCalculationApprox(ArrayList<Node> internalPreOrderGeneNodes, Node domainNode) {
			//exchange root out-mapping information
			int minimumInAltCost = Integer.MAX_VALUE;
			ArrayList<Node>outNodes = new ArrayList<Node>();
			//find minimum inAltCost first
			for (GeneTreeRooted geneTree : geneTrees) {
				Node rootGene = geneTree.getRoot();
				DomainMapping tentativeMapping = getDomainMappingFromNodes(domainNode, rootGene);
				if(tentativeMapping.getInAltCost()<minimumInAltCost) {
						minimumInAltCost = tentativeMapping.getInAltCost();
						outNodes = tentativeMapping.getInAltNodes();
				}
			}
				//exchange root out-mapping cost information and nodes
				for (GeneTreeRooted geneTree : geneTrees) {
					Node rootGene = geneTree.getRoot();
					DomainMapping tentativeMapping = getDomainMappingFromNodes(domainNode, rootGene);
					if(tentativeMapping.getInAltCost()>minimumInAltCost) {
						tentativeMapping.setOutCost(minimumInAltCost);
						tentativeMapping.updateOutNodes(this, outNodes);
						
					}
			}
				// for each species node in pre order
				for (Node internalGeneNode : internalPreOrderGeneNodes) {

					// let s' = left(s) and s" = right(s)
					Node leftGeneChild = internalGeneNode.getLeft();
					Node rightGeneChild = internalGeneNode.getRight();

					// get the mappings (g, s), (g, s'), (g, s")
					DomainMapping tentativeMapping = getDomainMappingFromNodes(domainNode, internalGeneNode);
					DomainMapping parentToLeftMapping = getDomainMappingFromNodes(domainNode, leftGeneChild);
					DomainMapping parentToRightMapping = getDomainMappingFromNodes(domainNode, rightGeneChild);

					setOutTransferCostsAndMappingApprox(parentToLeftMapping, tentativeMapping, parentToRightMapping);
			
					
					
					setOutTransferCostsAndMappingApprox(parentToRightMapping, tentativeMapping, parentToLeftMapping);
					
					}

		}
	
	 
	private void processDomainToGeneDGTLOutCalculation(ArrayList<Node> internalPreOrderGeneNodes, Node domainNode) {
			//exchange root out-mapping information
		int minimumInAltCost = Integer.MAX_VALUE/10;
		ArrayList<Node>outNodes = new ArrayList<Node>();
		//find minimum inAltCost first
		for (GeneTreeRooted geneTree : geneTrees) {
			Node rootGene = geneTree.getRoot();
			DomainMapping tentativeMapping = getDomainMappingFromNodes(domainNode, rootGene);
			if(tentativeMapping.getInAltCost()<minimumInAltCost) {
					minimumInAltCost = tentativeMapping.getInAltCost();
					outNodes = tentativeMapping.getInAltNodes();
			}
		}
			//exchange root out-mapping cost information and nodes
			for (GeneTreeRooted geneTree : geneTrees) {
				Node rootGene = geneTree.getRoot();
				DomainMapping tentativeMapping = getDomainMappingFromNodes(domainNode, rootGene);
				if(tentativeMapping.getInAltCost()>minimumInAltCost) {
					if(checkIntraSpeciesTransfer(tentativeMapping, outNodes)) {
						tentativeMapping.setIntraDifferentTreeOutCost(minimumInAltCost);
						tentativeMapping.updateIntraSpeciesDifferentOutNodes(this, outNodes);
					}
					if(checkInterSpeciesTransfer(tentativeMapping, outNodes)) {
						tentativeMapping.setInterDifferentTreeOutCost(minimumInAltCost);
						tentativeMapping.updateInterSpeciesDifferentOutNodes(this, outNodes);
						
					}
				}
			}
				// for each gene node in pre order
				for (Node internalGeneNode : internalPreOrderGeneNodes) {

					// let g' = left(g) and g" = right(g)
					Node leftGeneChild = internalGeneNode.getLeft();
					Node rightGeneChild = internalGeneNode.getRight();

					// get the mappings (d, g), (d, g'), (d, g")
					DomainMapping tentativeMapping = getDomainMappingFromNodes(domainNode, internalGeneNode);
					DomainMapping parentToLeftMapping = getDomainMappingFromNodes(domainNode, leftGeneChild);
					DomainMapping parentToRightMapping = getDomainMappingFromNodes(domainNode, rightGeneChild);
				//	if(tentativeMapping.toString().equals("[D41_29, H13_29]")|| tentativeMapping.toString().equals("[D36_72, H7]") ) {
				//		System.out.println("Debug point");
				//	}
					setOutTransferCostsAndMapping(parentToLeftMapping, tentativeMapping, parentToRightMapping);
			
					
					
					setOutTransferCostsAndMapping(parentToRightMapping, tentativeMapping, parentToLeftMapping);
					
				

	}
	}
	
	private boolean checkIntraSpeciesTransfer(DomainMapping tentativeMapping, ArrayList<Node> outNodes) {
		//take current gene node
		Node geneNode = tentativeMapping.getSecondNode();
		// iterate and check all other outnodes for intra species transfer
		for(Node outGeneNode: outNodes) {
			if(checkIntraSpeciesCompatibility(geneNode, outGeneNode)) {
				return true;
			}
		}
		return false;
	}
	private boolean checkInterSpeciesTransfer(DomainMapping tentativeMapping, ArrayList<Node> outNodes) {
		//take current gene node
		Node geneNode = tentativeMapping.getSecondNode();
		// iterate and check all other outnodes for inter species transfer
		for(Node outGeneNode: outNodes) {
			if(!checkIntraSpeciesCompatibility(geneNode, outGeneNode)) {
				return true;
			}
		}
		return false;
	}
	private void setOutTransferCostsAndMappingApprox(DomainMapping parentToLeftMapping, DomainMapping tentativeMapping,
		DomainMapping parentToRightMapping) {
		int newOutCost = Math.min(tentativeMapping.getOutCost(), parentToRightMapping.getInAltCost());
		parentToLeftMapping.setOutCost(newOutCost);
		parentToLeftMapping.getOutNodes().clear();
		if (parentToLeftMapping.getOutCost()==tentativeMapping.getOutCost()) {
					
			for(Node outNode: tentativeMapping.getOutNodes()) {
						parentToLeftMapping.addOutSpeciesNode(outNode);
			}
		}
		else{
		
				for(Node inAltNode: parentToRightMapping.getInAltNodes()) {
						parentToLeftMapping.addOutSpeciesNode(inAltNode);
				}
		}
		
		
	}
	private void setOutTransferCostsAndMapping(DomainMapping parentToLeftMapping, DomainMapping tentativeMapping,
			DomainMapping parentToRightMapping) {
		
		handleIntraSameOutCostAndMapping(parentToLeftMapping,tentativeMapping,parentToRightMapping );
		handleIntraDifferentOutCostAndMapping(parentToLeftMapping, tentativeMapping);
		
		handleInterSameOutCostAndMapping(parentToLeftMapping,tentativeMapping,parentToRightMapping );
		handleInterDifferentOutCostAndMapping(parentToLeftMapping, tentativeMapping);
		
		
	}
	
	private void handleInterSameOutCostAndMapping(DomainMapping parentToLeftMapping,
			DomainMapping tentativeMapping, DomainMapping parentToRightMapping) {
	//	if(parentToLeftMapping.toString().equals("[D41_29, H13_29]")|| tentativeMapping.toString().equals("[D36_72, H7]") ) {
	//		System.out.println("Debug point");
	//	}
		if(checkInterSameTreeCompatibility(parentToLeftMapping, tentativeMapping) && checkInterSameTreeInAltCompatibility(parentToLeftMapping, parentToRightMapping)) {
			 int newInterSameOutCost = getInterSameOutCost(parentToLeftMapping, tentativeMapping, parentToRightMapping);
			if(newInterSameOutCost<parentToLeftMapping.getInterSameTreeOutCost())	
			 {
				parentToLeftMapping.setInterSameTreeOutCost(newInterSameOutCost);
				setInterSameOutNodes(parentToLeftMapping, tentativeMapping, parentToRightMapping);
			 }
		}
		else if(checkInterSameTreeCompatibility(parentToLeftMapping, tentativeMapping)) {
			 int newInterSameOutCost = getInterSameOutCost(parentToLeftMapping, tentativeMapping);
			 if(newInterSameOutCost<parentToLeftMapping.getInterSameTreeOutCost())	
			 {
				parentToLeftMapping.setInterSameTreeOutCost(newInterSameOutCost);
				setInterSameOutNodes(parentToLeftMapping, tentativeMapping);
			 }
		}
		
		else if(checkInterSameTreeInAltCompatibility(parentToLeftMapping, parentToRightMapping)) {
			 int newInterSameOutCost = getInterSameOutCostFromInAlt(parentToLeftMapping, parentToRightMapping);
			 if(newInterSameOutCost<parentToLeftMapping.getInterSameTreeOutCost())	
			 {
			 parentToLeftMapping.setInterSameTreeOutCost(newInterSameOutCost);
				setInterSameOutNodesFromInAlt(parentToLeftMapping, parentToRightMapping);
			 }
		}
		
	}
	
	private void handleIntraSameOutCostAndMapping(DomainMapping parentToLeftMapping,
			DomainMapping tentativeMapping, DomainMapping parentToRightMapping) {
		if(checkIntraSameTreeCompatibility(parentToLeftMapping, tentativeMapping) && checkIntraSameTreeInAltCompatibility(parentToLeftMapping, parentToRightMapping)) {
			 int newIntraSameOutCost = getIntraSameOutCost(parentToLeftMapping, tentativeMapping, parentToRightMapping);
			 if(newIntraSameOutCost<parentToLeftMapping.getIntraSameTreeOutCost())
				{	parentToLeftMapping.setIntraSameTreeOutCost(newIntraSameOutCost);
				setIntraSameOutNodes(parentToLeftMapping, tentativeMapping, parentToRightMapping);
				}
		}
		else if(checkIntraSameTreeCompatibility(parentToLeftMapping, tentativeMapping)) {
			 int newIntraSameOutCost = getIntraSameOutCost(parentToLeftMapping, tentativeMapping);
			 if(newIntraSameOutCost<parentToLeftMapping.getIntraSameTreeOutCost())
				{
				parentToLeftMapping.setIntraSameTreeOutCost(newIntraSameOutCost);
				setIntraSameOutNodes(parentToLeftMapping, tentativeMapping);
				}
		}
		
		else if(checkIntraSameTreeInAltCompatibility(parentToLeftMapping, parentToRightMapping)) {
			 int newIntraSameOutCost = getIntraSameOutCostFromInAlt(parentToLeftMapping, parentToRightMapping);
			if(newIntraSameOutCost<parentToLeftMapping.getIntraSameTreeOutCost())
			{	
				parentToLeftMapping.setIntraSameTreeOutCost(newIntraSameOutCost);
				setIntraSameOutNodesFromInAlt(parentToLeftMapping, parentToRightMapping);
			}
		}
		
	}

	private void handleIntraDifferentOutCostAndMapping(DomainMapping parentToLeftMapping,
			DomainMapping tentativeMapping) {
		if(checkIntraDifferentTreeCompatibility(parentToLeftMapping, tentativeMapping)) {
			 int newIntraDifferentOutCost = getIntraDifferentOutCost(parentToLeftMapping, tentativeMapping);
			 if(newIntraDifferentOutCost<parentToLeftMapping.getIntraDifferentTreeOutCost()) {
				 parentToLeftMapping.setIntraDifferentTreeOutCost(newIntraDifferentOutCost);
				setIntraDifferentOutNodes(parentToLeftMapping, tentativeMapping);
			 }
				
			}
	}

	private void handleInterDifferentOutCostAndMapping(DomainMapping parentToLeftMapping, DomainMapping tentativeMapping) {
		if(checkInterDifferentTreeCompatibility(parentToLeftMapping, tentativeMapping)) {
		 int newInterDifferentOutCost = getInterDifferentOutCost(parentToLeftMapping, tentativeMapping);
		 if(newInterDifferentOutCost<parentToLeftMapping.getInterDifferentTreeOutCost()) {
			parentToLeftMapping.setInterDifferentTreeOutCost(newInterDifferentOutCost);
			setInterDifferentOutNodes(parentToLeftMapping, tentativeMapping);
		 }
		}
	
	}
	private void setInterSameOutNodesFromInAlt(DomainMapping parentToLeftMapping, DomainMapping paretToRightMapping) {
		Node firstGene = parentToLeftMapping.getSecondNode();
		parentToLeftMapping.getInterSameTreeOutNodes().clear();
		if(parentToLeftMapping.getInterSameTreeOutCost()==paretToRightMapping.getInAltCost()) {
			for(Node secondGene: paretToRightMapping.getInAltNodes()) {
				if(!checkIntraSpeciesCompatibility(firstGene, secondGene))
				{	
					parentToLeftMapping.getInterSameTreeOutNodes().add(secondGene);
				}
			}
		}
	}
	private void setIntraSameOutNodesFromInAlt(DomainMapping parentToLeftMapping, DomainMapping paretToRightMapping) {
		Node firstGene = parentToLeftMapping.getSecondNode();
		parentToLeftMapping.getIntraSameTreeOutNodes().clear();
		if(parentToLeftMapping.getIntraSameTreeOutCost()==paretToRightMapping.getInAltCost()) {
			for(Node secondGene: paretToRightMapping.getInAltNodes()) {
				if(checkIntraSpeciesCompatibility(firstGene, secondGene))
				{	
					parentToLeftMapping.getIntraSameTreeOutNodes().add(secondGene);
				}
			}
		}
		
		
		
		
	}
	private void setIntraSameOutNodes(DomainMapping parentToLeftMapping, DomainMapping tentativeMapping) {
		Node firstGene = parentToLeftMapping.getSecondNode();
		parentToLeftMapping.getIntraSameTreeOutNodes().clear();
		if(parentToLeftMapping.getIntraSameTreeOutCost()==tentativeMapping.getInterSameTreeOutCost()) {
			for(Node secondGene: tentativeMapping.getInterSameTreeOutNodes()) {
				if(checkIntraSpeciesCompatibility(firstGene, secondGene))
				{	
					parentToLeftMapping.getIntraSameTreeOutNodes().add(secondGene);
				}
				}
		}
		
		if(parentToLeftMapping.getIntraSameTreeOutCost()==tentativeMapping.getIntraSameTreeOutCost()) {
			for(Node secondGene: tentativeMapping.getIntraSameTreeOutNodes()) {
				if(checkIntraSpeciesCompatibility(firstGene, secondGene))
				{	
					parentToLeftMapping.getIntraSameTreeOutNodes().add(secondGene);
				}
				}
		}
		
		
	}
	
	private void setInterSameOutNodes(DomainMapping parentToLeftMapping, DomainMapping tentativeMapping) {
		Node firstGene = parentToLeftMapping.getSecondNode();
		parentToLeftMapping.getInterSameTreeOutNodes().clear();
		if(parentToLeftMapping.getInterSameTreeOutCost()==tentativeMapping.getInterSameTreeOutCost()) {
			for(Node secondGene: tentativeMapping.getInterSameTreeOutNodes()) {
				if(!checkIntraSpeciesCompatibility(firstGene, secondGene))
				{	
					parentToLeftMapping.getInterSameTreeOutNodes().add(secondGene);
				}
				}
		}
		
		if(parentToLeftMapping.getInterSameTreeOutCost()==tentativeMapping.getIntraSameTreeOutCost()) {
			for(Node secondGene: tentativeMapping.getIntraSameTreeOutNodes()) {
				if(!checkIntraSpeciesCompatibility(firstGene, secondGene))
				{	
					parentToLeftMapping.getInterSameTreeOutNodes().add(secondGene);
				}
				}
		}
		
	}
	
	private void setInterSameOutNodes(DomainMapping parentToLeftMapping, DomainMapping tentativeMapping, DomainMapping parentToRightMapping) {
		Node firstGene = parentToLeftMapping.getSecondNode();
		parentToLeftMapping.getInterSameTreeOutNodes().clear();
		if(parentToLeftMapping.getInterSameTreeOutCost()==tentativeMapping.getInterSameTreeOutCost()) {
			for(Node secondGene: tentativeMapping.getInterSameTreeOutNodes()) {
				if(!checkIntraSpeciesCompatibility(firstGene, secondGene))
				{	
					parentToLeftMapping.getInterSameTreeOutNodes().add(secondGene);
				}
				}
		}
		
		if(parentToLeftMapping.getInterSameTreeOutCost()==tentativeMapping.getIntraSameTreeOutCost()) {
			for(Node secondGene: tentativeMapping.getIntraSameTreeOutNodes()) {
				if(!checkIntraSpeciesCompatibility(firstGene, secondGene))
				{	
					parentToLeftMapping.getInterSameTreeOutNodes().add(secondGene);
				}
				}
		}
		if(parentToLeftMapping.getInterSameTreeOutCost()==parentToRightMapping.getInAltCost()) {
			for(Node secondGene: tentativeMapping.getInAltNodes()) {
				if(!checkIntraSpeciesCompatibility(firstGene, secondGene))
				{	
					parentToLeftMapping.getInterSameTreeOutNodes().add(secondGene);
				}
				}
		}
		
	}
	private void setIntraSameOutNodes(DomainMapping parentToLeftMapping, DomainMapping tentativeMapping, DomainMapping parentToRightMapping) {
		Node firstGene = parentToLeftMapping.getSecondNode();
		parentToLeftMapping.getIntraSameTreeOutNodes().clear();
		if(parentToLeftMapping.getIntraSameTreeOutCost()==tentativeMapping.getInterSameTreeOutCost()) {
			for(Node secondGene: tentativeMapping.getInterSameTreeOutNodes()) {
				if(checkIntraSpeciesCompatibility(firstGene, secondGene))
				{	
					parentToLeftMapping.getIntraSameTreeOutNodes().add(secondGene);
				}
				}
		}
		
		if(parentToLeftMapping.getIntraSameTreeOutCost()==tentativeMapping.getIntraSameTreeOutCost()) {
			for(Node secondGene: tentativeMapping.getIntraSameTreeOutNodes()) {
				if(checkIntraSpeciesCompatibility(firstGene, secondGene))
				{	
					parentToLeftMapping.getIntraSameTreeOutNodes().add(secondGene);
				}
			}
		}
		if(parentToLeftMapping.getIntraSameTreeOutCost()==parentToRightMapping.getInAltCost()) {
			for(Node secondGene: tentativeMapping.getInAltNodes()) {
				if(checkIntraSpeciesCompatibility(firstGene, secondGene))
				{	
					parentToLeftMapping.getIntraSameTreeOutNodes().add(secondGene);
				}
				}
		}
		
	}
	private void setIntraDifferentOutNodes(DomainMapping parentToLeftMapping, DomainMapping tentativeMapping) {
		Node firstGene = parentToLeftMapping.getSecondNode();
		parentToLeftMapping.getIntraDifferentTreeOutNodes().clear();
		if(parentToLeftMapping.getIntraDifferentTreeOutCost()==tentativeMapping.getInterDifferentTreeOutCost()) {
			for(Node secondGene: tentativeMapping.getInterDifferentTreeOutNodes()) {
				if(checkIntraSpeciesCompatibility(firstGene, secondGene)
				    && ! checkSameGeneTreeCompatibility(firstGene, secondGene))
				{	
					parentToLeftMapping.getIntraDifferentTreeOutNodes().add(secondGene);
				}
				}
		}
		
		if(parentToLeftMapping.getIntraDifferentTreeOutCost()==tentativeMapping.getIntraDifferentTreeOutCost()) {
			for(Node secondGene: tentativeMapping.getIntraDifferentTreeOutNodes()) {
				if(checkIntraSpeciesCompatibility(firstGene, secondGene)
				    && !checkSameGeneTreeCompatibility(firstGene, secondGene))
				{	
					parentToLeftMapping.getIntraDifferentTreeOutNodes().add(secondGene);
				}
				}
		}
	}
	private void setInterDifferentOutNodes(DomainMapping parentToLeftMapping, DomainMapping tentativeMapping) {
		Node firstGene = parentToLeftMapping.getSecondNode();
		parentToLeftMapping.getInterDifferentTreeOutNodes().clear();
		if(parentToLeftMapping.getInterDifferentTreeOutCost()==tentativeMapping.getInterDifferentTreeOutCost()) {
			for(Node secondGene: tentativeMapping.getInterDifferentTreeOutNodes()) {
				if(!checkIntraSpeciesCompatibility(firstGene, secondGene)
				    && ! checkSameGeneTreeCompatibility(firstGene, secondGene))
				{	
					parentToLeftMapping.getInterDifferentTreeOutNodes().add(secondGene);
				}
				}
		}
		
		if(parentToLeftMapping.getInterDifferentTreeOutCost()==tentativeMapping.getIntraDifferentTreeOutCost()) {
			for(Node secondGene: tentativeMapping.getIntraDifferentTreeOutNodes()) {
				if(!checkIntraSpeciesCompatibility(firstGene, secondGene)
				    && ! checkSameGeneTreeCompatibility(firstGene, secondGene))
				{	
					parentToLeftMapping.getInterDifferentTreeOutNodes().add(secondGene);
				}
				}
		}
	}
	private int getInterSameOutCostFromInAlt(DomainMapping parentToLeftMapping, DomainMapping parentToRightMapping ) {
		Node firstGene = parentToLeftMapping.getSecondNode();
		int minimumInterSameCost = Integer.MAX_VALUE/10;
		//Node secondGene = secondMapping.getSecondNode();
		for(Node secondGene: parentToRightMapping.getInAltNodes()) {
		if(!checkIntraSpeciesCompatibility(firstGene, secondGene))
		{	
			minimumInterSameCost = parentToRightMapping.getInAltCost();
			break;
		}
		}
	
		
		
		return minimumInterSameCost;
	}
	private int getIntraSameOutCostFromInAlt(DomainMapping parentToLeftMapping, DomainMapping parentToRightMapping ) {
		Node firstGene = parentToLeftMapping.getSecondNode();
		int minimumIntraSameCost = Integer.MAX_VALUE/10;
		//Node secondGene = secondMapping.getSecondNode();
		for(Node secondGene: parentToRightMapping.getInAltNodes()) {
		if(checkIntraSpeciesCompatibility(firstGene, secondGene))
		{	if(minimumIntraSameCost>parentToRightMapping.getInAltCost()) {
			minimumIntraSameCost = parentToRightMapping.getInAltCost();
			break;
			}
		}
		}
	
		
		
		return minimumIntraSameCost;
	}

	private int getIntraSameOutCost(DomainMapping parentToLeftMapping, DomainMapping tentativeMapping ) {
		Node firstGene = parentToLeftMapping.getSecondNode();
		int minimumIntraSameCost = Integer.MAX_VALUE/10;
		//Node secondGene = secondMapping.getSecondNode();
		for(Node secondGene: tentativeMapping.getIntraSameTreeOutNodes()) {
		if(checkIntraSpeciesCompatibility(firstGene, secondGene))
		{	
			if(minimumIntraSameCost>tentativeMapping.getIntraSameTreeOutCost())
			{minimumIntraSameCost = tentativeMapping.getIntraSameTreeOutCost();
			break;}
		}
		}
		for(Node secondGene: tentativeMapping.getInterSameTreeOutNodes()) {
			if(checkIntraSpeciesCompatibility(firstGene, secondGene))
			{
				if(minimumIntraSameCost>tentativeMapping.getInterSameTreeOutCost()) {
					minimumIntraSameCost = tentativeMapping.getInterSameTreeOutCost();
					break;
				}	
			}
		}
		
		
		
		return minimumIntraSameCost;
	}
	private int getInterSameOutCost(DomainMapping parentToLeftMapping, DomainMapping tentativeMapping) {
		Node firstGene = parentToLeftMapping.getSecondNode();
		int minimumInterSameCost = Integer.MAX_VALUE/10;
		//Node secondGene = secondMapping.getSecondNode();
		for(Node secondGene: tentativeMapping.getIntraSameTreeOutNodes()) {
		if(!checkIntraSpeciesCompatibility(firstGene, secondGene))
		{	
			minimumInterSameCost = tentativeMapping.getIntraSameTreeOutCost();
			break;
		}
		}
		for(Node secondGene: tentativeMapping.getInterSameTreeOutNodes()) {
			if(!checkIntraSpeciesCompatibility(firstGene, secondGene))
			{
				if(minimumInterSameCost>tentativeMapping.getInterSameTreeOutCost()) {
					minimumInterSameCost = tentativeMapping.getInterSameTreeOutCost();
					break;
				}	
			}
		}
		
		
		
		return minimumInterSameCost;
	}
	private int getInterSameOutCost(DomainMapping parentToLeftMapping, DomainMapping tentativeMapping, DomainMapping parentToRightMapping) {
		Node firstGene = parentToLeftMapping.getSecondNode();
		int minimumInterSameCost = Integer.MAX_VALUE/10;
		//Node secondGene = secondMapping.getSecondNode();
		for(Node secondGene: tentativeMapping.getIntraSameTreeOutNodes()) {
		if(!checkIntraSpeciesCompatibility(firstGene, secondGene))
		{	
			minimumInterSameCost = tentativeMapping.getIntraSameTreeOutCost();
			break;
		}
		}
		for(Node secondGene: tentativeMapping.getInterSameTreeOutNodes()) {
			if(!checkIntraSpeciesCompatibility(firstGene, secondGene))
			{
				if(minimumInterSameCost>tentativeMapping.getInterSameTreeOutCost()) {
					minimumInterSameCost = tentativeMapping.getInterSameTreeOutCost();
					break;
				}	
			}
		}
		
		for(Node secondGene: parentToRightMapping.getInAltNodes()) {
			if(!checkIntraSpeciesCompatibility(firstGene, secondGene))
			{
				if(minimumInterSameCost>parentToRightMapping.getInAltCost()) {
					minimumInterSameCost = parentToRightMapping.getInAltCost();
					break;
				}	
			}
		}
		
		return minimumInterSameCost;
	}
	
	private int getIntraSameOutCost(DomainMapping parentToLeftMapping, DomainMapping tentativeMapping, DomainMapping parentToRightMapping) {
		Node firstGene = parentToLeftMapping.getSecondNode();
		int minimumIntraSameCost = Integer.MAX_VALUE/10;
		//Node secondGene = secondMapping.getSecondNode();
		for(Node secondGene: tentativeMapping.getIntraSameTreeOutNodes()) {
		if(checkIntraSpeciesCompatibility(firstGene, secondGene))
		{	
			minimumIntraSameCost = tentativeMapping.getIntraSameTreeOutCost();
			break;
		}
		}
		for(Node secondGene: tentativeMapping.getInterSameTreeOutNodes()) {
			if(checkIntraSpeciesCompatibility(firstGene, secondGene))
			{
				if(minimumIntraSameCost>tentativeMapping.getInterSameTreeOutCost()) {
					minimumIntraSameCost = tentativeMapping.getInterSameTreeOutCost();
					break;
				}	
			}
		}
		
		for(Node secondGene: parentToRightMapping.getInAltNodes()) {
			if(checkIntraSpeciesCompatibility(firstGene, secondGene))
			{
				if(minimumIntraSameCost>parentToRightMapping.getInAltCost()) {
					minimumIntraSameCost = parentToRightMapping.getInAltCost();
					break;
				}	
			}
		}
		
		return minimumIntraSameCost;
	}
	private int getIntraDifferentOutCost(DomainMapping parentToLeftMapping, DomainMapping tentativeMapping) {
		Node firstGene = parentToLeftMapping.getSecondNode();
		int minimumIntraDifferentCost = Integer.MAX_VALUE/10;
		
		for(Node secondGene: tentativeMapping.getInterDifferentTreeOutNodes()) {
		if(checkIntraSpeciesCompatibility(firstGene, secondGene)
		    && ! checkSameGeneTreeCompatibility(firstGene, secondGene))
		{	
			minimumIntraDifferentCost = tentativeMapping.getInterDifferentTreeOutCost();
			break;
		}
		}
		for(Node secondGene: tentativeMapping.getIntraDifferentTreeOutNodes()) {
			if(checkIntraSpeciesCompatibility(firstGene, secondGene)
			    && ! checkSameGeneTreeCompatibility(firstGene, secondGene))
			{
				if(minimumIntraDifferentCost>tentativeMapping.getIntraDifferentTreeOutCost()) {
					minimumIntraDifferentCost = tentativeMapping.getIntraDifferentTreeOutCost();
					break;
				}	
			}
		}
		return minimumIntraDifferentCost;
	}
	
	private int getInterDifferentOutCost(DomainMapping parentToLeftMapping, DomainMapping tentativeMapping) {
		Node firstGene = parentToLeftMapping.getSecondNode();
		int minimumInterDifferentCost = Integer.MAX_VALUE/10;
		//Node secondGene = secondMapping.getSecondNode();
		for(Node secondGene: tentativeMapping.getInterDifferentTreeOutNodes()) {
		if(!checkIntraSpeciesCompatibility(firstGene, secondGene)
		    && ! checkSameGeneTreeCompatibility(firstGene, secondGene))
		{	
			minimumInterDifferentCost = tentativeMapping.getInterDifferentTreeOutCost();
			break;
		}
		}
		for(Node secondGene: tentativeMapping.getIntraDifferentTreeOutNodes()) {
			if(!checkIntraSpeciesCompatibility(firstGene, secondGene)
			    && ! checkSameGeneTreeCompatibility(firstGene, secondGene))
			{
				if(minimumInterDifferentCost>tentativeMapping.getIntraDifferentTreeOutCost()) {
					minimumInterDifferentCost = tentativeMapping.getIntraDifferentTreeOutCost();
					break;
				}	
			}
		}
		return minimumInterDifferentCost;
	}



	



	private boolean checkIntraSameTreeInAltCompatibility(DomainMapping firstMapping,
			DomainMapping inAltMapping) {
		Node firstGene = firstMapping.getSecondNode();
		ArrayList<Node>inAltNodes = inAltMapping.getInAltNodes();
		for(Node inAltNode: inAltNodes) {
			if(checkIntraSpeciesCompatibility(firstGene, inAltNode)
				    && checkSameGeneTreeCompatibility(firstGene, inAltNode))
					return true;
		}
		return false;
	}

	
	
	
	
	private boolean checkIntraSameTreeCompatibility(DomainMapping firstMapping, DomainMapping secondMapping) {
		Node firstGene = firstMapping.getSecondNode();
		
		for(Node secondGene: secondMapping.getIntraSameTreeOutNodes()) {
		if(checkIntraSpeciesCompatibility(firstGene, secondGene)
		    && checkSameGeneTreeCompatibility(firstGene, secondGene))
			return true;
		}
		for(Node secondGene: secondMapping.getInterSameTreeOutNodes()) {
			if(checkIntraSpeciesCompatibility(firstGene, secondGene)
			    && checkSameGeneTreeCompatibility(firstGene, secondGene))
				return true;
		}
		return false;
	}
	
	private boolean checkIntraDifferentTreeCompatibility(DomainMapping firstMapping, DomainMapping secondMapping) {
		Node firstGene = firstMapping.getSecondNode();
		//Node secondGene = secondMapping.getSecondNode();
		for(Node secondGene: secondMapping.getInterDifferentTreeOutNodes()) {
		if(checkIntraSpeciesCompatibility(firstGene, secondGene)
		    && ! checkSameGeneTreeCompatibility(firstGene, secondGene))
			return true;
		}
		for(Node secondGene: secondMapping.getIntraDifferentTreeOutNodes()) {
			if(checkIntraSpeciesCompatibility(firstGene, secondGene)
			    && ! checkSameGeneTreeCompatibility(firstGene, secondGene))
				return true;
		}
		return false;
	}
	
	private boolean checkInterSameTreeCompatibility(DomainMapping firstMapping, DomainMapping secondMapping) {
		Node firstGene = firstMapping.getSecondNode();
		
		for(Node secondGene: secondMapping.getIntraSameTreeOutNodes()) {
		if(!checkIntraSpeciesCompatibility(firstGene, secondGene)
		    && checkSameGeneTreeCompatibility(firstGene, secondGene))
			return true;
		}
		for(Node secondGene: secondMapping.getInterSameTreeOutNodes()) {
			if(!checkIntraSpeciesCompatibility(firstGene, secondGene)
			    && checkSameGeneTreeCompatibility(firstGene, secondGene))
				return true;
		}
		return false;
	}
	private boolean checkInterSameTreeInAltCompatibility(DomainMapping firstMapping, DomainMapping secondMapping) {
	Node firstGene = firstMapping.getSecondNode();
		
		for(Node secondGene: secondMapping.getInAltNodes()) {
		if(!checkIntraSpeciesCompatibility(firstGene, secondGene)
		    && checkSameGeneTreeCompatibility(firstGene, secondGene))
			return true;
		}
		
		return false;
	}
	private boolean checkInterDifferentTreeCompatibility(DomainMapping firstMapping, DomainMapping secondMapping) {
		Node firstGene = firstMapping.getSecondNode();
		//Node secondGene = secondMapping.getSecondNode();
		for(Node secondGene: secondMapping.getInterDifferentTreeOutNodes()) {
		if(!checkIntraSpeciesCompatibility(firstGene, secondGene)
		    && ! checkSameGeneTreeCompatibility(firstGene, secondGene))
			return true;
		}
		for(Node secondGene: secondMapping.getIntraDifferentTreeOutNodes()) {
			if(!checkIntraSpeciesCompatibility(firstGene, secondGene)
			    && ! checkSameGeneTreeCompatibility(firstGene, secondGene))
				return true;
		}
		return false;
	}
	
	
	
	private void initializeGeneToSpeciesLeafMappings(ArrayList<Node>geneLeaves, ArrayList<Node> speciesLeaves) {
		// for each leaf node of gene trees
		for (Node geneLeaf: geneLeaves) {
			GeneNodeRooted geneLeafRooted = (GeneNodeRooted) geneLeaf;
			//for each leaf node of species trees
			for (Node speciesLeaf: speciesLeaves) {
				if(geneLeafRooted.getSpeciesName().equals(speciesLeaf.getName())) {
					GeneMapping mapping = getGeneMappingFromNodes(geneLeaf, speciesLeaf);
					// set options to 1
					mapping.setNumberOfSolutions(new BigInteger("1"));
					// set cost and in-cost to zero
					mapping.setCost(0);
					mapping.setInCost(0);
					mapping.setInAltCost(0);

					ArrayList<RootedBinaryTreeMapping> optimalMappingsForSingleGeneNode = new ArrayList<>();

					optimalMappingsForSingleGeneNode.add(mapping);
					mapping.getOptions().add(0);
					mapping.getOptionWeights().add(new BigInteger("1"));
					mapping.addInSpeciesNode(speciesLeaf);
					mapping.addInAltSpeciesNode(speciesLeaf);
					// allPossibleMappings[leafGeneNodeIndex][ancestorSpeciesNodeIndex].getOptions().add(0);
					// add leaf to leaf match to optimal mappings
					optimalMappings.put(geneLeaf, optimalMappingsForSingleGeneNode);
					// get all ancestors of this species node
					ArrayList<Node> ancestorSpeciesNodes = speciesLeaf.getAllAncestors();

					// for each ancestor s' of this species node
					for (Node ancestorSpeciesNode : ancestorSpeciesNodes) {

						// get distance from species node to ancestor node
						int distance = speciesLeaf.getDistance(ancestorSpeciesNode);

						// set inAlt(g, s') = 0 and in(g, s') = distance*LOSS
						RootedBinaryTreeMapping ancestorMapping = getGeneMappingFromNodes(geneLeaf,
								ancestorSpeciesNode);
						ancestorMapping.setInCost(GENE_LOSS * distance);
						ancestorMapping.setInAltCost(0);
						ancestorMapping.addInSpeciesNode(speciesLeaf);
						ancestorMapping.addInAltSpeciesNode(speciesLeaf);

					}
				}
			}
		}
		
		for (Node geneLeaf : geneLeaves) {
		//	System.out.println(((GeneNodeRooted) geneLeaf).getFullName());
			Node actualMappedNode = optimalMappings.get(geneLeaf).get(0).getSecondNode();
			ArrayList<Node> speciesNodes = speciesTree.getAllNodesByPostOrder();
			for (Node speciesNode : speciesNodes) {
				if (!speciesNode.equals(actualMappedNode) && !actualMappedNode.isDescendant(speciesNode)) {
					RootedBinaryTreeMapping mapping = getGeneMappingFromNodes(geneLeaf, speciesNode);
					mapping.addOutSpeciesNode(actualMappedNode);
					mapping.setOutCost(0);
				}
			}
		}
	}

	private void initDToGLeafMappingsApproximately(ArrayList<Node> leafDomainNodes, ArrayList<Node> leafGeneNodes) {


		// for each leaf node of domain tree
		for (Node domainLeafNode : leafDomainNodes) {
			DomainNodeRooted domainLeafNodeRooted = (DomainNodeRooted) domainLeafNode;
			// for each leaf node of gene trees
			for (Node geneLeafNode : leafGeneNodes) {
				GeneNodeRooted leafGeneNodeRooted = (GeneNodeRooted)geneLeafNode;
				// if gene name and domain nodes' gene names are equal
				if (domainLeafNodeRooted.getGeneName().equals(leafGeneNodeRooted.getDomainPart())) {
					//update the gene tree as used
					geneTrees.get(leafGeneNodeRooted.getGeneTreeID()).setUsed(true);

					DomainMapping mapping = getDomainMappingFromNodes(domainLeafNode, geneLeafNode);
					// set number of solutions to 1
					mapping.setNumberOfSolutions(new BigInteger("1"));
					// set cost and in-cost, in-alt cost to zero
					mapping.setCost(0);
					mapping.setInCost(0);
					mapping.setInAltCost(0);

			
					ArrayList<RootedBinaryTreeMapping> optimalMappingsForSingleGeneNode = new ArrayList<>();

					optimalMappingsForSingleGeneNode.add(mapping);
					mapping.getOptions().add(0);
					mapping.getOptionWeights().add(new BigInteger("1"));
					mapping.addInSpeciesNode(geneLeafNode);
					mapping.addInAltSpeciesNode(geneLeafNode);
					
					// add leaf to leaf match to optimal mappings
					optimalMappings.put(domainLeafNode, optimalMappingsForSingleGeneNode);
					// get all ancestors of this species node
					ArrayList<Node> ancestorGeneNodes = geneLeafNode.getAllAncestors();

					// for each ancestor s' of this species node
					for (Node ancestorGeneNode : ancestorGeneNodes) {

						// get distance from species node to ancestor node
						int distance = geneLeafNode.getDistance(ancestorGeneNode);

						// set inAlt(g, s') = 0 and in(g, s') = distance*LOSS
						RootedBinaryTreeMapping ancestorMapping = getDomainMappingFromNodes(domainLeafNode,
								ancestorGeneNode);
						ancestorMapping.setInCost(DOMAIN_LOSS * distance);
						ancestorMapping.setInAltCost(0);
						ancestorMapping.addInSpeciesNode(geneLeafNode);
						ancestorMapping.addInAltSpeciesNode(geneLeafNode);

					}

				}
			}
		}
		// set the out costs of non mapped nodes
		for (Node domainLeafNode : leafDomainNodes) {			
			Node actualMappedNode = optimalMappings.get(domainLeafNode).get(0).getSecondNode();
			ArrayList<Node> geneNodes = getAllGeneNodesByPostOrder();
			for (Node geneNode : geneNodes) {
				if (!geneNode.equals(actualMappedNode) 
						&& 
						!actualMappedNode.isDescendant(geneNode)
						) {
					DomainMapping mapping = getDomainMappingFromNodes(domainLeafNode, geneNode);
				
					mapping.setOutCost(0);
					mapping.addOutSpeciesNode(actualMappedNode);
				}
			}
		}
	
	}
	
	private void initializeDomainToGeneLeafMappings(ArrayList<Node> leafDomainNodes, ArrayList<Node> leafGeneNodes) {

		// for each leaf node of domain tree
		for (Node domainLeafNode : leafDomainNodes) {
			DomainNodeRooted domainLeafNodeRooted = (DomainNodeRooted) domainLeafNode;
			// for each leaf node of gene trees
			for (Node geneLeafNode : leafGeneNodes) {
				GeneNodeRooted geneLeafRooted = (GeneNodeRooted)geneLeafNode;
				// if gene name and domain nodes' gene names are equal
				if (domainLeafNodeRooted.getGeneName().equals(geneLeafRooted.getDomainPart())) {
					//mark the gene tree as used
					geneTrees.get(geneLeafRooted.getGeneTreeID()).setUsed(true);
					
					DomainMapping mapping = getDomainMappingFromNodes(domainLeafNode, geneLeafNode);
					// set number of solutions to 1
					//mapping.setNumberOfSolutions(new BigInteger("1"));
					// set cost and in-cost, in-alt cost to zero
					mapping.setCost(0);
					mapping.setInCost(0);
					mapping.setInAltCost(0);

			
					ArrayList<RootedBinaryTreeMapping> optimalMappingsForSingleGeneNode = new ArrayList<>();

					optimalMappingsForSingleGeneNode.add(mapping);
					mapping.getOptions().add(0);
					//mapping.getOptionWeights().add(new BigInteger("1"));
					mapping.addInSpeciesNode(geneLeafNode);
					mapping.addInAltSpeciesNode(geneLeafNode);
					
					// add leaf to leaf match to optimal mappings
					optimalMappings.put(domainLeafNode, optimalMappingsForSingleGeneNode);
					// get all ancestors of this species node
					ArrayList<Node> ancestorGeneNodes = geneLeafNode.getAllAncestors();

					// for each ancestor s' of this species node
					for (Node ancestorGeneNode : ancestorGeneNodes) {

						// get distance from species node to ancestor node
						int distance = geneLeafNode.getDistance(ancestorGeneNode);

						// set inAlt(g, s') = 0 and in(g, s') = distance*LOSS
						RootedBinaryTreeMapping ancestorMapping = getDomainMappingFromNodes(domainLeafNode,
								ancestorGeneNode);
						ancestorMapping.setInCost(DOMAIN_LOSS * distance);
						ancestorMapping.setInAltCost(0);
						ancestorMapping.addInSpeciesNode(geneLeafNode);
						ancestorMapping.addInAltSpeciesNode(geneLeafNode);

					}

				}
			}
		}
		// set the out costs of non mapped nodes
		for (Node domainLeafNode : leafDomainNodes) {			
			Node actualMappedNode = optimalMappings.get(domainLeafNode).get(0).getSecondNode();
			ArrayList<Node> geneNodes = getAllGeneNodesByPostOrder();
			for (Node geneNode : geneNodes) {
				if (!geneNode.equals(actualMappedNode) && !actualMappedNode.isDescendant(geneNode)) {
					DomainMapping mapping = getDomainMappingFromNodes(domainLeafNode, geneNode);
				//	mapping.addOutSpeciesNode(actualMappedNode);
				//	mapping.setOutCost(0);
					if(checkIntraSpeciesCompatibility(actualMappedNode, geneNode)) {
						if(checkSameGeneTreeCompatibility(actualMappedNode, geneNode)) {
							mapping.setIntraSameTreeOutCost(0);
							mapping.getIntraSameTreeOutNodes().add(actualMappedNode);
						}
						else {
							mapping.setIntraDifferentTreeOutCost(0);
							mapping.getIntraDifferentTreeOutNodes().add(actualMappedNode);
						}
					}
					else {
						if(checkSameGeneTreeCompatibility(actualMappedNode, geneNode)) {
							mapping.setInterSameTreeOutCost(0);
							mapping.getInterSameTreeOutNodes().add(actualMappedNode);
								
						}
						else {
							mapping.setInterDifferentTreeOutCost(0);
							mapping.getInterDifferentTreeOutNodes().add(actualMappedNode);
							
						}
					}
				}
			}
		}
	}
	public boolean checkSameGeneTreeCompatibility(Node firstNode, Node secondNode) {
		// get first and second rooted genenode
		GeneNodeRooted firstGeneNode = (GeneNodeRooted) firstNode;
		GeneNodeRooted secondGeneNode = (GeneNodeRooted) secondNode;
		// get the tree IDs
		int firstGeneTreeID = firstGeneNode.getGeneTreeID();
		int secondGeneTreeID = secondGeneNode.getGeneTreeID();
		// if the IDs are equal then they are in the same  tree
		if(!firstGeneNode.equals(secondGeneNode)&& firstGeneTreeID == secondGeneTreeID)return true;
		return false;
	}
	public boolean checkIntraSpeciesCompatibility(Node firstNode, Node secondNode) {
		// get first and second rooted genenode
		GeneNodeRooted firstGeneNode = (GeneNodeRooted) firstNode;
		GeneNodeRooted secondGeneNode = (GeneNodeRooted) secondNode;
		// get the mapped species nodes
		Node firstMappedSpecies =  firstGeneNode.getMapping().getSecondNode();
		Node secondMappedSpecies =  secondGeneNode.getMapping().getSecondNode();
		// if they equal then they are intra species transfer compatible
		if(!firstNode.equals(secondNode) && firstMappedSpecies.equals(secondMappedSpecies))return true;
		return false;
	}

	public DomainMapping getDomainMappingFromNodes(Node domainNode, Node geneNode) {
		int domainNodeIndex = domainNode.getMappingIndex();
		int geneNodeIndex = geneNode.getMappingIndex();
		return allPossibleDomainMappings[domainNodeIndex][geneNodeIndex];
	}
	private GeneMapping getGeneMappingFromNodes(Node geneNode, Node speciesNode) {
		int geneNodeIndex = geneNode.getMappingIndex();
		int speciesNodeIndex = speciesNode.getPostOrderNumber();
		return allPossibleGeneMappings[geneNodeIndex][speciesNodeIndex];
	}
	private ArrayList<Node> getAllPreOrderInternalGeneNodes() {
		 ArrayList<Node>allInternalGeneNodes= new ArrayList<Node>();
		 for (GeneTreeRooted geneTree: geneTrees) {
			 ArrayList<Node>internalGeneNodes = geneTree.getAllPreOrderInternalNodes();
			 allInternalGeneNodes.addAll(internalGeneNodes);
		 }
		return allInternalGeneNodes;
	}

	public void performPreprocessings() {
		
		speciesTree.computeAllPostOrderNumbers();
		speciesTree.computeAllInOrderNumbers();
		speciesTree.computeLeafSet();
		
		speciesTree.computeAllPostOrderInternalNodes();
		speciesTree.setMappingIndexes();
		speciesTree.setAllBranches();
		for (DomainTreeRooted domainTree: domainTrees) {
			domainTree.computeAllPostOrderNumbers();
			domainTree.computeAllInOrderNumbers();
			domainTree.computeLeafSet();
		//	domainTree.computeAllPreOrderInternalNodes();
			domainTree.computeAllPostOrderInternalNodes();
			domainTree.setAllBranches();
		}
		setMappingIndexesOfDomainTrees();
		for (GeneTreeRooted geneTree: geneTrees) {
			geneTree.computeAllPostOrderNumbers();
			geneTree.computeAllInOrderNumbers();
			geneTree.computeLeafSet();
		//	geneTree.computeAllPreOrderInternalNodes();
			geneTree.computeAllPostOrderInternalNodes();
			geneTree.setAllBranches();
		}
		setMappingIndexesOfGeneTrees();
		
	//	
	}
	

private void setMappingIndexesOfDomainTrees() {
		int index = 0;
		ArrayList<Node> domainNodes = getAllDomainNodesByPostOrder();
		for(Node domainNode: domainNodes)
		{	
			domainNode.setMappingIndex(index);
			index++;
		}
	}
private void setMappingIndexesOfGeneTrees() {
	int index = 0;
	ArrayList<Node> geneNodes = getAllGeneNodesByPostOrder();
	for(Node geneNode: geneNodes)
	{	
		geneNode.setMappingIndex(index);
		index++;
	}
}


private void sampleReconciliationOfDomainToGeneApprox(DomainTreeRooted domainTree) {

		ArrayList<Node> allPreOrderDomainNodes = domainTree.getAllNodesByPreOrder();
	
		for (Node domainNode : allPreOrderDomainNodes) {
			RootedBinaryTreeMapping optimalMapping = null;
			ArrayList<RootedBinaryTreeMapping> optimalMappingsOfDomainNode = optimalMappings.get(domainNode);
			DomainNodeRooted domainNodeRooted = (DomainNodeRooted) domainNode;
			if (domainNode.isRoot()) {

				optimalMapping = getRandomSolution(optimalMappingsOfDomainNode);//getGreedyOptionSolution(optimalMappingsOfDomainNode);

			} else {
				DomainNodeRooted parentNode = (DomainNodeRooted) domainNodeRooted.getParent();

				RootedBinaryTreeMapping parentMapping = parentNode.getMapping();
				Node parentImage = parentMapping.getSecondNode();
				Node rightOfParentImage = parentMapping.getSecondNode().getRight();
				Node leftOfParentImage = parentMapping.getSecondNode().getLeft();

				int parentOption = parentNode.getOption();
				if (parentOption == 1) {
					optimalMapping = getDomainMappingFromNodes(domainNode, parentImage);
				} else if ((parentOption /100 == 2) || (parentOption /100== 15)) {
					if (parentNode.getLeft().equals(domainNodeRooted)) {
						optimalMapping = getRandomInMappingFromPair(domainNode, parentImage);
					} else {
						
						//inter species different gene family domain transfer
						 if(parentOption==222 || parentOption==1522) {
							optimalMapping = getOutMappingFromPairRandomly(domainNode, parentImage);
							parentNode.getMapping().setAdditiveTransferRecipient(optimalMapping.getSecondNode());
							
						}
					}
					
				} else if (parentOption/100 == 3 || parentOption/100 == 16) {
					if (parentNode.getRight().equals(domainNodeRooted)) {
						optimalMapping = getRandomInMappingFromPair(domainNode, parentImage);
						parentMapping.updateInTransferLossesWithBranch(optimalMapping.getSecondNode());
						
					} else {
						 if(parentOption%100==22) {
							optimalMapping = getOutMappingFromPairRandomly(domainNode, parentImage);
							parentNode.getMapping().setAdditiveTransferRecipient(optimalMapping.getSecondNode());							
						}
					}
					
				} else if (parentOption == 4) {
					if (parentNode.getLeft().equals(domainNodeRooted)) {
						optimalMapping = getRandomInMappingFromPair(domainNode, leftOfParentImage);

					} else {

						optimalMapping = getRandomInMappingFromPair(domainNode, rightOfParentImage);

					}
				
				} else if (parentOption == 5) {
					if (parentNode.getLeft().equals(domainNodeRooted)) {
						optimalMapping = getRandomInMappingFromPair(domainNode, rightOfParentImage);
					} else {
						optimalMapping = getRandomInMappingFromPair(domainNode, leftOfParentImage);
					}
				
				} else if (parentOption == 6) {
					if (parentNode.getLeft().equals(domainNodeRooted)) {
						optimalMapping = getDomainMappingFromNodes(domainNode, parentImage);
					} else {
						optimalMapping = getRandomInMappingFromPair(domainNode, rightOfParentImage);// inSearchWithLossTracking(optimalMappingsOfGeneNode,
						// rightOfParentImage);
					}
				} else if (parentOption == 7) {
					if (parentNode.getLeft().equals(domainNodeRooted)) {
						optimalMapping = getDomainMappingFromNodes(domainNode, parentImage);
					} else {
						optimalMapping = getGreedyInMappingFromPair(domainNode, leftOfParentImage);
					}
				} else if (parentOption == 8) {
					if (parentNode.getRight().equals(domainNodeRooted)) {
						optimalMapping = getDomainMappingFromNodes(domainNode, parentImage);
					
					} else {
						optimalMapping = getRandomInMappingFromPair(domainNode, rightOfParentImage);
					
					}
				} else if (parentOption == 9) {
					if (parentNode.getRight().equals(domainNodeRooted)) {
						optimalMapping = getDomainMappingFromNodes(domainNode, parentImage);
					
					} else {
						optimalMapping = getRandomInMappingFromPair(domainNode, leftOfParentImage);
					
					}
				} else if (parentOption == 10) {
					optimalMapping = getDomainMappingFromNodes(domainNode, parentImage);
				} else if (parentOption == 11) {
					if (parentNode.getLeft().equals(domainNodeRooted)) {
						optimalMapping = getRandomInMappingFromPair(domainNode, leftOfParentImage);
					} else {
						optimalMapping = getRandomInMappingFromPair(domainNode, rightOfParentImage);
					}
				
				} else if (parentOption == 12) {
					if (parentNode.getLeft().equals(domainNodeRooted)) {
						optimalMapping = getRandomInMappingFromPair(domainNode, rightOfParentImage);
					
					} else {
						optimalMapping = getRandomInMappingFromPair(domainNode, leftOfParentImage);
					
					}
					
				} else if (parentOption == 13) {
					optimalMapping = getRandomInMappingFromPair(domainNode, leftOfParentImage);
				} else if (parentOption == 14) {
					optimalMapping = getRandomInMappingFromPair(domainNode, rightOfParentImage);
				}
			}
	//		System.out.println(domainNodeRooted);
	//		System.out.println(optimalMapping);
			domainNodeRooted.setMappedGeneNode((GeneNodeRooted)optimalMapping.getSecondNode());
			setDomainOptionRandomly(domainNodeRooted, optimalMapping);
			domainNodeRooted.setMapping(optimalMapping);
			domainNodeRooted.setEventByOption();

		}

	}
	private void sampleGreedyOptionSolutionOfDomainTree(DomainTreeRooted domainTree) {

		ArrayList<Node> allPreOrderDomainNodes = domainTree.getAllNodesByPreOrder();
	
		for (Node domainNode : allPreOrderDomainNodes) {
			RootedBinaryTreeMapping optimalMapping = null;
			ArrayList<RootedBinaryTreeMapping> optimalMappingsOfDomainNode = optimalMappings.get(domainNode);
			DomainNodeRooted domainNodeRooted = (DomainNodeRooted) domainNode;
			if (domainNode.isRoot()) {

				optimalMapping = getRandomSolution(optimalMappingsOfDomainNode);//getGreedyOptionSolution(optimalMappingsOfDomainNode);

			} else {
				DomainNodeRooted parentNode = (DomainNodeRooted) domainNodeRooted.getParent();

				RootedBinaryTreeMapping parentMapping = parentNode.getMapping();
				Node parentImage = parentMapping.getSecondNode();
				Node rightOfParentImage = parentMapping.getSecondNode().getRight();
				Node leftOfParentImage = parentMapping.getSecondNode().getLeft();

				int parentOption = parentNode.getOption();
			//	System.out.println("Parent option: "+parentOption);
				if (parentOption == 1) {
					optimalMapping = getDomainMappingFromNodes(domainNode, parentImage);
				} else if ((parentOption /100 == 2) || (parentOption /100== 15)) {
					if (parentNode.getLeft().equals(domainNodeRooted)) {
						optimalMapping = getRandomInMappingFromPair(domainNode, parentImage);
						parentMapping.updateInTransferLossesWithBranch(optimalMapping.getSecondNode());
					} else {
						//intra species same gene family domain transfer
						if(parentOption==211 || parentOption==1511) {
							optimalMapping = getIntraSameOutMappingFromPair(domainNode, parentImage);
							parentNode.getMapping().setAdditiveTransferRecipient(optimalMapping.getSecondNode());
							
						}
						//intra species different gene family domain transfer
						else if(parentOption==212 || parentOption==1512) {
							optimalMapping = getIntraDifferentOutMappingFromPair(domainNode, parentImage);
							parentNode.getMapping().setAdditiveTransferRecipient(optimalMapping.getSecondNode());
							
						}
						//inter species same gene family domain transfer
						else if(parentOption==221 || parentOption==1521) {
							optimalMapping = getInterSameOutMappingFromPair(domainNode, parentImage);
							parentNode.getMapping().setAdditiveTransferRecipient(optimalMapping.getSecondNode());
							
						}
						//inter species different gene family domain transfer
						else if(parentOption==222 || parentOption==1522) {
							optimalMapping = getInterDifferentOutMappingFromPair(domainNode, parentImage);
							parentNode.getMapping().setAdditiveTransferRecipient(optimalMapping.getSecondNode());
							
						}
					//	optimalMapping = getGreedyOutMappingFromPair(domainNode, parentImage);
					//	parentNode.getMapping().setAdditiveTransferRecipient(optimalMapping.getSecondNode());
					}
					
				} else if (parentOption/100 == 3 || parentOption/100 == 16) {
					if (parentNode.getRight().equals(domainNodeRooted)) {
						optimalMapping = getRandomInMappingFromPair(domainNode, parentImage);
						parentMapping.updateInTransferLossesWithBranch(optimalMapping.getSecondNode());
						
					} else {
						//intra species same gene family domain transfer
						if(parentOption%100==11) {
							optimalMapping = getIntraSameOutMappingFromPair(domainNode, parentImage);
							parentNode.getMapping().setAdditiveTransferRecipient(optimalMapping.getSecondNode());
							
						}
						//intra species different gene family domain transfer
						else if(parentOption%100==12) {
							optimalMapping = getIntraDifferentOutMappingFromPair(domainNode, parentImage);
							parentNode.getMapping().setAdditiveTransferRecipient(optimalMapping.getSecondNode());
							
						}
						//inter species same gene family domain transfer
						else if(parentOption%100==21) {
							optimalMapping = getInterSameOutMappingFromPair(domainNode, parentImage);
							parentNode.getMapping().setAdditiveTransferRecipient(optimalMapping.getSecondNode());
							
						}
						//inter species different gene family domain transfer
						else if(parentOption%100==22) {
							optimalMapping = getInterDifferentOutMappingFromPair(domainNode, parentImage);
							parentNode.getMapping().setAdditiveTransferRecipient(optimalMapping.getSecondNode());
							
						}
					}
					
				} else if (parentOption == 4) {
					if (parentNode.getLeft().equals(domainNodeRooted)) {
						optimalMapping = getRandomInMappingFromPair(domainNode, leftOfParentImage);

					} else {

						optimalMapping = getRandomInMappingFromPair(domainNode, rightOfParentImage);

					}
					parentMapping.updateLossesWithBranch(optimalMapping.getSecondNode());

				} else if (parentOption == 5) {
					if (parentNode.getLeft().equals(domainNodeRooted)) {
						optimalMapping = getRandomInMappingFromPair(domainNode, rightOfParentImage);
					} else {
						optimalMapping = getRandomInMappingFromPair(domainNode, leftOfParentImage);
					}
					parentMapping.updateLossesWithBranch(optimalMapping.getSecondNode());

				} else if (parentOption == 6) {
					if (parentNode.getLeft().equals(domainNodeRooted)) {
						optimalMapping = getDomainMappingFromNodes(domainNode, parentImage);
						parentMapping.addLoss(parentImage.getLeft());
						parentMapping.addLostBranch(parentImage.getLeftBranch());
					} else {
						optimalMapping = getRandomInMappingFromPair(domainNode, rightOfParentImage);// inSearchWithLossTracking(optimalMappingsOfGeneNode,
						// rightOfParentImage);
						parentMapping.updateLossesWithBranch(optimalMapping.getSecondNode());
					}
				} else if (parentOption == 7) {
					if (parentNode.getLeft().equals(domainNodeRooted)) {
						optimalMapping = getDomainMappingFromNodes(domainNode, parentImage);
						parentMapping.addLoss(parentImage.getRight());
						parentMapping.addLostBranch(parentImage.getRightBranch());
					} else {
						optimalMapping = getRandomInMappingFromPair(domainNode, leftOfParentImage);
						parentMapping.updateLossesWithBranch(optimalMapping.getSecondNode());
					}
				} else if (parentOption == 8) {
					if (parentNode.getRight().equals(domainNodeRooted)) {
						optimalMapping = getDomainMappingFromNodes(domainNode, parentImage);
						parentMapping.addLoss(parentImage.getLeft());
						parentMapping.addLostBranch(parentImage.getLeftBranch());

					} else {
						optimalMapping = getRandomInMappingFromPair(domainNode, rightOfParentImage);
						parentMapping.updateLossesWithBranch(optimalMapping.getSecondNode());

					}
				} else if (parentOption == 9) {
					if (parentNode.getRight().equals(domainNodeRooted)) {
						optimalMapping = getDomainMappingFromNodes(domainNode, parentImage);
						parentMapping.addLoss(parentImage.getRight());
						parentMapping.addLostBranch(parentImage.getRightBranch());

					} else {
						optimalMapping = getRandomInMappingFromPair(domainNode, leftOfParentImage);
						parentMapping.updateLossesWithBranch(optimalMapping.getSecondNode());

					}
				} else if (parentOption == 10) {
					optimalMapping = getDomainMappingFromNodes(domainNode, parentImage);
				} else if (parentOption == 11) {
					if (parentNode.getLeft().equals(domainNodeRooted)) {
						optimalMapping = getRandomInMappingFromPair(domainNode, leftOfParentImage);
						parentMapping.addLoss(parentImage.getRight());
						parentMapping.addLostBranch(parentImage.getRightBranch());
					} else {
						optimalMapping = getRandomInMappingFromPair(domainNode, rightOfParentImage);
						parentMapping.addLoss(parentImage.getLeft());
						parentMapping.addLostBranch(parentImage.getLeftBranch());
					}
					parentMapping.updateLossesWithBranch(optimalMapping.getSecondNode());

				} else if (parentOption == 12) {
					if (parentNode.getLeft().equals(domainNodeRooted)) {
						optimalMapping = getRandomInMappingFromPair(domainNode, rightOfParentImage);
						parentMapping.addLoss(parentImage.getRight());
						parentMapping.addLostBranch(parentImage.getRightBranch());

					} else {
						optimalMapping = getRandomInMappingFromPair(domainNode, leftOfParentImage);
						parentMapping.addLoss(parentImage.getLeft());
						parentMapping.addLostBranch(parentImage.getLeftBranch());

					}
					parentMapping.updateLossesWithBranch(optimalMapping.getSecondNode());

				} else if (parentOption == 13) {
					optimalMapping = getRandomInMappingFromPair(domainNode, leftOfParentImage);
					parentMapping.addLoss(parentImage.getRight());
					parentMapping.addLostBranch(parentImage.getRightBranch());
					parentMapping.updateLossesWithBranch(optimalMapping.getSecondNode());
				} else if (parentOption == 14) {
					optimalMapping = getRandomInMappingFromPair(domainNode, rightOfParentImage);
					parentMapping.addLoss(parentImage.getLeft());
					parentMapping.addLostBranch(parentImage.getLeftBranch());
					parentMapping.updateLossesWithBranch(optimalMapping.getSecondNode());
				}
			}
		//	System.out.println(domainNodeRooted);
		//	System.out.println(optimalMapping);
		//	System.out.println(optimalMapping.getCost());
			domainNodeRooted.setMappedGeneNode((GeneNodeRooted)optimalMapping.getSecondNode());
			setDomainOptionRandomly(domainNodeRooted, optimalMapping);
			domainNodeRooted.setMapping(optimalMapping);
			domainNodeRooted.setEventByOption();

		}

	}
	private RootedBinaryTreeMapping getRandomSolution(ArrayList<RootedBinaryTreeMapping> optimalMappingsOfDomainNode) {
		Collections.shuffle(optimalMappingsOfDomainNode);
		
		return optimalMappingsOfDomainNode.get(0);
	}

	private DomainMapping getIntraSameOutMappingFromPair(Node domainNode, Node geneNode) {
		DomainMapping mapping = getDomainMappingFromNodes(domainNode, geneNode);

		return mapping.getIntraSameOutMappingRandomly(this);
	}
	private DomainMapping getIntraDifferentOutMappingFromPair(Node domainNode, Node geneNode) {
		DomainMapping mapping = getDomainMappingFromNodes(domainNode, geneNode);

		return mapping.getIntraDifferentOutMappingRandomly(this);
	}
	private DomainMapping getInterSameOutMappingFromPair(Node domainNode, Node geneNode) {
		DomainMapping mapping = getDomainMappingFromNodes(domainNode, geneNode);

		return mapping.getInterSameOutMappingRandomly(this);
	}
	private RootedBinaryTreeMapping getOutMappingFromPairRandomly(Node domainNode, Node geneNode) {
		DomainMapping mapping = getDomainMappingFromNodes(domainNode, geneNode);
		RootedBinaryTreeMapping outMapping = mapping.getOutMappingRandomly(this);
		return outMapping;
	}
	
	private DomainMapping getInterDifferentOutMappingFromPair(Node domainNode, Node geneNode) {
		DomainMapping mapping = getDomainMappingFromNodes(domainNode, geneNode);

		return mapping.getInterDifferentOutMappingRandomly(this);
	}

	private void setGeneOptionRandomly(GeneNodeRooted geneNode, RootedBinaryTreeMapping optimalMapping) {
		int greedyOption = optimalMapping.getRandomOption();
	//	System.out.println(greedyOption);
		geneNode.setOption(greedyOption);
	}
	
	private void setDomainOptionRandomly(DomainNodeRooted domainNode, RootedBinaryTreeMapping optimalMapping) {
		int greedyOption = optimalMapping.getRandomOption();
		domainNode.setOption(greedyOption);
	}
	private RootedBinaryTreeMapping getUniformOutMappingFromPair(Node geneNode, Node speciesNode) {
		RootedBinaryTreeMapping mapping = getGeneMappingFromNodes(geneNode, speciesNode);

		return mapping.getOutMappingUniformRandomly(this);
	}

	
private RootedBinaryTreeMapping getRandomOutMappingFromPair(Node firstNode,	Node secondNode) {
		
		RootedBinaryTreeMapping mapping =null;
		if(firstNode instanceof GeneNodeRooted) {
			mapping= getGeneMappingFromNodes(firstNode, secondNode);
		}
		else {
			mapping = getDomainMappingFromNodes(firstNode, secondNode);
		}
		
		return mapping.getOutMappingRandomly(this);

	}
	private RootedBinaryTreeMapping getGreedyInMappingFromPair(Node firstNode,	Node secondNode) {
			
		RootedBinaryTreeMapping mapping =null;
		if(firstNode instanceof GeneNodeRooted) {
			mapping= getGeneMappingFromNodes(firstNode, secondNode);
		}
		else {
			mapping = getDomainMappingFromNodes(firstNode, secondNode);
		}
		
		return mapping.getInMappingGreedily(this);

	}
	private RootedBinaryTreeMapping getRandomInMappingFromPair(Node firstNode,	Node secondNode) {
		
		RootedBinaryTreeMapping mapping =null;
		if(firstNode instanceof GeneNodeRooted) {
			mapping= getGeneMappingFromNodes(firstNode, secondNode);
		}
		else {
			mapping = getDomainMappingFromNodes(firstNode, secondNode);
		}
		
		return mapping.getInMappingRandomly(this);

	}
	public int binaryChoice() {
		Random random = new Random();
		return random.nextInt(2);
	}


	private void printOutputOfDGSReconciliation(PrintWriter writer) {
		//System.out.println("Trying to print");
		//writer.println("\n\n ------------ Reconciliation for Gene Tree 1 (rooted) -------------");
		
		writer.println("Domain Tree:");
		for(DomainTreeRooted domainTree: domainTrees)
			writer.println(domainTree.getNewickOutputFormat() + "\n");
		writer.println("Gene Trees:");
		for(GeneTreeRooted geneTree: geneTrees)
		{	writer.println(geneTree.getGeneTreeName()+":");
			writer.println(geneTree.getNewickOutputFormat() + "\n");
		}
		writer.println("Species Tree:");
		writer.println(speciesTree.getNewickOutputFormat() + "\n");
		
		ArrayList<Node> allPostOrderDomainNodes = getAllDomainNodesByPostOrder();
		int numberOfDomainDuplication = 0;
		//int numberOfDomainTransfer = 0;
		int numberOfGeneDuplication = 0;
		int numberOfGeneTransfer = 0;
		int numberOfGeneLoss = 0;
		int numberOfWithinGeneIntraSpeciesTransfer = 0;
		int numberOfWithinGeneInterSpeciesTransfer = 0;
		int numberOfAcrossGeneIntraSpeciesTransfer = 0;
		int numberOfAcrossGeneInterSpeciesTransfer = 0;
		int numberOfDomainLoss = 0;
		writer.println("\nReconciliation between domain tree with gene trees:");
		for (Node node : allPostOrderDomainNodes) {
			DomainNodeRooted domainNode = (DomainNodeRooted) node;
			if (domainNode.isLeaf()) {
				writer.println(domainNode.getFullName() + ": " + "Leaf Node");
				continue;
			}
			// writer.println("Internal node name: "+node.getInternalNodeName());
			String mappedNodeName = "";
			if(domainNode.getMappedGeneNode().isLeaf())
				mappedNodeName = domainNode.getMappedGeneNode().getFullName();
			else
				mappedNodeName = domainNode.getMappedGeneNode().getInternalNodeName();
			writer.print(domainNode.getInternalNodeName() + " = " + domainNode.getLCAFormat() + ": " + domainNode.getEvent()
					+ ", Mapping --> " + mappedNodeName);

			if (domainNode.getEvent().contains("Transfer")) {
				String recipientName = "";
				if(domainNode.getMapping().getAdditiveTransferRecipient().isLeaf()) {
					recipientName = ((GeneNodeRooted) domainNode.getMapping().getAdditiveTransferRecipient()).getFullName();
				}
				else
					recipientName =  ((GeneNodeRooted) domainNode.getMapping().getAdditiveTransferRecipient()).getInternalNodeName();
				writer.print(", Recipient --> " + recipientName);
				String event = domainNode.getEvent();
				if (event.contains("Within-gene-family Intraspecies")) {
					numberOfWithinGeneIntraSpeciesTransfer++;
				}
				else if (event.contains("Within-gene-family Interspecies")) {
					numberOfWithinGeneInterSpeciesTransfer++;
				}
				else if (event.contains("Across-gene-family Intraspecies")) {
					numberOfAcrossGeneIntraSpeciesTransfer++;
				}
				else {
					numberOfAcrossGeneInterSpeciesTransfer++;
				}
			//	numberOfDomainTransfer++;
		
			}
			if (domainNode.getEvent().equals("Duplication"))
				numberOfDomainDuplication++;

			if (domainNode.getMapping().getLosses().size() > 0) {
				numberOfDomainLoss += domainNode.getMapping().getLostEdges().size();
			}
			writer.println();

		}
		writer.println("\nReconciliation between gene trees with species trees:");
		
	
		for(GeneTreeRooted geneTree: geneTrees) {
			if(!geneTree.isUsed())
				continue;
			writer.println("Gene Tree: "+(geneTree.getID()+1));
		for(Node node: geneTree.getAllNodesByPostOrder()) {
			GeneNodeRooted geneNode = (GeneNodeRooted) node;
			if(geneNode.isLeaf()) {
				writer.println(geneNode.getFullName()+": "+"Leaf Node");
				continue;
			} 
			writer.print(geneNode.getInternalNodeName() + " = " + geneNode.getLCAFormat() + ": Gene " + geneNode.getEvent()
			+ ", Mapping --> " + geneNode.getMappedSpeciesNode().getName());

			if (geneNode.getEvent().contains("Transfer") || geneNode.getEvent().contains("TRANSFER") ) {
				writer.print(", Recipient --> " + geneNode.getMapping().getAdditiveTransferRecipient());
				numberOfGeneTransfer++;
			}
			if (geneNode.getEvent().equals("Duplication") || geneNode.getEvent().contains("DUPLICATION") )
				numberOfGeneDuplication++;

			if (geneNode.getMapping().getLosses().size() > 0) {
				numberOfGeneLoss += geneNode.getMapping().getLostEdges().size();
			}
			writer.println();
			
		}
		numberOfGeneLoss = getTotalGeneLoss();
		writer.println();
		}
		
	
		numberOfDomainLoss = (reconciliationCost - geneSpeciesReconciliationCost)- numberOfDomainDuplication*DOMAIN_DUPLICATION
				-numberOfAcrossGeneInterSpeciesTransfer*INTER_SPECIES_DOMAIN_TRANSFER_DIFFERENT- numberOfWithinGeneInterSpeciesTransfer*INTER_SPECIES_DOMAIN_TRANSFER_SAME
				-numberOfAcrossGeneIntraSpeciesTransfer*INTRA_SPECIES_DOMAIN_TRANSFER_DIFFERENT - numberOfWithinGeneIntraSpeciesTransfer*INTRA_SPECIES_DOMAIN_TRANSFER_SAME;
		writer.print("\nMinimal Gen-DGS reconciliation cost: " + reconciliationCost);
		writer.print("\nDomain-Gene reconciliation cost: " + (reconciliationCost-geneSpeciesReconciliationCost));
		writer.print(" (Domain Duplications: " + numberOfDomainDuplication + ", within-gene-family intra-species domain transfers: " + numberOfWithinGeneIntraSpeciesTransfer + ", within-gene-family inter-species domain transfers: "+ numberOfWithinGeneInterSpeciesTransfer
				+", across-gene-family intra-species domain transfers: "+ numberOfAcrossGeneIntraSpeciesTransfer
		+", across-gene-family inter-species domain transfers: "+numberOfAcrossGeneInterSpeciesTransfer+
				", Domain losses: "
				+ numberOfDomainLoss + ")");
		numberOfGeneLoss= geneSpeciesReconciliationCost - numberOfGeneTransfer*GENE_TRANSFER - numberOfGeneDuplication*GENE_DUPLICATION;
		writer.print("\nGene-Species reconciliation cost: " + geneSpeciesReconciliationCost);
		writer.println(" (Gene Duplications: " + numberOfGeneDuplication + ", Gene Transfers: " + numberOfGeneTransfer + ", Gene Losses: "
				+ numberOfGeneLoss + ")");
		writer.close();
	}
	private int getTotalGeneLoss() {
		int numberOfGeneLoss = 0;
		for(GeneTreeRooted geneTree: geneTrees) {
			numberOfGeneLoss+=geneTree.getNumberOfLoss();
		}
		return numberOfGeneLoss;
	}
	
	private void handleIntToIntGeneMapping(RootedBinaryTreeMapping tentativeMapping, Node geneNode,
			Node speciesNode, RootedBinaryTreeMapping leftToParentMapping,
			RootedBinaryTreeMapping rightToParentMapping) {

		// get left child s' and right child s" of species node
		Node leftSpeciesChild = speciesNode.getLeft();
		Node rightSpeciesChild = speciesNode.getRight();
		// get left child g' and right child g" of gene node
		Node leftGeneChild = geneNode.getLeft();
		Node rightGeneChild = geneNode.getRight();
		// let the gene node to left child of species node mapping be (g, s')
		RootedBinaryTreeMapping parentToRightMapping = getGeneMappingFromNodes(geneNode, rightSpeciesChild);

		// let the gene node to right child of species node mapping be (g, s")
		RootedBinaryTreeMapping parentToLeftMapping = getGeneMappingFromNodes(geneNode, leftSpeciesChild);
		// let the left child of gene node to left child of species node mapping be (g',
		// s')
		RootedBinaryTreeMapping leftToLeftMapping = getGeneMappingFromNodes(leftGeneChild, leftSpeciesChild);

		// let the left child of gene node to right child of species node mapping be
		// (g', s")
		RootedBinaryTreeMapping leftToRightMapping = getGeneMappingFromNodes(leftGeneChild, rightSpeciesChild);

		// let the right child of gene node to left child of species node mapping be
		// (g", s')
		RootedBinaryTreeMapping rightToLeftMapping = getGeneMappingFromNodes(rightGeneChild, leftSpeciesChild);

		// let the right child of gene node to right child of species node mapping be
		// (g", s")
		RootedBinaryTreeMapping rightToRightMapping = getGeneMappingFromNodes(rightGeneChild, rightSpeciesChild);

		// speciationCost = min{in(g',s') + in (g",s"), in(g",s') + in (g', s")}
		int speciationCost = Math.min(leftToLeftMapping.getInCost() + rightToRightMapping.getInCost(),
				leftToRightMapping.getInCost() + rightToLeftMapping.getInCost());
		tentativeMapping.setSpeciationCost(speciationCost);

		// save all duplication option costs in an array

		int[] duplicationOptions = getDuplicationOptions(leftToParentMapping, rightToParentMapping, leftToLeftMapping,
				leftToRightMapping, rightToLeftMapping, rightToRightMapping);

		// Set duplicationCost = DUPLICATION + min(duplicationOptions)
		int duplicationCost = GENE_DUPLICATION + getMinimum(duplicationOptions);
		tentativeMapping.setDuplicationCost(duplicationCost);

		// s is not root
		if (!speciesNode.isRoot()) {

			// transferCost(g,s) = TRANSFER + min(in(g',s) + out(g",s), in(g",s) + out(g's))
			int transferCost = GENE_TRANSFER + Math.min(leftToParentMapping.getInCost() + rightToParentMapping.getOutCost(),
					leftToParentMapping.getOutCost() + rightToParentMapping.getInCost());
			tentativeMapping.setTransferCost(transferCost);
		}

		// c(g,s) = min(speciationCost, duplicationCost, transferCost)
		tentativeMapping.setCost(getMinimumValue(tentativeMapping.getSpeciationCost(),
				tentativeMapping.getDuplicationCost(), tentativeMapping.getTransferCost()));

		int previousInCost = tentativeMapping.getInCost();
		// in(g,s) = min{c(g,s), in(g,s') + LOSS, in(g, s") + LOSS}
		tentativeMapping.setInCost(getMinimumValue(tentativeMapping.getCost(), parentToLeftMapping.getInCost() + GENE_LOSS,
				parentToRightMapping.getInCost() + GENE_LOSS));

		updateInGeneNodeForMultipleOptima(tentativeMapping, previousInCost);

		int previousInAltCost = tentativeMapping.getInAltCost();
		// inAlt(g,s) = min{c(g,s), inAlt(g,s'), inAlt(g, s")}
		tentativeMapping.setInAltCost(getMinimumValue(tentativeMapping.getCost(), parentToLeftMapping.getInAltCost(),
				parentToRightMapping.getInAltCost()));

		updateInAltNodeForGeneToSpecies(tentativeMapping, leftSpeciesChild, rightSpeciesChild, previousInAltCost);

	}


	private void updateInAltNodeForGeneToSpecies(RootedBinaryTreeMapping tentativeMapping, Node leftSpeciesChild,
			Node rightSpeciesChild, int previousInAltCost) {


		if (tentativeMapping.getInAltCost() < previousInAltCost) {
			tentativeMapping.getInAltSpeciesNodes().clear();
		}
		Node geneNode = tentativeMapping.getFirstNode();
		// get left child s' and right child s" of species node

		// let the gene node to left child of species node mapping be (g, s')
		RootedBinaryTreeMapping parentToRightMapping = getGeneMappingFromNodes(geneNode, rightSpeciesChild);

		// let the gene node to right child of species node mapping be (g, s")
		RootedBinaryTreeMapping parentToLeftMapping = getGeneMappingFromNodes(geneNode, leftSpeciesChild);
		if (tentativeMapping.getInAltCost() == parentToRightMapping.getInAltCost()) {
			for (Node inAltNode : parentToRightMapping.getInAltSpeciesNodes()) {
				if (!tentativeMapping.getInAltSpeciesNodes().contains(inAltNode))
					tentativeMapping.addInAltSpeciesNode(inAltNode);
			}
		}
		if (tentativeMapping.getInAltCost() == parentToLeftMapping.getInAltCost()) {
			for (Node inAltNode : parentToLeftMapping.getInAltSpeciesNodes())

			{
				if (!tentativeMapping.getInAltSpeciesNodes().contains(inAltNode))
					tentativeMapping.addInAltSpeciesNode(inAltNode);
			}
		}
		if (tentativeMapping.getInAltCost() == tentativeMapping.getCost()) {
			if (!tentativeMapping.getInAltSpeciesNodes().contains(tentativeMapping.getSecondNode()))
				tentativeMapping.addInAltSpeciesNode(tentativeMapping.getSecondNode());
		}

	
	}

	private void updateInGeneNodeForMultipleOptima(RootedBinaryTreeMapping tentativeMapping, int previousInCost) {
		if (tentativeMapping.getInCost() < previousInCost) {
			tentativeMapping.getInSpeciesNodes().clear();
		}
		Node speciesNode = tentativeMapping.getSecondNode();
		Node geneNode = tentativeMapping.getFirstNode();
		// get left child s' and right child s" of species node
		Node leftSpeciesChild = speciesNode.getLeft();
		Node rightSpeciesChild = speciesNode.getRight();
		// let the gene node to left child of species node mapping be (g, s')
		RootedBinaryTreeMapping parentToRightMapping = getGeneMappingFromNodes(geneNode, rightSpeciesChild);

		// let the gene node to right child of species node mapping be (g, s")
		RootedBinaryTreeMapping parentToLeftMapping = getGeneMappingFromNodes(geneNode, leftSpeciesChild);
		if (tentativeMapping.getInCost() == parentToLeftMapping.getInCost() + GENE_LOSS) {
			for (Node inNode : parentToLeftMapping.getInSpeciesNodes()) {
				if (!tentativeMapping.getInSpeciesNodes().contains(inNode)) {
					tentativeMapping.addInSpeciesNode(inNode);
				}
			}

		}
		if (tentativeMapping.getInCost() == parentToRightMapping.getInCost() + GENE_LOSS) {
			for (Node inNode : parentToRightMapping.getInSpeciesNodes()) {
				if (!tentativeMapping.getInSpeciesNodes().contains(inNode)) {
					tentativeMapping.addInSpeciesNode(inNode);
				}
			}
		}
		if (tentativeMapping.getInCost() == tentativeMapping.getCost()) {
			if (!tentativeMapping.getInSpeciesNodes().contains(speciesNode))
				tentativeMapping.addInSpeciesNode(speciesNode);
		}

	
	}

	private void handleIntToIntDomainMappingApprox(DomainMapping tentativeMapping, Node domainNode,
			Node geneNode, DomainMapping leftToParentMapping,
			DomainMapping rightToParentMapping) {

		// get left child s' and right child s" of gene node
		Node leftGeneChild = geneNode.getLeft();
		Node rightGeneChild = geneNode.getRight();
		// get left child g' and right child g" of gene node
		Node leftDomainChild = domainNode.getLeft();
		Node rightDomainChild = domainNode.getRight();
		// let the gene node to left child of species node mapping be (g, s')
		RootedBinaryTreeMapping parentToRightMapping = getDomainMappingFromNodes(domainNode, rightGeneChild);

		// let the gene node to right child of species node mapping be (g, s")
		RootedBinaryTreeMapping parentToLeftMapping = getDomainMappingFromNodes(domainNode, leftGeneChild);
		// let the left child of gene node to left child of species node mapping be (g',
		// s')
		RootedBinaryTreeMapping leftToLeftMapping = getDomainMappingFromNodes(leftDomainChild, leftGeneChild);

		// let the left child of gene node to right child of species node mapping be
		// (g', s")
		RootedBinaryTreeMapping leftToRightMapping = getDomainMappingFromNodes(leftDomainChild, rightGeneChild);

		// let the right child of gene node to left child of species node mapping be
		// (g", s')
		RootedBinaryTreeMapping rightToLeftMapping = getDomainMappingFromNodes(rightDomainChild, leftGeneChild);

		// let the right child of gene node to right child of species node mapping be
		// (g", s")
		RootedBinaryTreeMapping rightToRightMapping = getDomainMappingFromNodes(rightDomainChild, rightGeneChild);

		// speciationCost = min{in(g',s') + in (g",s"), in(g",s') + in (g', s")}
		int speciationCost = Math.min(leftToLeftMapping.getInCost() + rightToRightMapping.getInCost(),
				leftToRightMapping.getInCost() + rightToLeftMapping.getInCost());
		tentativeMapping.setSpeciationCost(speciationCost);

		// save all duplication option costs in an array

		int[] duplicationOptions = getDuplicationOptions(leftToParentMapping, rightToParentMapping, leftToLeftMapping,
				leftToRightMapping, rightToLeftMapping, rightToRightMapping);

		// Set duplicationCost = DUPLICATION + min(duplicationOptions)
		int duplicationCost = DOMAIN_DUPLICATION + getMinimum(duplicationOptions);
		tentativeMapping.setDuplicationCost(duplicationCost);

		// s is not root
		if (!geneNode.isRoot()) 
		{
			// transferCost(g,s) = TRANSFER + min(in(g',s) + out(g",s), in(g",s) + out(g's))						
			int transferCost =INTER_SPECIES_DOMAIN_TRANSFER_DIFFERENT +  Math.min(
					leftToParentMapping.getInCost() + rightToParentMapping.getOutCost(),
					leftToParentMapping.getOutCost() + rightToParentMapping.getInCost());
			tentativeMapping.setTransferCost(transferCost);

		}

		// c(g,s) = min(speciationCost, duplicationCost, transferCost)
		tentativeMapping.setCost(getMinimumValue(tentativeMapping.getSpeciationCost(),
				tentativeMapping.getDuplicationCost(), tentativeMapping.getTransferCost()));

		int previousInCost = tentativeMapping.getInCost();
		// in(g,s) = min{c(g,s), in(g,s') + LOSS, in(g, s") + LOSS}
		tentativeMapping.setInCost(getMinimumValue(tentativeMapping.getCost(), parentToLeftMapping.getInCost() + DOMAIN_LOSS,
				parentToRightMapping.getInCost() + DOMAIN_LOSS));

		updateInNodeForMultipleOptima(tentativeMapping, previousInCost);

		int previousInAltCost = tentativeMapping.getInAltCost();
		// inAlt(g,s) = min{c(g,s), inAlt(g,s'), inAlt(g, s")}
		tentativeMapping.setInAltCost(getMinimumValue(tentativeMapping.getCost(), parentToLeftMapping.getInAltCost(),
				parentToRightMapping.getInAltCost()));

		updateInAltNodeForMultipleOptima(tentativeMapping, leftGeneChild, rightGeneChild, previousInAltCost);

	}

	private void handleIntToIntDomainMapping(DomainMapping tentativeMapping, Node domainNode,
			Node geneNode, DomainMapping leftToParentMapping,
			DomainMapping rightToParentMapping) {

		// get left child s' and right child s" of gene node
		Node leftGeneChild = geneNode.getLeft();
		Node rightGeneChild = geneNode.getRight();
		// get left child g' and right child g" of gene node
		Node leftDomainChild = domainNode.getLeft();
		Node rightDomainChild = domainNode.getRight();
		// let the gene node to left child of species node mapping be (g, s')
		RootedBinaryTreeMapping parentToRightMapping = getDomainMappingFromNodes(domainNode, rightGeneChild);

		// let the gene node to right child of species node mapping be (g, s")
		RootedBinaryTreeMapping parentToLeftMapping = getDomainMappingFromNodes(domainNode, leftGeneChild);
		// let the left child of gene node to left child of species node mapping be (g',
		// s')
		RootedBinaryTreeMapping leftToLeftMapping = getDomainMappingFromNodes(leftDomainChild, leftGeneChild);

		// let the left child of gene node to right child of species node mapping be
		// (g', s")
		RootedBinaryTreeMapping leftToRightMapping = getDomainMappingFromNodes(leftDomainChild, rightGeneChild);

		// let the right child of gene node to left child of species node mapping be
		// (g", s')
		RootedBinaryTreeMapping rightToLeftMapping = getDomainMappingFromNodes(rightDomainChild, leftGeneChild);

		// let the right child of gene node to right child of species node mapping be
		// (g", s")
		RootedBinaryTreeMapping rightToRightMapping = getDomainMappingFromNodes(rightDomainChild, rightGeneChild);

		// speciationCost = min{in(g',s') + in (g",s"), in(g",s') + in (g', s")}
		int speciationCost = Math.min(leftToLeftMapping.getInCost() + rightToRightMapping.getInCost(),
				leftToRightMapping.getInCost() + rightToLeftMapping.getInCost());
		tentativeMapping.setSpeciationCost(speciationCost);

		// save all duplication option costs in an array

		int[] duplicationOptions = getDuplicationOptions(leftToParentMapping, rightToParentMapping, leftToLeftMapping,
				leftToRightMapping, rightToLeftMapping, rightToRightMapping);

		// Set duplicationCost = DUPLICATION + min(duplicationOptions)
		int duplicationCost = DOMAIN_DUPLICATION + getMinimum(duplicationOptions);
		tentativeMapping.setDuplicationCost(duplicationCost);

		// s is not root
		if (!geneNode.isRoot()) 
		{
			// transferCost(g,s) = TRANSFER + min(in(g',s) + out(g",s), in(g",s) + out(g's))
			// check all four combinations of domain transfer costs
			int intraSpeciesSameTreeTransferCost = INTRA_SPECIES_DOMAIN_TRANSFER_SAME +  Math.min(
					leftToParentMapping.getInCost() + rightToParentMapping.getIntraSameTreeOutCost(),
					leftToParentMapping.getIntraSameTreeOutCost() + rightToParentMapping.getInCost());
			int intraSpeciesDifferentTreeTransferCost = INTRA_SPECIES_DOMAIN_TRANSFER_DIFFERENT +  Math.min(
					leftToParentMapping.getInCost() + rightToParentMapping.getIntraDifferentTreeOutCost(),
					leftToParentMapping.getIntraDifferentTreeOutCost() + rightToParentMapping.getInCost());
			int interSpeciesSameTreeTransferCost = INTER_SPECIES_DOMAIN_TRANSFER_SAME +  Math.min(
					leftToParentMapping.getInCost() + rightToParentMapping.getInterSameTreeOutCost(),
					leftToParentMapping.getInterSameTreeOutCost() + rightToParentMapping.getInCost());
			int interSpeciesDifferentTreeTransferCost = INTER_SPECIES_DOMAIN_TRANSFER_DIFFERENT +  Math.min(
					leftToParentMapping.getInCost() + rightToParentMapping.getInterDifferentTreeOutCost(),
					leftToParentMapping.getInterDifferentTreeOutCost() + rightToParentMapping.getInCost());
			
			int transferCost = getMinimumOfFourTransferCost(intraSpeciesSameTreeTransferCost, 
					intraSpeciesDifferentTreeTransferCost, interSpeciesSameTreeTransferCost, 
					interSpeciesDifferentTreeTransferCost);
			tentativeMapping.setTransferCost(transferCost);

		}

		// c(g,s) = min(speciationCost, duplicationCost, transferCost)
		tentativeMapping.setCost(getMinimumValue(tentativeMapping.getSpeciationCost(),
				tentativeMapping.getDuplicationCost(), tentativeMapping.getTransferCost()));

		int previousInCost = tentativeMapping.getInCost();
		// in(g,s) = min{c(g,s), in(g,s') + LOSS, in(g, s") + LOSS}
		tentativeMapping.setInCost(getMinimumValue(tentativeMapping.getCost(), parentToLeftMapping.getInCost() + DOMAIN_LOSS,
				parentToRightMapping.getInCost() + DOMAIN_LOSS));

		updateInNodeForMultipleOptima(tentativeMapping, previousInCost);

		int previousInAltCost = tentativeMapping.getInAltCost();
		// inAlt(g,s) = min{c(g,s), inAlt(g,s'), inAlt(g, s")}
		tentativeMapping.setInAltCost(getMinimumValue(tentativeMapping.getCost(), parentToLeftMapping.getInAltCost(),
				parentToRightMapping.getInAltCost()));

		updateInAltNodeForMultipleOptima(tentativeMapping, leftGeneChild, rightGeneChild, previousInAltCost);

	}

	private void updateInAltNodeForMultipleOptima(RootedBinaryTreeMapping tentativeMapping, Node leftSpeciesChild,
			Node rightSpeciesChild, int previousCost) {

		if (tentativeMapping.getInAltCost() < previousCost) {
			tentativeMapping.getInAltSpeciesNodes().clear();
		}
		if(tentativeMapping.getInAltCost()>=Integer.MAX_VALUE/10)
			return;
		Node geneNode = tentativeMapping.getFirstNode();
		// get left child s' and right child s" of species node

		// let the gene node to left child of species node mapping be (g, s')
		RootedBinaryTreeMapping parentToRightMapping = getDomainMappingFromNodes(geneNode, rightSpeciesChild);

		// let the gene node to right child of species node mapping be (g, s")
		RootedBinaryTreeMapping parentToLeftMapping = getDomainMappingFromNodes(geneNode, leftSpeciesChild);
		if (tentativeMapping.getInAltCost() == parentToRightMapping.getInAltCost() ) {
			for (Node inAltNode : parentToRightMapping.getInAltSpeciesNodes()) {
				if (!tentativeMapping.getInAltSpeciesNodes().contains(inAltNode))
					tentativeMapping.addInAltSpeciesNode(inAltNode);
			}
		}
		if (tentativeMapping.getInAltCost() == parentToLeftMapping.getInAltCost() ) {
			for (Node inAltNode : parentToLeftMapping.getInAltSpeciesNodes())

			{
				if (!tentativeMapping.getInAltSpeciesNodes().contains(inAltNode))
					tentativeMapping.addInAltSpeciesNode(inAltNode);
			}
		}
		if (tentativeMapping.getInAltCost() == tentativeMapping.getCost()) {
			if (!tentativeMapping.getInAltSpeciesNodes().contains(tentativeMapping.getSecondNode()))
				tentativeMapping.addInAltSpeciesNode(tentativeMapping.getSecondNode());
		}

	}

	private void updateInNodeForMultipleOptima(RootedBinaryTreeMapping tentativeMapping, int previousInCost) {
		
		if(tentativeMapping.getInCost()>=Integer.MAX_VALUE/10)
			return;
		
		if (tentativeMapping.getInCost() < previousInCost) {
			tentativeMapping.getInSpeciesNodes().clear();
		}
		
		Node speciesNode = tentativeMapping.getSecondNode();
		Node geneNode = tentativeMapping.getFirstNode();
		// get left child s' and right child s" of species node
		Node leftSpeciesChild = speciesNode.getLeft();
		Node rightSpeciesChild = speciesNode.getRight();
		// let the gene node to left child of species node mapping be (g, s')
		RootedBinaryTreeMapping parentToRightMapping = getDomainMappingFromNodes(geneNode, rightSpeciesChild);

		// let the gene node to right child of species node mapping be (g, s")
		RootedBinaryTreeMapping parentToLeftMapping = getDomainMappingFromNodes(geneNode, leftSpeciesChild);
		if (tentativeMapping.getInCost() == parentToLeftMapping.getInCost() + DOMAIN_LOSS) {
			for (Node inNode : parentToLeftMapping.getInSpeciesNodes()) {
				if (!tentativeMapping.getInSpeciesNodes().contains(inNode)) {
					tentativeMapping.addInSpeciesNode(inNode);
				}
			}

		}
		if (tentativeMapping.getInCost() == parentToRightMapping.getInCost() + DOMAIN_LOSS) {
			for (Node inNode : parentToRightMapping.getInSpeciesNodes()) {
				if (!tentativeMapping.getInSpeciesNodes().contains(inNode)) {
					tentativeMapping.addInSpeciesNode(inNode);
				}
			}
		}
		if (tentativeMapping.getInCost() == tentativeMapping.getCost()) {
			if (!tentativeMapping.getInSpeciesNodes().contains(speciesNode))
				tentativeMapping.addInSpeciesNode(speciesNode);
		}

	}

	private void updateOutSpeciesNodeForMultipleOptima(RootedBinaryTreeMapping tentativeMapping,
			RootedBinaryTreeMapping parentToLeftMapping, RootedBinaryTreeMapping parentToRightMapping) {

		if (parentToLeftMapping.getOutCost() == tentativeMapping.getOutCost()) {
			for (Node outNode : tentativeMapping.getOutSpeciesNodes()) {
				if (!parentToLeftMapping.getOutSpeciesNodes().contains(outNode))
					parentToLeftMapping.addOutSpeciesNode(outNode);
			}
		}
		if (parentToLeftMapping.getOutCost() == parentToRightMapping.getInAltCost()) {
			for (Node inAltNode : parentToRightMapping.getInAltSpeciesNodes()) {
				if (!parentToLeftMapping.getOutSpeciesNodes().contains(inAltNode))
					parentToLeftMapping.addOutSpeciesNode(inAltNode);
			}
		}
		if (parentToRightMapping.getOutCost() == tentativeMapping.getOutCost()) {
			for (Node outNode : tentativeMapping.getOutSpeciesNodes()) {
				if (!parentToRightMapping.getOutSpeciesNodes().contains(outNode))
					parentToRightMapping.addOutSpeciesNode(outNode);
			}
		}
		if (parentToRightMapping.getOutCost() == parentToLeftMapping.getInAltCost()) {
			for (Node inAltNode : parentToLeftMapping.getInAltSpeciesNodes()) {
				if (!parentToRightMapping.getOutSpeciesNodes().contains(inAltNode))
					parentToRightMapping.addOutSpeciesNode(inAltNode);
			}
		}

	}
	private void setOptionsForIntToIntDomainMappingApprox(Node domainNode, Node geneNode, Node leftDomainChild,
			Node rightDomainChild, Node leftGeneChild, Node rightGeneChild,
			DomainMapping tentativeMapping) {

		// let the left child to species node mapping be (g', s)
		DomainMapping leftToParentMapping = getDomainMappingFromNodes(leftDomainChild, geneNode);

		// let the right child to species node mapping be (g", s)
		DomainMapping rightToParentMapping = getDomainMappingFromNodes(rightDomainChild, geneNode);

		// let the left child of gene node to left child of species node mapping be (g',
		// s')
		DomainMapping leftToLeftMapping = getDomainMappingFromNodes(leftDomainChild, leftGeneChild);

		// let the left child of gene node to right child of species node mapping be
		// (g', s")
		DomainMapping leftToRightMapping = getDomainMappingFromNodes(leftDomainChild, rightGeneChild);

		// let the right child of gene node to left child of species node mapping be
		// (g", s')
		DomainMapping rightToLeftMapping = getDomainMappingFromNodes(rightDomainChild, leftGeneChild);

		// let the right child of gene node to right child of species node mapping be
		// (g", s")
		DomainMapping rightToRightMapping = getDomainMappingFromNodes(rightDomainChild, rightGeneChild);

		// speciationCost = min{in(g',s') + in (g",s"), in(g",s') + in (g', s")}
		int speciationCost = Math.min(leftToLeftMapping.getInCost() + rightToRightMapping.getInCost(),
				leftToRightMapping.getInCost() + rightToLeftMapping.getInCost());
		if (tentativeMapping.getCost() >= Integer.MAX_VALUE / 10)
			return;
		if (tentativeMapping.getCost() == speciationCost) {
			if (tentativeMapping.getCost() == leftToLeftMapping.getInCost() + rightToRightMapping.getInCost()) {
				tentativeMapping.getOptions().add(4);
				tentativeMapping.addWeightByInAndIn(leftToLeftMapping, rightToRightMapping, this);
			}
			if (tentativeMapping.getCost() == leftToRightMapping.getInCost() + rightToLeftMapping.getInCost()) {
				tentativeMapping.getOptions().add(5);
				tentativeMapping.addWeightByInAndIn(leftToRightMapping, rightToLeftMapping, this);
			}
		}

		int[] duplicationOptions = getDuplicationOptions(leftToParentMapping, rightToParentMapping, leftToLeftMapping,
				leftToRightMapping, rightToLeftMapping, rightToRightMapping);

		// Set duplicationCost = DUPLICATION + min(duplicationOptions)
		int duplicationCost = DOMAIN_DUPLICATION + getMinimum(duplicationOptions);
		if (tentativeMapping.getCost() == duplicationCost) {
			if (tentativeMapping.getCost() - DOMAIN_DUPLICATION == leftToParentMapping.getCost()
					+ rightToRightMapping.getInCost() + DOMAIN_LOSS) {
				tentativeMapping.getOptions().add(6);
				tentativeMapping.addWeightByNormalAndIn(leftToParentMapping, rightToRightMapping, this);
			}
			if (tentativeMapping.getCost() - DOMAIN_DUPLICATION == leftToParentMapping.getCost()
					+ rightToLeftMapping.getInCost() + DOMAIN_LOSS) {
				tentativeMapping.getOptions().add(7);
				tentativeMapping.addWeightByNormalAndIn(leftToParentMapping, rightToLeftMapping, this);

			}
			if (tentativeMapping.getCost() - DOMAIN_DUPLICATION == rightToParentMapping.getCost()
					+ leftToRightMapping.getInCost() + DOMAIN_LOSS) {
				tentativeMapping.getOptions().add(8);
				tentativeMapping.addWeightByNormalAndIn(rightToParentMapping, leftToRightMapping, this);
			}
			if (tentativeMapping.getCost() - DOMAIN_DUPLICATION == rightToParentMapping.getCost()
					+ leftToLeftMapping.getInCost() + DOMAIN_LOSS) {
				tentativeMapping.getOptions().add(9);
				tentativeMapping.addWeightByNormalAndIn(rightToParentMapping, leftToLeftMapping, this);

			}
			if (tentativeMapping.getCost() - DOMAIN_DUPLICATION == rightToParentMapping.getCost()
					+ leftToParentMapping.getCost()) {
				tentativeMapping.getOptions().add(10);
				tentativeMapping.addWeightByNormalAndNormal(rightToParentMapping, leftToParentMapping);

			}
			if (tentativeMapping.getCost() - DOMAIN_DUPLICATION == leftToLeftMapping.getInCost()
					+ rightToRightMapping.getInCost() + 2 * DOMAIN_LOSS) {
				tentativeMapping.getOptions().add(11);
				tentativeMapping.addWeightByInAndIn(leftToLeftMapping, rightToRightMapping, this);
			}
			if (tentativeMapping.getCost() - DOMAIN_DUPLICATION == leftToRightMapping.getInCost()
					+ rightToLeftMapping.getInCost() + 2 * DOMAIN_LOSS) {
				tentativeMapping.getOptions().add(12);
				tentativeMapping.addWeightByInAndIn(leftToRightMapping, rightToLeftMapping, this);

			}
			if (tentativeMapping.getCost() - DOMAIN_DUPLICATION == leftToLeftMapping.getInCost()
					+ rightToLeftMapping.getInCost() + 2 * DOMAIN_LOSS) {
				tentativeMapping.getOptions().add(13);
				tentativeMapping.addWeightByInAndIn(leftToLeftMapping, rightToLeftMapping, this);

			}
			if (tentativeMapping.getCost() - DOMAIN_DUPLICATION == leftToRightMapping.getInCost()
					+ rightToRightMapping.getInCost() + 2 * DOMAIN_LOSS) {
				tentativeMapping.getOptions().add(14);
				tentativeMapping.addWeightByInAndIn(leftToRightMapping, rightToRightMapping, this);

			}
		}
		if (!geneNode.isRoot()) 
		{

			// transferCost(g,s) = TRANSFER + min(in(g',s) + out(g",s), in(g",s) + out(g's))
			
			
			if (tentativeMapping.getCost() == INTER_SPECIES_DOMAIN_TRANSFER_DIFFERENT + leftToParentMapping.getInCost()
						+ rightToParentMapping.getOutCost()) {
					tentativeMapping.getOptions().add(1522);
					tentativeMapping.addWeightByInOut(leftToParentMapping, rightToParentMapping, this);
			}
			
			// transferCost(g,s) = TRANSFER + min(in(g',s) + out(g",s), in(g",s) + out(g's))
			else if(tentativeMapping.getCost() == INTER_SPECIES_DOMAIN_TRANSFER_DIFFERENT + leftToParentMapping.getOutCost()
			+ rightToParentMapping.getInCost()){
					tentativeMapping.getOptions().add(1622);
					tentativeMapping.addWeightByInOut(rightToParentMapping, leftToParentMapping, this);
			}
		}

	}
	private void setOptionsForIntToIntDomainMapping(Node domainNode, Node geneNode, Node leftDomainChild,
			Node rightDomainChild, Node leftGeneChild, Node rightGeneChild,
			DomainMapping tentativeMapping) {

		// let the left child to species node mapping be (g', s)
		DomainMapping leftToParentMapping = getDomainMappingFromNodes(leftDomainChild, geneNode);

		// let the right child to species node mapping be (g", s)
		DomainMapping rightToParentMapping = getDomainMappingFromNodes(rightDomainChild, geneNode);

		// let the left child of gene node to left child of species node mapping be (g',
		// s')
		DomainMapping leftToLeftMapping = getDomainMappingFromNodes(leftDomainChild, leftGeneChild);

		// let the left child of gene node to right child of species node mapping be
		// (g', s")
		DomainMapping leftToRightMapping = getDomainMappingFromNodes(leftDomainChild, rightGeneChild);

		// let the right child of gene node to left child of species node mapping be
		// (g", s')
		DomainMapping rightToLeftMapping = getDomainMappingFromNodes(rightDomainChild, leftGeneChild);

		// let the right child of gene node to right child of species node mapping be
		// (g", s")
		DomainMapping rightToRightMapping = getDomainMappingFromNodes(rightDomainChild, rightGeneChild);

		// speciationCost = min{in(g',s') + in (g",s"), in(g",s') + in (g', s")}
		int speciationCost = Math.min(leftToLeftMapping.getInCost() + rightToRightMapping.getInCost(),
				leftToRightMapping.getInCost() + rightToLeftMapping.getInCost());
		if (tentativeMapping.getCost() >= Integer.MAX_VALUE / 10)
			return;
		if (tentativeMapping.getCost() == speciationCost) {
			if (tentativeMapping.getCost() == leftToLeftMapping.getInCost() + rightToRightMapping.getInCost()) {
				tentativeMapping.getOptions().add(4);
				tentativeMapping.addWeightByInAndIn(leftToLeftMapping, rightToRightMapping, this);
			}
			if (tentativeMapping.getCost() == leftToRightMapping.getInCost() + rightToLeftMapping.getInCost()) {
				tentativeMapping.getOptions().add(5);
				tentativeMapping.addWeightByInAndIn(leftToRightMapping, rightToLeftMapping, this);
			}
		}

		int[] duplicationOptions = getDuplicationOptions(leftToParentMapping, rightToParentMapping, leftToLeftMapping,
				leftToRightMapping, rightToLeftMapping, rightToRightMapping);

		// Set duplicationCost = DUPLICATION + min(duplicationOptions)
		int duplicationCost = DOMAIN_DUPLICATION + getMinimum(duplicationOptions);
		if (tentativeMapping.getCost() == duplicationCost) {
			if (tentativeMapping.getCost() - DOMAIN_DUPLICATION == leftToParentMapping.getCost()
					+ rightToRightMapping.getInCost() + DOMAIN_LOSS) {
				tentativeMapping.getOptions().add(6);
				tentativeMapping.addWeightByNormalAndIn(leftToParentMapping, rightToRightMapping, this);
			}
			if (tentativeMapping.getCost() - DOMAIN_DUPLICATION == leftToParentMapping.getCost()
					+ rightToLeftMapping.getInCost() + DOMAIN_LOSS) {
				tentativeMapping.getOptions().add(7);
				tentativeMapping.addWeightByNormalAndIn(leftToParentMapping, rightToLeftMapping, this);

			}
			if (tentativeMapping.getCost() - DOMAIN_DUPLICATION == rightToParentMapping.getCost()
					+ leftToRightMapping.getInCost() + DOMAIN_LOSS) {
				tentativeMapping.getOptions().add(8);
				tentativeMapping.addWeightByNormalAndIn(rightToParentMapping, leftToRightMapping, this);
			}
			if (tentativeMapping.getCost() - DOMAIN_DUPLICATION == rightToParentMapping.getCost()
					+ leftToLeftMapping.getInCost() + DOMAIN_LOSS) {
				tentativeMapping.getOptions().add(9);
				tentativeMapping.addWeightByNormalAndIn(rightToParentMapping, leftToLeftMapping, this);

			}
			if (tentativeMapping.getCost() - DOMAIN_DUPLICATION == rightToParentMapping.getCost()
					+ leftToParentMapping.getCost()) {
				tentativeMapping.getOptions().add(10);
				tentativeMapping.addWeightByNormalAndNormal(rightToParentMapping, leftToParentMapping);

			}
			if (tentativeMapping.getCost() - DOMAIN_DUPLICATION == leftToLeftMapping.getInCost()
					+ rightToRightMapping.getInCost() + 2 * DOMAIN_LOSS) {
				tentativeMapping.getOptions().add(11);
				tentativeMapping.addWeightByInAndIn(leftToLeftMapping, rightToRightMapping, this);
			}
			if (tentativeMapping.getCost() - DOMAIN_DUPLICATION == leftToRightMapping.getInCost()
					+ rightToLeftMapping.getInCost() + 2 * DOMAIN_LOSS) {
				tentativeMapping.getOptions().add(12);
				tentativeMapping.addWeightByInAndIn(leftToRightMapping, rightToLeftMapping, this);

			}
			if (tentativeMapping.getCost() - DOMAIN_DUPLICATION == leftToLeftMapping.getInCost()
					+ rightToLeftMapping.getInCost() + 2 * DOMAIN_LOSS) {
				tentativeMapping.getOptions().add(13);
				tentativeMapping.addWeightByInAndIn(leftToLeftMapping, rightToLeftMapping, this);

			}
			if (tentativeMapping.getCost() - DOMAIN_DUPLICATION == leftToRightMapping.getInCost()
					+ rightToRightMapping.getInCost() + 2 * DOMAIN_LOSS) {
				tentativeMapping.getOptions().add(14);
				tentativeMapping.addWeightByInAndIn(leftToRightMapping, rightToRightMapping, this);

			}
		}
		if (!geneNode.isRoot()) {

			// transferCost(g,s) = TRANSFER + min(in(g',s) + out(g",s), in(g",s) + out(g's))
			if (tentativeMapping.getCost() == INTRA_SPECIES_DOMAIN_TRANSFER_SAME + leftToParentMapping.getInCost()
					+ rightToParentMapping.getIntraSameTreeOutCost()) {
				tentativeMapping.getOptions().add(1511);
				tentativeMapping.addWeightByIntraSameInOut(leftToParentMapping, rightToParentMapping, this);
			}
			// transferCost(g,s) = TRANSFER + min(in(g',s) + out(g",s), in(g",s) + out(g's))
			if (tentativeMapping.getCost() == INTRA_SPECIES_DOMAIN_TRANSFER_DIFFERENT + leftToParentMapping.getInCost()
								+ rightToParentMapping.getIntraDifferentTreeOutCost()) {
							tentativeMapping.getOptions().add(1512);
				tentativeMapping.addWeightByIntraDifferentInOut(leftToParentMapping, rightToParentMapping, this);
			}
			if (tentativeMapping.getCost() == INTER_SPECIES_DOMAIN_TRANSFER_SAME + leftToParentMapping.getInCost()
				+ rightToParentMapping.getInterSameTreeOutCost()) {
				tentativeMapping.getOptions().add(1521);
				tentativeMapping.addWeightByInterSameInOut(leftToParentMapping, rightToParentMapping, this);
			}
			// transferCost(g,s) = TRANSFER + min(in(g',s) + out(g",s), in(g",s) + out(g's))
			if (tentativeMapping.getCost() == INTER_SPECIES_DOMAIN_TRANSFER_DIFFERENT + leftToParentMapping.getInCost()
						+ rightToParentMapping.getInterDifferentTreeOutCost()) {
					tentativeMapping.getOptions().add(1522);
					tentativeMapping.addWeightByInterDifferentInOut(leftToParentMapping, rightToParentMapping, this);
			}
			if (tentativeMapping.getCost() == INTRA_SPECIES_DOMAIN_TRANSFER_SAME + leftToParentMapping.getIntraSameTreeOutCost()
					+ rightToParentMapping.getInCost()) {
				tentativeMapping.getOptions().add(1611);
				tentativeMapping.addWeightByIntraSameInOut(rightToParentMapping, leftToParentMapping, this);
			}
			if (tentativeMapping.getCost() == INTRA_SPECIES_DOMAIN_TRANSFER_DIFFERENT + leftToParentMapping.getIntraDifferentTreeOutCost()
					+ rightToParentMapping.getInCost()) {
				tentativeMapping.getOptions().add(1612);
				tentativeMapping.addWeightByIntraDifferentInOut(rightToParentMapping, leftToParentMapping, this);
			}
			if (tentativeMapping.getCost() == INTER_SPECIES_DOMAIN_TRANSFER_SAME + leftToParentMapping.getInterSameTreeOutCost()
				+ rightToParentMapping.getInCost()) {
				tentativeMapping.getOptions().add(1621);
				tentativeMapping.addWeightByInterSameInOut(rightToParentMapping, leftToParentMapping, this);
			}
			if (tentativeMapping.getCost() == INTER_SPECIES_DOMAIN_TRANSFER_DIFFERENT + leftToParentMapping.getInterDifferentTreeOutCost()
			+ rightToParentMapping.getInCost()) {
					tentativeMapping.getOptions().add(1622);
					tentativeMapping.addWeightByInterDifferentInOut(rightToParentMapping, leftToParentMapping, this);
			}
		}

	}
	
	private void setOptionsForIntToLeafGeneMapping(Node geneNode, Node speciesNode, Node leftGeneChild,
			Node rightSpeciesChild, GeneMapping tentativeMapping) {

		GeneMapping leftToParentMapping = getGeneMappingFromNodes(leftGeneChild, speciesNode);

		// let the right child to species node mapping be (g", s)
		GeneMapping rightToParentMapping = getGeneMappingFromNodes(rightSpeciesChild, speciesNode);
		int duplicationCost = GENE_DUPLICATION + leftToParentMapping.getCost() + rightToParentMapping.getCost();

		if (tentativeMapping.getCost() == duplicationCost) {
			tentativeMapping.getOptions().add(1);
			tentativeMapping.addWeightByChildren(leftToParentMapping, rightToParentMapping);
		}
		// if s is not root
		if (!speciesNode.isRoot()) {

			// transferCost(g,s) = TRANSFER + min(in(g',s) + out(g",s), in(g",s) + out(g's))
			if (tentativeMapping.getCost() == GENE_TRANSFER + leftToParentMapping.getInCost()
					+ rightToParentMapping.getOutCost()) {
				tentativeMapping.getOptions().add(2);
				tentativeMapping.addWeightByInOut(leftToParentMapping, rightToParentMapping, this);
			}
			if (tentativeMapping.getCost() == GENE_TRANSFER + leftToParentMapping.getOutCost()
					+ rightToParentMapping.getInCost()) {
				tentativeMapping.getOptions().add(3);
				tentativeMapping.addWeightByInOut(rightToParentMapping, leftToParentMapping, this);
			}
		}

	}
	private void setOptionsForIntToLeafDomainMappingApprox(Node domainNode, Node geneNode, Node leftDomainChild,
			Node rightDomainChild, DomainMapping tentativeMapping) {

		DomainMapping leftToParentMapping = getDomainMappingFromNodes(leftDomainChild, geneNode);

		// let the right child to species node mapping be (g", s)
		DomainMapping rightToParentMapping = getDomainMappingFromNodes(rightDomainChild, geneNode);
		int duplicationCost = DOMAIN_DUPLICATION + leftToParentMapping.getCost() + rightToParentMapping.getCost();

		if (tentativeMapping.getCost() == duplicationCost) {
			tentativeMapping.getOptions().add(1);
			tentativeMapping.addWeightByChildren(leftToParentMapping, rightToParentMapping);
		}
		// if s is not root
	//	if (!geneNode.isRoot())
		//{

			
			// transferCost(g,s) = TRANSFER + min(in(g',s) + out(g",s), in(g",s) + out(g's))
			if (tentativeMapping.getCost() == INTER_SPECIES_DOMAIN_TRANSFER_DIFFERENT + leftToParentMapping.getInCost()
						+ rightToParentMapping.getOutCost()) {
					tentativeMapping.getOptions().add(222);
					tentativeMapping.addWeightByInOut(leftToParentMapping, rightToParentMapping, this);
			}
			
			if (tentativeMapping.getCost() == INTER_SPECIES_DOMAIN_TRANSFER_DIFFERENT + leftToParentMapping.getOutCost()
			+ rightToParentMapping.getInCost()) {
					tentativeMapping.getOptions().add(322);
					tentativeMapping.addWeightByInOut(rightToParentMapping, leftToParentMapping, this);
			}
	//	}

	}


	private void setOptionsForIntToLeafDomainMapping(Node domainNode, Node geneNode, Node leftDomainChild,
			Node rightDomainChild, DomainMapping tentativeMapping) {

		DomainMapping leftToParentMapping = getDomainMappingFromNodes(leftDomainChild, geneNode);

		// let the right child to species node mapping be (g", s)
		DomainMapping rightToParentMapping = getDomainMappingFromNodes(rightDomainChild, geneNode);
		int duplicationCost = DOMAIN_DUPLICATION + leftToParentMapping.getCost() + rightToParentMapping.getCost();

		if (tentativeMapping.getCost() == duplicationCost) {
			tentativeMapping.getOptions().add(1);
			tentativeMapping.addWeightByChildren(leftToParentMapping, rightToParentMapping);
		}
		// if s is not root
	//	if (!geneNode.isRoot())
		{

			// transferCost(g,s) = TRANSFER + min(in(g',s) + out(g",s), in(g",s) + out(g's))
			if (tentativeMapping.getCost() == INTRA_SPECIES_DOMAIN_TRANSFER_SAME + leftToParentMapping.getInCost()
					+ rightToParentMapping.getIntraSameTreeOutCost()) {
				tentativeMapping.getOptions().add(211);
			}
			// transferCost(g,s) = TRANSFER + min(in(g',s) + out(g",s), in(g",s) + out(g's))
			if (tentativeMapping.getCost() == INTRA_SPECIES_DOMAIN_TRANSFER_DIFFERENT + leftToParentMapping.getInCost()
								+ rightToParentMapping.getIntraDifferentTreeOutCost()) {
							tentativeMapping.getOptions().add(212);
			}
			if (tentativeMapping.getCost() == INTER_SPECIES_DOMAIN_TRANSFER_SAME + leftToParentMapping.getInCost()
				+ rightToParentMapping.getInterSameTreeOutCost()) {
				tentativeMapping.getOptions().add(221);
				tentativeMapping.addWeightByInterSameInOut(leftToParentMapping, rightToParentMapping, this);
			}
			// transferCost(g,s) = TRANSFER + min(in(g',s) + out(g",s), in(g",s) + out(g's))
			if (tentativeMapping.getCost() == INTER_SPECIES_DOMAIN_TRANSFER_DIFFERENT + leftToParentMapping.getInCost()
						+ rightToParentMapping.getInterDifferentTreeOutCost()) {
					tentativeMapping.getOptions().add(222);
				}
			if (tentativeMapping.getCost() == INTRA_SPECIES_DOMAIN_TRANSFER_SAME + leftToParentMapping.getIntraSameTreeOutCost()
					+ rightToParentMapping.getInCost()) {
				tentativeMapping.getOptions().add(311);
			}
			if (tentativeMapping.getCost() == INTRA_SPECIES_DOMAIN_TRANSFER_DIFFERENT + leftToParentMapping.getIntraDifferentTreeOutCost()
					+ rightToParentMapping.getInCost()) {
				tentativeMapping.getOptions().add(312);
			}
			if (tentativeMapping.getCost() == INTER_SPECIES_DOMAIN_TRANSFER_SAME + leftToParentMapping.getInterSameTreeOutCost()
				+ rightToParentMapping.getInCost()) {
				tentativeMapping.getOptions().add(321);
			}
			if (tentativeMapping.getCost() == INTER_SPECIES_DOMAIN_TRANSFER_DIFFERENT + leftToParentMapping.getInterDifferentTreeOutCost()
			+ rightToParentMapping.getInCost()) {
					tentativeMapping.getOptions().add(322);
			}

		}

	}

	





	private int[] getDuplicationOptions(RootedBinaryTreeMapping leftToParentMapping,
			RootedBinaryTreeMapping rightToParentMapping, RootedBinaryTreeMapping leftToLeftMapping,
			RootedBinaryTreeMapping leftToRightMapping, RootedBinaryTreeMapping rightToLeftMapping,
			RootedBinaryTreeMapping rightToRightMapping) {

		int[] duplicationOptions = { leftToParentMapping.getCost() + rightToRightMapping.getInCost() + DOMAIN_LOSS,
				leftToParentMapping.getCost() + rightToLeftMapping.getInCost() + DOMAIN_LOSS,
				rightToParentMapping.getCost() + leftToRightMapping.getInCost() + DOMAIN_LOSS,
				rightToParentMapping.getCost() + leftToLeftMapping.getInCost() + DOMAIN_LOSS,
				rightToParentMapping.getCost() + leftToParentMapping.getCost(),
				leftToLeftMapping.getInCost() + rightToRightMapping.getInCost() + 2 * DOMAIN_LOSS,
				leftToRightMapping.getInCost() + rightToLeftMapping.getInCost() + 2 * DOMAIN_LOSS,
				leftToLeftMapping.getInCost() + rightToLeftMapping.getInCost() + 2 * DOMAIN_LOSS,
				leftToRightMapping.getInCost() + rightToRightMapping.getInCost() + 2 * DOMAIN_LOSS };

		return duplicationOptions;
	}
	private void handleIntToLeafGeneMapping(Node geneNode, Node speciesNode, Node leftDomainChild, Node rightDomainChild,
			RootedBinaryTreeMapping tentativeMapping, RootedBinaryTreeMapping leftToParentMapping,
			RootedBinaryTreeMapping rightToParentMapping) {
		// internal node to leaf node speciation cost should be infinity
		tentativeMapping.setSpeciationCost(Integer.MAX_VALUE / 10);

		// duplicationCost(g,s) = DUPLICATION + c(g',s) + c(g",s)
		int duplicationCost = GENE_DUPLICATION + leftToParentMapping.getCost() + rightToParentMapping.getCost();
		tentativeMapping.setDuplicationCost(duplicationCost);

		// if s is not root
		if (!speciesNode.isRoot()) {
			// transferCost(g,s) = TRANSFER + min(in(g',s) + out(g",s), in(g",s) + out(g's))
			int transferCost = GENE_TRANSFER + Math.min(leftToParentMapping.getInCost() + rightToParentMapping.getOutCost(),
					leftToParentMapping.getOutCost() + rightToParentMapping.getInCost());
			tentativeMapping.setTransferCost(transferCost);
		}

		// c(g,s) = min(speciationCost, duplicationCost, transferCost)
		tentativeMapping.setCost(getMinimumValue(tentativeMapping.getSpeciationCost(),
				tentativeMapping.getDuplicationCost(), tentativeMapping.getTransferCost()));

		// in(g,s) = c(g,s)
		tentativeMapping.setInCost(tentativeMapping.getCost());

		tentativeMapping.getInSpeciesNodes().clear();
		tentativeMapping.addInSpeciesNode(speciesNode);

		// inAlt(g,s) = c(g,s)
		tentativeMapping.setInAltCost(tentativeMapping.getCost());

		tentativeMapping.getInAltSpeciesNodes().clear();
		tentativeMapping.addInAltSpeciesNode(speciesNode);
	}
	
	private void handleIntToLeafDomainMappingApprox(Node domainNode, Node geneNode, Node leftDomainChild, Node rightDomainChild,
			DomainMapping tentativeMapping, DomainMapping leftToParentMapping,
			DomainMapping rightToParentMapping) {
		// internal node to leaf node speciation cost should be infinity
		tentativeMapping.setSpeciationCost(Integer.MAX_VALUE / 10);

		// duplicationCost(g,s) = DUPLICATION + c(g',s) + c(g",s)
		int duplicationCost = DOMAIN_DUPLICATION + leftToParentMapping.getCost() + rightToParentMapping.getCost();
		tentativeMapping.setDuplicationCost(duplicationCost);

		// if s is not root
		//if (!geneNode.isRoot())
		{
			
			
			int transferCost =  INTER_SPECIES_DOMAIN_TRANSFER_DIFFERENT +  Math.min(
					leftToParentMapping.getInCost() + rightToParentMapping.getOutCost(),
					leftToParentMapping.getOutCost() + rightToParentMapping.getInCost());
			tentativeMapping.setTransferCost(transferCost);

		}

		// c(g,s) = min(speciationCost, duplicationCost, transferCost)
		tentativeMapping.setCost(getMinimumValue(tentativeMapping.getSpeciationCost(),
				tentativeMapping.getDuplicationCost(), tentativeMapping.getTransferCost()));

		// in(g,s) = c(g,s)
		tentativeMapping.setInCost(tentativeMapping.getCost());

		tentativeMapping.getInSpeciesNodes().clear();
		tentativeMapping.addInSpeciesNode(geneNode);

		// inAlt(g,s) = c(g,s)
		tentativeMapping.setInAltCost(tentativeMapping.getCost());

		tentativeMapping.getInAltSpeciesNodes().clear();
		tentativeMapping.addInAltSpeciesNode(geneNode);

	}
	private void handleIntToLeafDomainMapping(Node domainNode, Node geneNode, Node leftDomainChild, Node rightDomainChild,
			DomainMapping tentativeMapping, DomainMapping leftToParentMapping,
			DomainMapping rightToParentMapping) {
		// internal node to leaf node speciation cost should be infinity
		tentativeMapping.setSpeciationCost(Integer.MAX_VALUE / 10);

		// duplicationCost(g,s) = DUPLICATION + c(g',s) + c(g",s)
		int duplicationCost = DOMAIN_DUPLICATION + leftToParentMapping.getCost() + rightToParentMapping.getCost();
		tentativeMapping.setDuplicationCost(duplicationCost);

		// if s is not root
	//	if (!geneNode.isRoot())
		{
			// transferCost(g,s) = TRANSFER + min(in(g',s) + out(g",s), in(g",s) + out(g's))
			// check all four combinations of domain transfer costs
			int intraSpeciesSameTreeTransferCost = INTRA_SPECIES_DOMAIN_TRANSFER_SAME +  Math.min(
					leftToParentMapping.getInCost() + rightToParentMapping.getIntraSameTreeOutCost(),
					leftToParentMapping.getIntraSameTreeOutCost() + rightToParentMapping.getInCost());
			int intraSpeciesDifferentTreeTransferCost = INTRA_SPECIES_DOMAIN_TRANSFER_DIFFERENT +  Math.min(
					leftToParentMapping.getInCost() + rightToParentMapping.getIntraDifferentTreeOutCost(),
					leftToParentMapping.getIntraDifferentTreeOutCost() + rightToParentMapping.getInCost());
			int interSpeciesSameTreeTransferCost = INTER_SPECIES_DOMAIN_TRANSFER_SAME +  Math.min(
					leftToParentMapping.getInCost() + rightToParentMapping.getInterSameTreeOutCost(),
					leftToParentMapping.getInterSameTreeOutCost() + rightToParentMapping.getInCost());
			int interSpeciesDifferentTreeTransferCost = INTER_SPECIES_DOMAIN_TRANSFER_DIFFERENT +  Math.min(
					leftToParentMapping.getInCost() + rightToParentMapping.getInterDifferentTreeOutCost(),
					leftToParentMapping.getInterDifferentTreeOutCost() + rightToParentMapping.getInCost());
		//	tentativeMapping.setIntraSameTreeOutCost(intraSameTreeOutCost);
			int transferCost = getMinimumOfFourTransferCost(intraSpeciesSameTreeTransferCost, 
					intraSpeciesDifferentTreeTransferCost, interSpeciesSameTreeTransferCost, 
					interSpeciesDifferentTreeTransferCost);
			tentativeMapping.setTransferCost(transferCost);

		}

		// c(g,s) = min(speciationCost, duplicationCost, transferCost)
		tentativeMapping.setCost(getMinimumValue(tentativeMapping.getSpeciationCost(),
				tentativeMapping.getDuplicationCost(), tentativeMapping.getTransferCost()));

		// in(g,s) = c(g,s)
		tentativeMapping.setInCost(tentativeMapping.getCost());

		tentativeMapping.getInSpeciesNodes().clear();
		tentativeMapping.addInSpeciesNode(geneNode);

		// inAlt(g,s) = c(g,s)
		tentativeMapping.setInAltCost(tentativeMapping.getCost());

		tentativeMapping.getInAltSpeciesNodes().clear();
		tentativeMapping.addInAltSpeciesNode(geneNode);

	}
	// get minimum of four numbers
	private int getMinimumOfFourTransferCost(int one,
			int two, int three,
			int four) {
		int minimum = Math.min(getMinimumValue(one, two, three), four);
		return minimum;
	}

	private void initializeAllGeneMappings() {
		
		int totalGeneNodes = getTotalGeneNodes();
		int totalSpeciesNodes = speciesTree.getAllNodesByPostOrder().size();
		ArrayList<Node>allSpeciesNodesbyPostOrder = speciesTree.getAllNodesByPostOrder();
		ArrayList<Node>allGeneNodesbyPostOrder = getAllGeneNodesByPostOrder();
		
		allPossibleGeneMappings = new GeneMapping[totalGeneNodes][totalSpeciesNodes];
		for (int i = 0; i < totalGeneNodes; i++) {
			Node geneNode = allGeneNodesbyPostOrder.get(i);
			for (int j = 0; j < totalSpeciesNodes; j++) {

				Node speciesNode = allSpeciesNodesbyPostOrder.get(j);
				allPossibleGeneMappings[i][j] = new GeneMapping(geneNode, speciesNode);

			}

		}

	}
	private void initializeAllDomainMappings() {
		int totalDomainNodes = getTotalDomainNodes();
		int totalGeneNodes = getTotalGeneNodes();
		ArrayList<Node>allDomainNodesbyPostOrder = getAllDomainNodesByPostOrder();
		ArrayList<Node>allGeneNodesbyPostOrder = getAllGeneNodesByPostOrder();
		
		allPossibleDomainMappings = new DomainMapping[totalDomainNodes][totalGeneNodes];
		for (int i = 0; i < totalDomainNodes; i++) {
			Node domainNode = allDomainNodesbyPostOrder.get(i);
			for (int j = 0; j < totalGeneNodes; j++) {

				Node geneNode = allGeneNodesbyPostOrder.get(j);
				allPossibleDomainMappings[i][j] = new DomainMapping(domainNode, geneNode);

			}

		}

	}

	private ArrayList<Node> getAllDomainNodesByPostOrder() {
	 ArrayList<Node>allDomainNodesByPostOrder= new ArrayList<Node>();
	 for (DomainTreeRooted domainTree: domainTrees) {
		 ArrayList<Node>allDomainNodes = domainTree.getAllNodesByPostOrder();
		 allDomainNodesByPostOrder.addAll(allDomainNodes);
	 }
	return allDomainNodesByPostOrder;
	}
	
	 private ArrayList<Node> getAllPostOrderInternalGeneNodes() {
		 ArrayList<Node>allGeneNodesByPostOrder= new ArrayList<Node>();
		 for (GeneTreeRooted geneTree: geneTrees) {
			 ArrayList<Node>allGeneNodes = geneTree.getAllPostOrderInternalNodes();
			 allGeneNodesByPostOrder.addAll(allGeneNodes);
		 }
		return allGeneNodesByPostOrder;
		}

	private ArrayList<Node> getAllGeneNodesByPostOrder() {
		 ArrayList<Node>allGeneNodesByPostOrder= new ArrayList<Node>();
		 for (GeneTreeRooted geneTree: geneTrees) {
			 ArrayList<Node>allGeneNodes = geneTree.getAllNodesByPostOrder();
			 allGeneNodesByPostOrder.addAll(allGeneNodes);
		 }
		return allGeneNodesByPostOrder;
	}
	private ArrayList<Node> getAllLeafGeneNodes() {
		 ArrayList<Node>allLeafGeneNodes= new ArrayList<Node>();
		 for (GeneTreeRooted geneTree: geneTrees) {
			 ArrayList<Node>geneLeaves = geneTree.getLeafSet();
			 allLeafGeneNodes.addAll(geneLeaves);
		 }
		return allLeafGeneNodes;
	}
	private int getTotalGeneNodes() {
		int totalCount = 0;
		for(GeneTreeRooted geneTree: geneTrees) {
			int size = geneTree.getAllNodesByPostOrder().size();
			totalCount+=size;
		} 
			return totalCount;
}

	private int getTotalDomainNodes() {
		int totalCount = 0;
	for(DomainTreeRooted domainTree: domainTrees) {
		int size = domainTree.getAllNodesByPostOrder().size();
		totalCount+=size;
	} 
		return totalCount;
	}

	
	private int getMinimum(int[] array) {
		int minimum = Integer.MAX_VALUE;
		for (int value : array) {
			if (value < minimum)
				minimum = value;
		}
		return minimum;
	}

	// get minimum value of three integers
	private int getMinimumValue(int a, int b, int c) {

		return Math.min(a, Math.min(b, c));
	}

	// perform some pre-processing of gene tree and species tree by pre-order,
	// post-order and in-order

	public void addGeneTree(GeneTreeRooted geneTree) {
		geneTrees.add(geneTree);
	}

	public void addDomainTree(DomainTreeRooted domainTree) {
		domainTrees.add(domainTree);
	}
	
	public RootedBinaryTreeMapping getMappingFromNodesByIndex(Node firstNode, Node secondNode, RootedBinaryTreeMapping mapping) {
		if (mapping instanceof DomainMapping)
			return getDomainMappingFromNodes(firstNode, secondNode);
		else
			return getGeneMappingFromNodes(firstNode, secondNode);
	}



	
}
