package mapping;

import java.math.BigInteger;
import java.util.*;

import dtl_algo.DTLReconciliation;
import edge.Edge;
import node.Node;
import node.SpeciesNodeRooted;

public class RootedBinaryTreeMapping {
	Node firstNode;
	Node secondNode;
	BigInteger numberOfSolutions = new BigInteger("0");
	ArrayList<Node> inNodes= new ArrayList<Node>();
	ArrayList<Node> outNodes= new ArrayList<Node>();
	ArrayList<Node> inAltNodes= new ArrayList<Node>();
	int cost = Integer.MAX_VALUE / 10;
	int inCost = Integer.MAX_VALUE / 10;
	int inAltCost = Integer.MAX_VALUE / 10;
	int outCost = Integer.MAX_VALUE / 10;
	int speciationCost = Integer.MAX_VALUE / 10;
	int duplicationCost = Integer.MAX_VALUE / 10;
	int transferCost = Integer.MAX_VALUE / 10;
	ArrayList<Integer> options = new ArrayList<Integer>();
	ArrayList<BigInteger>optionWeights = new ArrayList<>();
	ArrayList<Edge> lostEdges = new ArrayList<>();
	ArrayList<Node> losses = new ArrayList<Node>();
	Node additiveTransferRecipient = null;
	
	
	public RootedBinaryTreeMapping(Node firstNode, Node secondNode) {
		
		this.firstNode = firstNode;
		this.secondNode = secondNode;
	}

	public BigInteger getNumberOfSolutions() {
		return numberOfSolutions;
	}

	public void setNumberOfSolutions(BigInteger numberOfSolutions) {
		this.numberOfSolutions = numberOfSolutions;
	}

	

	public ArrayList<Edge> getLostEdges() {
		return lostEdges;
	}

	public void setLostEdges(ArrayList<Edge> lostEdges) {
		this.lostEdges = lostEdges;
	}

	public Node getAdditiveTransferRecipient() {
		return additiveTransferRecipient;
	}

	public void setAdditiveTransferRecipient(Node additiveTransferRecipient) {
		this.additiveTransferRecipient = additiveTransferRecipient;
	}

	



	public ArrayList<Node> getOutSpeciesNodes() {
		return outNodes;
	}

	public void setOutSpeciesNodes(ArrayList<Node> outSpeciesNodes) {
		this.outNodes = outSpeciesNodes;
	}

	public ArrayList<Node> getInAltSpeciesNodes() {
		return inAltNodes;
	}

	public void setInAltSpeciesNodes(ArrayList<Node> inAltSpeciesNodes) {
		this.inAltNodes = inAltSpeciesNodes;
	}

	public ArrayList<Node> getInSpeciesNodes() {
		return inNodes;
	}

	public void setInSpeciesNodes(ArrayList<Node> inSpeciesNodes) {
		this.inNodes = inSpeciesNodes;
	}

	public ArrayList<Integer> getOptions() {
		return options;
	}

	public void setOptions(ArrayList<Integer> options) {
		this.options = options;
	}

	



	

	public ArrayList<BigInteger> getOptionWeights() {
		return optionWeights;
	}

	public void setOptionWeights(ArrayList<BigInteger> optionWeights) {
		this.optionWeights = optionWeights;
	}

	public int getCost() {
		return cost;
	}

	public void setCost(int cost) {
		this.cost = cost;
	}

	public int getInCost() {
		return inCost;
	}

	public void setInCost(int inCost) {
		this.inCost = inCost;
	}

	public int getInAltCost() {
		return inAltCost;
	}

	public void setInAltCost(int inAltCost) {
		this.inAltCost = inAltCost;
	}

	public int getOutCost() {
		return outCost;
	}

	public void setOutCost(int outCost) {
		this.outCost = outCost;
	}

	public int getSpeciationCost() {
		return speciationCost;
	}

	public void setSpeciationCost(int speciationCost) {
		this.speciationCost = speciationCost;
	}

	public int getDuplicationCost() {
		return duplicationCost;
	}

	public void setDuplicationCost(int duplicationCost) {
		this.duplicationCost = duplicationCost;
	}

	public int getTransferCost() {
		return transferCost;
	}

	public void setTransferCost(int transferCost) {
		this.transferCost = transferCost;
	}

	public ArrayList<Node> getLosses() {
		return losses;
	}

	public void setLosses(ArrayList<Node> losses) {
		this.losses = losses;
	}



	



	public void updateLosses(ArrayList<Node> nodeList1,
			ArrayList<Node> nodeList2) {
		losses.addAll(nodeList1);
		losses.addAll(nodeList2);
		
	}

	public void printLossInformation() {
		System.out.print("The number of losses: "+losses.size()+" and the lost nodes are: ");
		for(Node node: losses) {
			System.out.print(node.getName()+" ");
		}
		System.out.println();
		System.out.print("The number of lost branches: "+lostEdges.size()+" and the lost branches are: ");
		for(Edge edge: lostEdges) {
			System.out.print(edge+" ");
		}
		System.out.println();
	}



	
	public void updateLossesWithBranch(Node childImage) {
		losses.addAll (childImage.getLosses(secondNode));
		concatenateBranchLosses( childImage.getLostBranches(secondNode));
		
	}
	public void addLoss(Node node) {
		losses.add(node);
		
	}

	public void updateLosses(Node leftChildImage, Node rightChildImage) {
		ArrayList<Node> leftChildLosses = leftChildImage.getLosses(secondNode);
		ArrayList<Node>rightChildLosses = rightChildImage.getLosses(secondNode);
		
		concatenateLosses( leftChildLosses);
		
		concatenateLosses( rightChildLosses);

	}

	public void updateLossesWithBranch(SpeciesNodeRooted leftChildImage, SpeciesNodeRooted rightChildImage) {
		ArrayList<Node> leftChildLosses = leftChildImage.getLosses(secondNode);
		ArrayList<Edge> leftChildLostBranches = leftChildImage.getLostBranches(secondNode);
		ArrayList<Node>rightChildLosses = rightChildImage.getLosses(secondNode);
		ArrayList<Edge> rightChildLostBranches = rightChildImage.getLostBranches(secondNode);
		
		concatenateLosses( leftChildLosses);
		
		concatenateLosses( rightChildLosses);

		concatenateBranchLosses(leftChildLostBranches);
		concatenateBranchLosses(rightChildLostBranches);
		
	}
	
	private void concatenateLosses( ArrayList<Node> newLosses) {
		for(Node node: newLosses) {
			losses.add(node);
		}
	}
	private void concatenateBranchLosses( ArrayList<Edge> newLosses) {
		for(Edge edge: newLosses) {
		addLostBranch(edge);
		
		}
	}
	public void addLostBranch(Edge branch) {
		lostEdges.add(branch);
		branch.incrementLossCount();
	}

	public void updateInTransferLossesWithBranch(Node childImage) {
		losses.addAll(childImage.getInTrasferLosses(secondNode));
		
		concatenateBranchLosses(childImage.getInTransferLostBranches(secondNode));
		
	}

	public Node getInSpeciesNodeRandomly() {
		if(inNodes.size()==0)return null;
		Random random = new Random();
		int index = random.nextInt(inNodes.size());
		return inNodes.get(index);
	}

	public void addInSpeciesNode(Node node) {
		inNodes.add(node);
	}

	public void addInAltSpeciesNode(Node speciesNode) {
		inAltNodes.add(speciesNode);
	}

	public Node getInAltSpeciesNodeRandomly() {
		if(inAltNodes.size()==0)return null;
		Random random = new Random();
		int index = random.nextInt(inAltNodes.size());
		return inAltNodes.get(index);
	}

	public Node getOutSpeciesNodeRandomly() {
		if(outNodes.size()==0)return null;
		Random random = new Random();
		int index = random.nextInt(outNodes.size());
		return outNodes.get(index);
	}

	public void addOutSpeciesNode(Node node) {
		if(node!=null)
		outNodes.add(node);
		
	}

	public void addWeightByChildren(RootedBinaryTreeMapping leftToParentMapping,
			RootedBinaryTreeMapping rightToParentMapping) {
		BigInteger newWeight = leftToParentMapping.numberOfSolutions.multiply(rightToParentMapping.numberOfSolutions);
		optionWeights.add(newWeight);
		
	}

	public void addWeightByInOut(RootedBinaryTreeMapping firstMapping,
			RootedBinaryTreeMapping secondMapping, DTLReconciliation reconciliation) {
		BigInteger inCounts = new BigInteger("0");
		BigInteger outCounts = new BigInteger("0");
		for(Node inNode: firstMapping.getInSpeciesNodes()) {
			Node geneNode = firstMapping.firstNode;
			RootedBinaryTreeMapping inMapping = reconciliation.getMappingFromNodesByIndex(geneNode, inNode, this);
			inCounts = inCounts.add(inMapping.numberOfSolutions);
		}
		for(Node outNode: secondMapping.getOutSpeciesNodes()) {
			Node geneNode = secondMapping.firstNode;
			RootedBinaryTreeMapping outMapping = reconciliation.getMappingFromNodesByIndex(geneNode, outNode, this);
			outCounts = outCounts.add(outMapping.numberOfSolutions);
		}
		
		optionWeights.add(inCounts.multiply(outCounts));
		
	}

	public void addWeightByInAndIn(RootedBinaryTreeMapping firstMapping,
			RootedBinaryTreeMapping secondMapping, DTLReconciliation reconciliation) {
		BigInteger firstInCount = new BigInteger("0");
		BigInteger secondInCount = new BigInteger("0");
		for(Node inNode: firstMapping.getInSpeciesNodes()) {
			Node geneNode = firstMapping.firstNode;
			RootedBinaryTreeMapping inMapping = reconciliation.getMappingFromNodesByIndex(geneNode, inNode, this);
			firstInCount =firstInCount.add(inMapping.numberOfSolutions);
		}
		for(Node inNode: secondMapping.getInSpeciesNodes()) {
			Node geneNode = secondMapping.firstNode;
			RootedBinaryTreeMapping inMapping = reconciliation.getMappingFromNodesByIndex(geneNode, inNode, this);
			secondInCount =secondInCount.add(inMapping.numberOfSolutions);
		}
		optionWeights.add(firstInCount.multiply(secondInCount));
		
	}

	public void addWeightByNormalAndIn(RootedBinaryTreeMapping firstMapping,
			RootedBinaryTreeMapping inMapping, DTLReconciliation reconciliation) {
		BigInteger inCount = new BigInteger("0");
		
		for(Node inNode: inMapping.getInSpeciesNodes()) {
			Node geneNode = inMapping.firstNode;
			RootedBinaryTreeMapping sampleInMapping = reconciliation.getMappingFromNodesByIndex(geneNode, inNode, this);
		inCount = 	inCount.add(sampleInMapping.numberOfSolutions);
		}
		optionWeights.add(firstMapping.numberOfSolutions.multiply(inCount));
		
		
	}

	public void addWeightByNormalAndNormal(RootedBinaryTreeMapping firstMapping,
			RootedBinaryTreeMapping secondMapping ) {
		optionWeights.add(firstMapping.numberOfSolutions.multiply(secondMapping.numberOfSolutions));
		
	}

	public void sumUpNumberOfOptions() {
		BigInteger sum = new BigInteger("0");
		for(BigInteger weight: optionWeights) {
			sum = sum.add(weight);
		}
		numberOfSolutions = sum;
	}
	public RootedBinaryTreeMapping getInMappingRandomly(DTLReconciliation reconciliation) {
		if(inNodes.size()==0)return null;
		Collections.shuffle(inNodes);
		Node inNode = inNodes.get(0);
		RootedBinaryTreeMapping mapping = reconciliation.getMappingFromNodesByIndex(firstNode, inNode, this);
		return mapping;
	}
	public RootedBinaryTreeMapping getInMappingGreedily(DTLReconciliation reconciliation) {

		if(inNodes.size()==0)return null;
		

		RootedBinaryTreeMapping greedyMapping = null;
		BigInteger numberOfSolutions  = new BigInteger("0");
		
		for(Node inNode : inNodes )
		{
			RootedBinaryTreeMapping mapping = reconciliation.getMappingFromNodesByIndex(firstNode, inNode, this);
			int comparison = mapping.getNumberOfSolutions().compareTo(numberOfSolutions);
			if(comparison==1) {
				numberOfSolutions = mapping.getNumberOfSolutions();
				greedyMapping = mapping;
			}
			else if (comparison==0) {
				int choice =reconciliation.binaryChoice(); 
				if(choice == 1) {
					numberOfSolutions = mapping.getNumberOfSolutions();
					greedyMapping = mapping;
				}
			}
			// here i am
		//	if(inNode.isCandidadateNode(geneNode)) {
			//	inNode.addCandidateNode(geneNode);
		//	inNode.incrementMappingCount();
		//	}
		}
			return greedyMapping;
		
	
	}
	public RootedBinaryTreeMapping getOutMappingRandomly(DTLReconciliation reconciliation) {

		if(outNodes.size()==0)return null;
		Collections.shuffle(outNodes);
		Node outNode = outNodes.get(0);
		RootedBinaryTreeMapping mapping = reconciliation.getMappingFromNodesByIndex(firstNode, outNode, this);		
		return mapping;
		
	
	}

	public RootedBinaryTreeMapping getOutMappingGreedily(DTLReconciliation reconciliation) {

		if(outNodes.size()==0)return null;
		

		RootedBinaryTreeMapping greedyMapping = null;
		BigInteger numberOfSolutions  = new BigInteger("0");
		
		for(Node outNode : outNodes )
		{
			RootedBinaryTreeMapping mapping = reconciliation.getMappingFromNodesByIndex(firstNode, outNode, this);
			int comparison = mapping.getNumberOfSolutions().compareTo(numberOfSolutions);
			if(comparison==1) {
				numberOfSolutions = mapping.getNumberOfSolutions();
				greedyMapping = mapping;
			}
			else if (comparison==0) {
				int choice =reconciliation.binaryChoice(); 
				if(choice == 1) {
					numberOfSolutions = mapping.getNumberOfSolutions();
					greedyMapping = mapping;
				}
			}
	//		outNode.incrementMappingCount();
		}
			return greedyMapping;
		
	
	}

	public RootedBinaryTreeMapping getInMappingUniformRandomly(DTLReconciliation reconciliation) {


		if(inNodes.size()==0)return null;
		RootedBinaryTreeMapping uniformMapping = null;
		BigInteger numberOfSolutions  = new BigInteger("0");
	
		
		for (Node inNode : inNodes)
		{
			RootedBinaryTreeMapping mapping = reconciliation.getMappingFromNodesByIndex(firstNode, inNode, this);
			numberOfSolutions =  numberOfSolutions.add(mapping.getNumberOfSolutions());
		}
		double random = Math.random() * numberOfSolutions.doubleValue();
		for (int i = 0; i < inNodes.size(); ++i)
		{
			Node inNode = inNodes.get(i);
			RootedBinaryTreeMapping mapping = reconciliation.getMappingFromNodesByIndex(firstNode, inNode,this);
		    random -= mapping.getNumberOfSolutions().doubleValue();
		    if (random <= 0.0d)
		    {
		        uniformMapping = mapping;
		        break;
		    }
		}
		if( uniformMapping==null) {
			return getInMappingRandomly(reconciliation);
		}
		return uniformMapping;	
	}

	public RootedBinaryTreeMapping getOutMappingUniformRandomly(DTLReconciliation reconciliation) {


		if(outNodes.size()==0)return null;
		RootedBinaryTreeMapping uniformMapping = null;
		BigInteger numberOfSolutions  = new BigInteger("0");
	
		
		for (Node inNode : outNodes)
		{
			RootedBinaryTreeMapping mapping = reconciliation.getMappingFromNodesByIndex(firstNode, inNode, this);
			numberOfSolutions =  numberOfSolutions.add(mapping.getNumberOfSolutions());
		}
		double random = Math.random() * numberOfSolutions.doubleValue();
		for (int i = 0; i < outNodes.size(); ++i)
		{
			Node outNode = outNodes.get(i);
			RootedBinaryTreeMapping mapping = reconciliation.getMappingFromNodesByIndex(firstNode, outNode, this);
		    random -= mapping.getNumberOfSolutions().doubleValue();
		    if (random <= 0.0d)
		    {
		        uniformMapping = mapping;
		        break;
		    }
		}
		if( uniformMapping==null) {
			return getOutMappingRandomly(reconciliation);
		}
		return uniformMapping;	
	}

	public boolean isTransferPossible() {
		if(cost== transferCost)return true;
		return false;
	}

	

	public Node getFirstNode() {
		return firstNode;
	}

	public void setFirstNode(Node firstNode) {
		this.firstNode = firstNode;
	}

	public Node getSecondNode() {
		return secondNode;
	}

	public void setSecondNode(Node secondNode) {
		this.secondNode = secondNode;
	}

	public ArrayList<Node> getInNodes() {
		return inNodes;
	}

	public void setInNodes(ArrayList<Node> inNodes) {
		this.inNodes = inNodes;
	}

	public ArrayList<Node> getOutNodes() {
		return outNodes;
	}

	public void setOutNodes(ArrayList<Node> outNodes) {
		this.outNodes = outNodes;
	}

	public ArrayList<Node> getInAltNodes() {
		return inAltNodes;
	}

	public void setInAltNodes(ArrayList<Node> inAltNodes) {
		this.inAltNodes = inAltNodes;
	}

	public RootedBinaryTreeMapping getTransferBiasedOutMapping(DTLReconciliation reconciliation) {


		if(outNodes.size()==0)return null;
		

		RootedBinaryTreeMapping greedyMapping = null;
		BigInteger numberOfSolutions  = new BigInteger("0");
		
		for(Node outNode : outNodes )
		{
			RootedBinaryTreeMapping mapping = reconciliation.getMappingFromNodesByIndex(firstNode, outNode, this);
			int comparison = mapping.getNumberOfSolutions().compareTo(numberOfSolutions);
			if(mapping.isTransferPossible()) {
				numberOfSolutions = mapping.getNumberOfSolutions();
				greedyMapping = mapping;
				break;
			}
			else if(comparison==1) {
				numberOfSolutions = mapping.getNumberOfSolutions();
				greedyMapping = mapping;
			}
			else if (comparison==0) {
				int choice =reconciliation.binaryChoice(); 
				if(choice == 1) {
					numberOfSolutions = mapping.getNumberOfSolutions();
					greedyMapping = mapping;
				}
			}
		}
			return greedyMapping;
		
	
	
	}
	public int getRandomOption() {
		Collections.shuffle(options);
		if(options.size()==0) {
			System.out.println("Error!");
		}
		return options.get(0);
	}
	public int getGreedyOption() {
		BigInteger greedyOptionValue  = new BigInteger("0");
		int greedyOption = 0;
		int i = 0;
		for(BigInteger weight : optionWeights )
		{

			int comparison = weight.compareTo(greedyOptionValue);
			
			 if(comparison==1) {
				greedyOptionValue = weight;
				greedyOption = options.get(i);
			}
			else if (comparison==0) {
				int choice = new Random().nextInt(2);
				if(choice == 1) {
					greedyOptionValue = weight;
					greedyOption = options.get(i);
				}
			}
			 i++;
		}
		return greedyOption;
	}


	@Override
	public String toString() {
		return "[" + firstNode + ", " + secondNode + "]";
	}
	
}
