package utility;

import java.io.*;
import java.util.*;

import dtl_algo.DTLReconciliation;
import node.*;

import tree.*;

public class FileUtility {
	// create species tree from input string (species tree part)
	public static SpeciesTreeRooted getSpeciesTreeFromFile(String treeString) {

		treeString = treeString.replaceAll("\\s", "");
		treeString = treeString.replaceAll(";", "");

		// System.out.println(treeString);
		int length = treeString.length();
		Stack<SpeciesNodeRooted> stack = new Stack<SpeciesNodeRooted>();
		String species = "";
		for (int i = 0; i < length; i++) {
			char presentCharacter = treeString.charAt(i);
			if (presentCharacter == '(' || presentCharacter == ',' || presentCharacter == ';')
				continue;
			else if (presentCharacter == ')') {
				SpeciesNodeRooted speciesRight = stack.pop();
				SpeciesNodeRooted speciesLeft = stack.pop();

				int nextStopCharIndex = getNextStopCharIndex(treeString, i);
				SpeciesNodeRooted parent = new SpeciesNodeRooted(speciesLeft.getName() + "$" + speciesRight.getName());
				
				if (nextStopCharIndex != i + 1 && nextStopCharIndex != -1) {
					String newParentName = (treeString.substring(i + 1, nextStopCharIndex));
					parent.setName(newParentName);
					i = nextStopCharIndex - 1;
				}
				parent.setLeft(speciesLeft);
				parent.setRight(speciesRight);
				// System.out.println(parent.getName());
				stack.push(parent);
			} else {
				species = "";
				while (treeString.charAt(i) != ',' && treeString.charAt(i) != ')') {
					species += treeString.charAt(i);
					i++;
				}
				if (treeString.charAt(i) == ')')
					i--;
				String fullName = species;
				
					
				//species = refineNameFromNewick(species);
				SpeciesNodeRooted speciesNode = new SpeciesNodeRooted(species);
				if(fullName.contains("_")) {
					speciesNode.setRealName(fullName);
				}
				
				stack.push(speciesNode);
				//System.out.println(speciesNode.getName());
			}
		}
		SpeciesNodeRooted root = stack.pop();
		root.setInternalNodeName(root.getName());
		SpeciesTreeRooted speciesTree = new SpeciesTreeRooted(root);
		return speciesTree;
	}

	

	private static int getNextStopCharIndex(String treeString, int i) {
		int nextCommaIndex = treeString.indexOf(',', i + 1);
		int nextSemiColonIndex = treeString.length();
		int nextBracketIndex = treeString.indexOf(')', i + 1);

		if (nextBracketIndex == -1)
			nextBracketIndex = Integer.MAX_VALUE;
		if (nextCommaIndex == -1)
			nextCommaIndex = Integer.MAX_VALUE;
		if (nextSemiColonIndex == -1)
			nextSemiColonIndex = Integer.MAX_VALUE;
		// System.out.println(nextSemiColonIndex);
		return Math.min(nextCommaIndex, Math.min(nextBracketIndex, nextSemiColonIndex));
	}

	public static String alignRangerSpeciesTreeFormat(Node node) {

		String output = "";
		String leftPart = "";
		String rightPart = "";
		if (node.getLeft().isLeaf()) {
			leftPart = node.getLeft().getName();
		} else {
			leftPart = alignRangerSpeciesTreeFormat(node.getLeft());
		}
		if (node.getRight().isLeaf()) {
			rightPart = node.getRight().getName();
		} else {
			
			rightPart = alignRangerSpeciesTreeFormat(node.getRight());

		}
		output = "(" + leftPart + "," + rightPart + ")" + node.getName();
		if (node.isRoot()) {

			output += ";";
		}

		return output;

	}
	public static void processInput(String speciesTreeFileName, String geneTreeFolderName, String domainTreeFileName,
			DTLReconciliation reconciliation) throws IOException {
			processSpeciesTree(speciesTreeFileName, reconciliation);
			processGeneTrees(geneTreeFolderName, reconciliation);
			processDomainTree(domainTreeFileName, reconciliation);
			
	}



	

	private static void processDomainTree(String domainTreeFileName, DTLReconciliation reconciliation) throws IOException{
		Scanner scanner = new Scanner(new File(domainTreeFileName));	
		String domainTreeString = scanner.nextLine();
		domainTreeString = domainTreeString.replaceAll("\\s", "");
		DomainTreeRooted domainTree = getDomainTreeFromFile(domainTreeString);
		setPreOrderNames(domainTree);
		String rangerDomainTreeFormat = alignRangerDomainTreeFormat(domainTree.getRoot());
		
		domainTree.setNewickOutputFormat(rangerDomainTreeFormat);
		reconciliation.addDomainTree(domainTree);
		scanner.close();
	}

	public static String alignRangerDomainTreeFormat(Node node) {

		String name = node.getInternalNodeName(); //"d" + (node.getPreOrderNumber() + 1);
		node.setName(name);
		String output = "";
		String leftPart = "";
		String rightPart = "";
		if (node.getLeft().isLeaf()) {
			DomainNodeRooted domainNodeLeft = (DomainNodeRooted) node.getLeft();
			leftPart = domainNodeLeft.getFullName();
		} else {
			leftPart = alignRangerDomainTreeFormat(node.getLeft());
		}
		if (node.getRight().isLeaf()) {
			DomainNodeRooted domainNodeRight = (DomainNodeRooted) node.getRight();
			rightPart = domainNodeRight.getFullName();
		} else {
		
				rightPart = alignRangerDomainTreeFormat(node.getRight());
		
		}
		output = "(" + leftPart + "," + rightPart + ")" + name;
		if (node.isRoot()) {

			output += ";";
		}

		return output;
	
	}


	private static void processGeneTrees(String geneTreeFolderName, DTLReconciliation reconciliation) throws IOException {
		File geneTreeFolder = new File(geneTreeFolderName);
		File[] geneTreeFiles = geneTreeFolder.listFiles();
		int ID = 0;
		for (File geneTreeFile: geneTreeFiles ) {
			Scanner scanner = new Scanner(geneTreeFile);
			String geneTreeName = getGeneTreeNameFromFile(geneTreeFile);
			String geneTreeFullName = getGeneTreeFullNameFromFile(geneTreeFile);
			String geneTreeString = scanner.nextLine();
			//System.out.println(geneTreeString);
			GeneTreeRooted geneTree = getGeneTreeFromFile(geneTreeString, ID, geneTreeName);
			setPreOrderNames(geneTree);
			String rangerGeneTreeFormat = alignRangerGeneTreeFormat(geneTree.getRoot(), ID);
			geneTree.setNewickOutputFormat(rangerGeneTreeFormat);
			geneTree.setID(ID);
			geneTree.setGeneTreeName(geneTreeFullName);
			reconciliation.addGeneTree(geneTree);
			ID++;
			scanner.close();
		}
	}

	private static String getGeneTreeNameFromFile(File geneTreeFile) {
		String fileName = geneTreeFile.getName();
		String [] partsOfName = fileName.split("/");
		String lastPart = partsOfName[partsOfName.length - 1];
		String geneTreeName = lastPart.replaceFirst("\\.", " ");
		 	geneTreeName = geneTreeName.split(" ")[0];
	//	System.out.println(geneTreeName);
		return geneTreeName;
	}

	private static String getGeneTreeFullNameFromFile(File geneTreeFile) {
		String fileName = geneTreeFile.getName();
		String [] partsOfName = fileName.split("/");
		String geneTreeName = partsOfName[partsOfName.length - 1];
		//String geneTreeName = lastPart.replaceFirst("\\.", " ");
		 //	geneTreeName = geneTreeName.split(" ")[0];
	//	System.out.println(geneTreeName);
		return geneTreeName;
	}
	private static void processSpeciesTree(String speciesTreeFileName, DTLReconciliation reconciliation)throws IOException  {
		Scanner scanner = new Scanner(new File(speciesTreeFileName));	
		String speciesString = scanner.nextLine();
		speciesString = speciesString.replaceAll("\\s", "");
		speciesString = speciesString.replaceAll(";","");
		SpeciesTreeRooted speciesTree = getSpeciesTreeFromFile(speciesString);
		setPreOrderNames(speciesTree);
		String rangerSpeciesTreeFormat = alignRangerSpeciesTreeFormat(speciesTree.getRoot());
		//System.out.println(rangerSpeciesTreeFormat);
		speciesTree.setNewickOutputFormat(rangerSpeciesTreeFormat);
		reconciliation.setSpeciesTree(speciesTree);
		scanner.close();
	}

	public static void processInput(String fileName, DTLReconciliation reconciliation) throws IOException {
		//byte[] encoded = Files.readAllBytes(Paths.get(fileName));
		//String s = new String(encoded, Charset.defaultCharset());
		
		Scanner scanner = new Scanner(new File(fileName));
		String s = "";
		while(scanner.hasNextLine()) {
			s = s + scanner.nextLine();
		}
		scanner.close();
		s = s.replaceAll("\\s", "");
		String[] splitPartsOfInput = s.split(";");
		if (splitPartsOfInput.length < 2) {
			reconciliation.setInputError(true);
			System.out.println("input error");
			System.out.println(fileName);
			return;
		}
		String speciesTreeString = splitPartsOfInput[0];
		String geneTreeString = splitPartsOfInput[1];

		if (geneTreeString.length() == 0 || speciesTreeString.length() == 0) {
			reconciliation.setInputError(true);
			System.out.println("input error");
			System.out.println(fileName);
			return;
		}
		SpeciesTreeRooted speciesTree = getSpeciesTreeFromFile(speciesTreeString);
	//	GeneTreeRooted geneTree = getGeneTreeFromFile(geneTreeString);

		setPreOrderNames(speciesTree);
		String rangerSpeciesTreeFormat = alignRangerSpeciesTreeFormat(speciesTree.getRoot());
		speciesTree.setNewickOutputFormat(rangerSpeciesTreeFormat);
		// System.out.println(rangerSpeciesTreeFormat);

	//	geneTree.computeAllPreOrderNumbers();
	//	String rangerGeneTreeFormat = alignRangerGeneTreeFormat(geneTree.getRoot());
	//	geneTree.setNewickOutputFormat(rangerGeneTreeFormat);
		// System.out.println(rangerGeneTreeFormat);

		reconciliation.setSpeciesTree(speciesTree);
////		reconciliation.setGeneTree(geneTree);
	}

	public static void setPreOrderNames(RootedTree speciesTree) {
		speciesTree.computeAllPreOrderNumbers();
		ArrayList<Node> allInternalPreOrderNodes = speciesTree.computeAllPreOrderInternalNodes();

		int index = 1;
		for (Node node : allInternalPreOrderNodes) {
			if( (node.getInternalNodeName()==null || node.getInternalNodeName().length()==0) &&  ( node instanceof GeneNodeRooted) )
			
			{	
				node.setName("m" + index);
				node.setInternalNodeName("m" + index+"_"+(((GeneNodeRooted)node).getGeneTreeName()));
			//	System.out.println(((GeneNodeRooted)node).getGeneTreeName());
			}
			if( (node.getInternalNodeName()==null || node.getInternalNodeName().contains("$"))  &&  ( node instanceof SpeciesNodeRooted) )
				
			{	
				node.setName("n" + index);
				node.setInternalNodeName("n" + index);
			}
			if ((node.getInternalNodeName()==null || node.getInternalNodeName().length()==0)&&  ( node instanceof DomainNodeRooted) )
				
			{	
				node.setName("d" + index);
				node.setInternalNodeName("d" + index);
			}
			index++;
			}
		// speciesTree.getAllNodesByPreOrder().clear();;
	}
	public static String alignRangerGeneTreeFormat(Node node, int ID) {
		//System.out.println(node);
		
		GeneNodeRooted geneNode = (GeneNodeRooted) node;
		if (node.isLeaf())
			return "("+geneNode.getFullName()+")";
		geneNode.setGeneTreeID(ID);
		String output = "";
		String leftPart = "";
		String rightPart = "";
		if (node.getLeft().isLeaf()) {
			GeneNodeRooted geneNodeLeft = (GeneNodeRooted) node.getLeft();
			leftPart = geneNodeLeft.getFullName();
		} else {
			leftPart = alignRangerGeneTreeFormat(node.getLeft(), ID);
		}
		if (node.getRight().isLeaf()) {
			GeneNodeRooted geneNodeRight = (GeneNodeRooted) node.getRight();
			rightPart = geneNodeRight.getFullName();
		} else {
			
				rightPart = alignRangerGeneTreeFormat(node.getRight(), ID);

		}
		output = "(" + leftPart + "," + rightPart + ")" + geneNode.getInternalNodeName();
		if (node.isRoot()) {

			output += ";";
		}

		return output;
	}


//create gene tree from input string (gene tree part)
	public static GeneTreeRooted getGeneTreeFromFile(String geneTreeString, int geneTreeID, String geneTreeName) {
		geneTreeString = geneTreeString.replaceAll("\\s", "");
		//geneTreeString = geneTreeString + ";";
		boolean flag = false;
		int length = geneTreeString.length();
		Stack<GeneNodeRooted> stack = new Stack<GeneNodeRooted>();
		String gene = "";
		GeneNodeRooted parent = null;
		for (int i = 0; i < length; i++) {
			
			char presentCharacter = geneTreeString.charAt(i);
			if (presentCharacter == '(' || presentCharacter == ',') {
				flag = false;
				continue;
			} else if (presentCharacter == ')') {
				GeneNodeRooted geneRight = stack.pop();
				GeneNodeRooted geneLeft = stack.pop();
				 parent = new GeneNodeRooted(geneLeft.getName() + "$" + geneRight.getName());
				 parent.setGeneTreeName(geneTreeName);
				 parent.setLeft(geneLeft);
				parent.setRight(geneRight);

				stack.push(parent);
				flag = true;
			} else {
				gene = "";
				while (geneTreeString.charAt(i) != ',' && geneTreeString.charAt(i) != ')'
						&& geneTreeString.charAt(i) != ';') {
					gene += geneTreeString.charAt(i);
					i++;
				}

				if (geneTreeString.charAt(i) == ')')
					i--;
				if (gene.contains(".")) {
					System.out.println(gene);
					String[] partsOfGeneName = gene.split(".");
					String geneName = partsOfGeneName[0];
					String realName = partsOfGeneName[1];
					GeneNodeRooted geneNode = new GeneNodeRooted(geneName);
					geneNode.setRealName(realName);
					if (!flag)
						stack.push(geneNode);
					else
					{
						parent.setInternalNodeName(gene);
						parent.setGeneTreeID(geneTreeID);
						parent.setGeneTreeName(geneTreeName);
						flag = false;
					}
				} 
				else if (gene.contains("_")) {
					String[] partsOfGeneName = gene.split("_");
					String geneName = partsOfGeneName[0];
					String speciesName = partsOfGeneName[1];
					String domainPart = "";
					if(partsOfGeneName.length==2) {
					
						speciesName = partsOfGeneName[1];
						domainPart = partsOfGeneName[0];
					}
					else if(partsOfGeneName.length==3) {
						geneName = gene;
					}
					
					GeneNodeRooted geneNode = new GeneNodeRooted(geneName);
					geneNode.setGeneTreeID(geneTreeID);
					geneNode.setGeneTreeName(geneTreeName);
					geneNode.setSpeciesName(speciesName);
					geneNode.setDomainPart(domainPart);
					if (!flag)
						stack.push(geneNode);
					else
						{
						flag = false;
						parent.setInternalNodeName(gene);
						}
				} else {
					GeneNodeRooted geneNode = new GeneNodeRooted(gene);
					geneNode.setGeneTreeID(geneTreeID);
					if (!flag)
						stack.push(geneNode);
					else
						{
						parent.setInternalNodeName(gene);						
						flag = false;
						}
				}
			}
		}
		if (stack.isEmpty())
			System.out.println("Tree: " + geneTreeString.length());
		GeneNodeRooted root = stack.pop();
		if (gene!=null && gene.length()>0)
			root.setInternalNodeName(gene);
		//System.out.println(root);
		GeneTreeRooted geneTree = new GeneTreeRooted(root);
		geneTree.setID(geneTreeID);
		return geneTree;
	}

	public static DomainTreeRooted getDomainTreeFromFile(String domainTreeString) {

		domainTreeString = domainTreeString.replaceAll("\\s", "");
		//geneTreeString = geneTreeString + ";";
		boolean flag = false;
		int length = domainTreeString.length();
		Stack<DomainNodeRooted> stack = new Stack<DomainNodeRooted>();
		String domain = "";
		DomainNodeRooted parent = null;
		for (int i = 0; i < length; i++) {
			
			char presentCharacter = domainTreeString.charAt(i);
			if (presentCharacter == '(' || presentCharacter == ',') {
				flag = false;
				continue;
			} else if (presentCharacter == ')') {
				DomainNodeRooted domainRight = stack.pop();
				DomainNodeRooted domainLeft = stack.pop();
				parent = new DomainNodeRooted(domainLeft.getName() + "$" + domainRight.getName());

		//		parent.setInternalNodeName(domain);
				parent.setLeft(domainLeft);
				parent.setRight(domainRight);

				stack.push(parent);
				flag = true;
			} else {
				domain = "";
				while (domainTreeString.charAt(i) != ',' && domainTreeString.charAt(i) != ')'
						&& domainTreeString.charAt(i) != ';') {
					domain += domainTreeString.charAt(i);
					i++;
				}

				if (domainTreeString.charAt(i) == ')')
					i--;
				if (domain.contains(".")) {
					String[] partsOfGeneName = domain.split(".");
					String geneName = partsOfGeneName[0];
					String realName = partsOfGeneName[1];
					DomainNodeRooted domainNode = new DomainNodeRooted(geneName);
					domainNode.setRealName(realName);
					if (!flag)
						stack.push(domainNode);
					else
					{
						parent.setInternalNodeName(domain);
						//parent.setGeneTreeID(geneTreeID);
						flag = false;
					}
				} else if (domain.contains("_")) {
					String[] partsOfDomainName = domain.split("_");
					DomainNodeRooted domainNode =null;
					if(partsOfDomainName.length==3) {
					String domainName = partsOfDomainName[0];
					String geneName = partsOfDomainName[1];
					String geneTreeName = partsOfDomainName[2];
					domainNode = new DomainNodeRooted(domainName);
						domainNode.setGeneTreeName(geneTreeName);
						domainNode.setGeneName(geneName);
						domainNode.setDomainName(domainName);
					}
					else if(partsOfDomainName.length==2) {
						String domainName = partsOfDomainName[0];
						String geneName = partsOfDomainName[1];
					
						domainNode = new DomainNodeRooted(domainName);
					
							domainNode.setGeneName(geneName);
							domainNode.setDomainName(domainName);
					}
					
					if (!flag)
						stack.push(domainNode);
					else
						{
						flag = false;
						parent.setInternalNodeName(domain);		
						}
				} else {
					DomainNodeRooted domainNode = new DomainNodeRooted(domain);
					//domainNode.setGeneTreeID(geneTreeID);
					if (!flag)
						stack.push(domainNode);
					else
						{
						parent.setInternalNodeName(domain);						
						flag = false;
						}
				}
			}
		}
		if (stack.isEmpty())
			System.out.println("Tree: " + domainTreeString.length());
		DomainNodeRooted root = stack.pop();
		root.setInternalNodeName(domain);
	//	System.out.println(root);
		DomainTreeRooted domainTree = new DomainTreeRooted(root);
		return domainTree;
	
	}

	
	
}
